package api

import (
	"bytes"
	"encoding/json"
	"net/http"
	"net/http/httptest"
	"testing"

	"git.inverse.ca/inverse/fingerbank-collector/common"
)

func TestListEndpointData(t *testing.T) {
	collector := fccommon.NewFingerbankCollector("")
	router := Setup(collector)

	// Create a few endpoints
	collector.FindEndpointDataMAC("00:11:22:33:44:55")
	collector.FindEndpointDataMAC("00:11:22:33:44:56")
	collector.FindEndpointDataMAC("00:11:22:33:44:57")

	req, _ := http.NewRequest("GET", "/endpoint_data", nil)
	recorder := httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	var eds []fccommon.ExpandedEndpointData

	resp := recorder.Result()

	if resp.StatusCode != 200 {
		t.Error("Wrong return code for endpoints data listing", resp.StatusCode)
	}

	dec := json.NewDecoder(resp.Body)
	err := dec.Decode(&eds)

	if err != nil {
		t.Errorf("Error while decoding response: %s", err)
	}

	if len(eds) != 3 {
		t.Errorf("Wrong length for endpoint data listing (%d)", len(eds))
	}
}

func TestGetEndpointData(t *testing.T) {
	collector := fccommon.NewFingerbankCollector("")
	router := Setup(collector)

	// Create a few endpoints
	collector.FindEndpointDataMAC("00:11:22:33:44:55")

	req, _ := http.NewRequest("GET", "/endpoint_data/00:11:22:33:44:55", nil)
	recorder := httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	var ed fccommon.EndpointData

	resp := recorder.Result()

	if resp.StatusCode != 200 {
		t.Error("Wrong return code for endpoints data GET", resp.StatusCode)
	}

	dec := json.NewDecoder(resp.Body)
	err := dec.Decode(&ed)

	if err != nil {
		t.Errorf("Error while decoding response: %s", err)
	}

	if ed.OUI != "001122" {
		t.Error("Wrong OUI for endpoint", ed.OUI)
	}

}

func TestUpdateEndpointData(t *testing.T) {
	collector := fccommon.NewFingerbankCollector("")
	router := Setup(collector)

	// Create an endpoint
	ed1, _ := collector.FindEndpointDataMAC("00:11:22:33:44:55")

	ed1.Update(func(ed *fccommon.EndpointData) bool {
		ed.DhcpFingerprint = "1,2,3,4"
		return true
	})

	// Updating an existing endpoint by MAC
	req, _ := http.NewRequest("PATCH", "/endpoint_data/00:11:22:33:44:55", bytes.NewBuffer([]byte(`{"most_accurate_user_agent":"test"}`)))
	recorder := httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	resp := recorder.Result()

	if resp.StatusCode != 200 {
		t.Errorf("Invalid status code returned while updating endpoint %d", resp.StatusCode)
	}

	if ed1.MostAccurateUserAgent != "test" {
		t.Error("User agent wasn't updated by update query")
	}

	if ed1.DhcpFingerprint != "1,2,3,4" {
		t.Error("DHCP fingerprint was altered even though it shouldn't have been touched")
	}

	collector.UpdateIPMAC("1.2.3.4", "00:11:22:33:44:55")

	// Updating an existing endpoint by IP
	req, _ = http.NewRequest("PATCH", "/endpoint_data/1.2.3.4", bytes.NewBuffer([]byte(`{"most_accurate_user_agent":"test2"}`)))
	recorder = httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	resp = recorder.Result()

	if resp.StatusCode != 200 {
		t.Errorf("Invalid status code returned while updating endpoint %d", resp.StatusCode)
	}

	if ed1.MostAccurateUserAgent != "test2" {
		t.Error("User agent wasn't updated by update query")
	}

	if ed1.DhcpFingerprint != "1,2,3,4" {
		t.Error("DHCP fingerprint was altered even though it shouldn't have been touched")
	}

	// Updating an endpoint by IP that doesn't exist should yield error
	req, _ = http.NewRequest("PATCH", "/endpoint_data/0.0.0.0", bytes.NewBuffer([]byte(`{"user_agent":"test2"}`)))
	recorder = httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	resp = recorder.Result()

	if resp.StatusCode != 404 {
		t.Errorf("Invalid status code returned while updating endpoint %d", resp.StatusCode)
	}

	// Updating an endpoint with a random device ID should yield error
	req, _ = http.NewRequest("PATCH", "/endpoint_data/zaymebabwe", bytes.NewBuffer([]byte(`{"user_agent":"test2"}`)))
	recorder = httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	resp = recorder.Result()

	if resp.StatusCode != 404 {
		t.Errorf("Invalid status code returned while updating endpoint %d", resp.StatusCode)
	}
}

func TestDeleteEndpointData(t *testing.T) {
	collector := fccommon.NewFingerbankCollector("")
	router := Setup(collector)

	// Create an endpoint
	ed, _ := collector.FindEndpointDataMAC("00:11:22:33:44:55")
	ed.MostAccurateUserAgent = "test"

	req, _ := http.NewRequest("DELETE", "/endpoint_data/00:11:22:33:44:55", nil)
	recorder := httptest.NewRecorder()
	router.ServeHTTP(recorder, req)

	resp := recorder.Result()

	if resp.StatusCode != 200 {
		t.Errorf("Wrong return code for endpoints data listing %d", resp.StatusCode)
	}

	ed, _ = collector.FindEndpointDataMAC("00:11:22:33:44:55")

	if ed.MostAccurateUserAgent != "" {
		t.Error("User-Agent is still filled even though the endpoint was deleted")
	}
}
