package boot_sequence

import (
	"sort"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/fvbommel/sortorder"
)

const DefaultTrackingDuration = 1 * time.Minute

//easyjson:json
type BootSequence struct {
	lock       *sync.Mutex
	Events     map[string]bool `json:"events"`
	StartTime  time.Time       `json:"start_time"`
	timer      *time.Timer
	seqStarted bool
}

func NewBootSequence() *BootSequence {
	bs := &BootSequence{}
	bs.Events = make(map[string]bool)
	bs.Init()
	return bs
}

func (bs *BootSequence) Init() {
	bs.timer = time.NewTimer(0)
	bs.lock = &sync.Mutex{}
	if !bs.timer.Stop() {
		<-bs.timer.C
	}
	if bs.Events == nil {
		bs.Events = map[string]bool{}
	}
}

func (bs *BootSequence) ResetLock() {
	bs.lock = &sync.Mutex{}
}

func (bs *BootSequence) Start() {
	bs.StartWithDuration(DefaultTrackingDuration)
}

func (bs *BootSequence) StartWithDuration(dur time.Duration) {
	bs.lock.Lock()
	defer bs.lock.Unlock()
	if !bs.timer.Stop() {
		select {
		case <-bs.timer.C:
		default:
		}
	}

	now := time.Now()
	bs.Events = map[string]bool{
		strconv.FormatInt(now.Unix()*1000, 10) + ":" + "boot_sequence_started": true,
	}
	bs.timer.Reset(dur)
	bs.seqStarted = true
	bs.StartTime = now
}

func (bs *BootSequence) SoftStart() {
	bs.SoftStartWithDuration(DefaultTrackingDuration)
}

func (bs *BootSequence) SoftStartWithDuration(dur time.Duration) {
	bs.checkSeqIsOver()
	if bs.seqStarted {
		return
	}

	bs.StartWithDuration(dur)
}

func (bs *BootSequence) IsInBootSeq() bool {
	return bs.seqStarted
}

func (bs *BootSequence) checkSeqIsOver() {
	select {
	case <-bs.timer.C:
		bs.seqStarted = false
	default:
	}
}

func (bs *BootSequence) Add(events ...string) bool {
	bs.lock.Lock()
	defer bs.lock.Unlock()
	bs.checkSeqIsOver()
	if bs.seqStarted {
		bs.alwaysAdd(events)
		return true
	}

	return false
}

func (bs *BootSequence) AlwaysAdd(events []string) {
	bs.lock.Lock()
	defer bs.lock.Unlock()
	bs.alwaysAdd(events)
}

func (bs *BootSequence) alwaysAdd(events []string) {
	for _, e := range events {
		bs.Events[e] = true
	}
}

func (bs *BootSequence) GetFirstN(n int) []string {
	bs.lock.Lock()
	defer bs.lock.Unlock()
	l := len(bs.Events)
	if n > l {
		n = l
	}
	cp := make([]string, l, l)
	i := 0
	for e, _ := range bs.Events {
		cp[i] = e
		i++
	}
	sort.Sort(sortorder.Natural(cp))
	return cp[:n]
}

func ParseEvent(event string) (int64, string) {
	pieces := strings.SplitN(event, ":", 2)
	switch len(pieces) {
	case 0:
		return 0, ""
	case 1:
		return 0, event
	default:
		ts, err := strconv.ParseInt(pieces[0], 10, 64)
		if err != nil {
			return 0, event
		}
		return ts, pieces[1]
	}
}
