package fccommon

import (
	"reflect"
	"strconv"
	"strings"

	"git.inverse.ca/inverse/fingerbank-collector/elru"
	"git.inverse.ca/inverse/fingerbank-collector/timedlock"
	"github.com/inverse-inc/go-utils/sharedutils"
)

type AllHostsCache struct {
	cache *elru.EncodableLRU
	lock  *timedlock.RWLock
}

func NewAllHostsCache(maxSize int) *AllHostsCache {
	cache, err := elru.NewEncodableLRU(maxSize)
	sharedutils.CheckError(err)

	ahc := &AllHostsCache{
		cache: cache,
		lock:  timedlock.NewRWLock(),
	}
	ahc.lock.Panic = true

	return ahc
}

func (ahc *AllHostsCache) Add(host string, port string) bool {
	lockid := ahc.lock.Lock()
	defer ahc.lock.Unlock(lockid)

	var m map[string]int
	var foundHost bool
	if m, foundHost = ahc.Get(host); !foundHost {
		m = map[string]int{}
	}

	_, foundPort := m[port]
	m[port]++

	ahc.cache.Add(host, m)
	return foundHost && foundPort
}

func (ahc *AllHostsCache) Get(host string) (map[string]int, bool) {
	if o, found := ahc.cache.Get(host); found {
		return o.(map[string]int), found
	} else {
		return map[string]int{}, found
	}
}

func (ahc *AllHostsCache) ToMap() map[string]map[string]int {
	m := map[string]map[string]int{}
	for _, k := range ahc.cache.Keys() {
		v, found := ahc.cache.Get(k)
		// The entry got expired by the time we fetched it
		if !found {
			continue
		}
		m[k.(string)] = v.(map[string]int)
	}
	return m
}

func (ahc *AllHostsCache) ToLegacyMap() map[string]map[int]int {
	m := ahc.ToMap()
	m2 := map[string]map[int]int{}
	for k, ports := range m {
		m2[k] = map[int]int{}
		for protoPortStr, counter := range ports {
			portStrParts := strings.Split(protoPortStr, ":")
			portStr := portStrParts[len(portStrParts)-1]
			portInt, err := strconv.Atoi(portStr)
			sharedutils.CheckError(err)
			m2[k][portInt] = counter
		}
	}
	return m2
}

func (ahc *AllHostsCache) MarshalJSON() ([]byte, error) {
	return ahc.cache.MarshalJSON()
}

func (ahc *AllHostsCache) UnmarshalJSON(data []byte) error {
	ahc.cache = elru.NewEncodableLRUForUnmarshal(reflect.TypeOf(map[string]int{}))
	ahc.lock = timedlock.NewRWLock()
	return ahc.cache.UnmarshalJSON(data)
}

func (ahc *AllHostsCache) RLock() uint64 {
	return ahc.lock.RLock()
}

func (ahc *AllHostsCache) RUnlock(lockid uint64) {
	ahc.lock.RUnlock(lockid)
}
