package fccommon

import (
	"bytes"
	"crypto/hmac"
	"crypto/sha256"
	"crypto/tls"
	"encoding/hex"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"io/ioutil"
	"net"
	"net/http"
	"strconv"
	"strings"
	"sync"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"

	"github.com/gin-gonic/gin"
	"github.com/go-ini/ini"
	"github.com/inverse-inc/go-utils/sharedutils"
	cache "github.com/patrickmn/go-cache"
)

var analysisLimitLock = &sync.Mutex{}
var analysisLimitCache = cache.New(5*time.Second, 1*time.Minute)

var nbaEnabled = flag.Bool(
	"network-behavior-analysis",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_NETWORK_BEHAVIOR_ANALYSIS", "false")),
	"Enables network behavior analysis of the endpoints.",
)

var nbaPoliciesLocation = flag.String(
	"network-behavior-policies",
	sharedutils.EnvOrDefault("COLLECTOR_NETWORK_BEHAVIOR_POLICIES", "/dev/null"),
	"The location of the network behavior policies configuration",
)

var endpointAnalysisWebhook = flag.String(
	"endpoint-analysis-webhook",
	sharedutils.EnvOrDefault("COLLECTOR_ENDPOINT_ANALYSIS_WEBHOOK", ""),
	"Where to send the web hooks when detecting a network behavior event event through the NBA analysis. Needs network-behavior-analysis to be enabled. Must be a valid HTTP/HTTPS URL. The validity of SSL will not be checked. Example: 'https://myserver.acme.com/collector_hook.php'. (COLLECTOR_ENDPOINT_ANALYSIS_WEBHOOK_USERNAME environment variable)",
)

var endpointAnalysisWebhookUsername = flag.String(
	"endpoint-analysis-webhook-username",
	sharedutils.EnvOrDefault("COLLECTOR_ENDPOINT_ANALYSIS_WEBHOOK_USERNAME", ""),
	"An optional username to send in the webhooks using HTTP basic auth. (COLLECTOR_ENDPOINT_ANALYSIS_WEBHOOK_USERNAME environment variable)",
)

var endpointAnalysisWebhookPassword = flag.String(
	"endpoint-analysis-webhook-password",
	sharedutils.EnvOrDefault("COLLECTOR_ENDPOINT_ANALYSIS_WEBHOOK_PASSWORD", ""),
	"An optional username to send in the webhooks using HTTP basic auth. (COLLECTOR_ENDPOINT_ANALYSIS_WEBHOOK_PASSWORD environment variable)",
)

func (fc *FingerbankCollector) buildNBAPoliciesFromFile(fname string) ([]*NBAPolicy, error) {
	var err error

	dat, err := ioutil.ReadFile(fname)
	if err != nil {
		return []*NBAPolicy{}, err
	}

	nbaPolicies, err := fc.buildNBAPolicies(dat)

	return nbaPolicies, err
}

func (fc *FingerbankCollector) buildNBAPolicies(iniContent []byte) ([]*NBAPolicy, error) {
	policies := []*NBAPolicy{}
	f, err := ini.Load(iniContent)
	if err != nil {
		return []*NBAPolicy{}, err
	}

	for _, section := range f.Sections() {
		if section.Name() == "DEFAULT" {
			continue
		}

		pol := new(NBAPolicy)
		err := section.MapTo(pol)

		if err != nil {
			return []*NBAPolicy{}, err
		}

		if pol.Status == "disabled" {
			continue
		}

		err = fc.buildNBAPolicyFromSection(section, pol)
		if err != nil {
			return []*NBAPolicy{}, err
		}
		policies = append(policies, pol)
	}

	return policies, nil
}

func (fc *FingerbankCollector) buildNBAPolicyFromSection(section *ini.Section, pol *NBAPolicy) error {
	newDevices := []int{}
	for _, deviceId := range pol.DevicesIncluded {
		newDevices = append(newDevices, deviceId)
		childsIds, err := fc.DeviceAllChildsIds(deviceId)
		if err != nil {
			return err
		}
		newDevices = append(newDevices, childsIds...)
	}
	pol.DevicesIncluded = newDevices

	newDevices = []int{}
	for _, deviceId := range pol.DevicesExcluded {
		newDevices = append(newDevices, deviceId)
		childsIds, err := fc.DeviceAllChildsIds(deviceId)
		if err != nil {
			return err
		}
		newDevices = append(newDevices, childsIds...)
	}
	pol.DevicesExcluded = newDevices

	pol.WatchedDeviceAttributes = map[string]bool{}
	for _, attr := range strings.Split(section.Key("watched_device_attributes").String(), ",") {
		if attr != "" {
			attr = strings.Trim(attr, " ")
			pol.WatchedDeviceAttributes[attr] = true
		}
	}

	pol.DeviceAttributesDiffThresholdOverrides = map[string]int{}
	for _, data := range strings.Split(section.Key("device_attributes_diff_threshold_overrides").String(), ",") {
		if data != "" {
			data = strings.Trim(data, " ")
			info := strings.Split(data, ":")
			if len(info) != 2 {
				fc.logger.Error("Invalid device_attributes_diff_threshold_overrides: " + data)
				continue
			}

			attr := info[0]
			if _, found := DIFF_THRESHOLD_HITS[attr]; !found {
				fc.logger.Error("Invalid data attribute in device_attributes_diff_threshold_overrides:" + attr)
				continue
			}

			diff, err := strconv.Atoi(info[1])
			if err != nil {
				fc.logger.Error(fmt.Sprintf("Invalid numerical threshold %s for attribute %s", info[1], attr))
				continue
			}

			pol.DeviceAttributesDiffThresholdOverrides[attr] = diff
		}
	}

	pol.WhitelistedIPs = []*net.IPNet{}
	for _, attr := range strings.Split(section.Key("whitelisted_ips").String(), ",") {
		if attr != "" {
			attr = strings.Trim(attr, " ")
			_, ipnet, err := net.ParseCIDR(attr)
			if err != nil {
				return err
			}
			pol.WhitelistedIPs = append(pol.WhitelistedIPs, ipnet)
		}
	}

	pol.BlacklistedPorts = map[uint16]bool{}
	for _, attr := range strings.Split(section.Key("blacklisted_ports").String(), ",") {
		if attr != "" {
			attr = strings.Trim(attr, " ")
			data := strings.Split(attr, "-")
			if len(data) == 1 {
				port, err := strconv.Atoi(data[0])
				if err != nil {
					fc.logger.Error(fmt.Sprintf("Invalid blacklisted port number %s", data[0]))
				}
				pol.BlacklistedPorts[uint16(port)] = true
			} else {
				begin, err := strconv.Atoi(data[0])
				if err != nil {
					fc.logger.Error(fmt.Sprintf("Invalid blacklisted port number %s", data[0]))
				}
				end, err := strconv.Atoi(data[1])
				if err != nil {
					fc.logger.Error(fmt.Sprintf("Invalid blacklisted port number %s", data[0]))
				}
				for port := begin; port <= end; port++ {
					pol.BlacklistedPorts[uint16(port)] = true
				}
			}
		}
	}

	pol.PolicyID = section.Name()

	return nil
}

func (fc *FingerbankCollector) AnalyseAllEndpoints() {
	for _, mac := range fc.ListEndpointsIds() {
		ed, _ := fc.FindEndpointDataMAC(mac)
		fc.AnalyseEndpoint(ed)
	}
}

func (fc *FingerbankCollector) GetNetworkProfile(deviceId int) (fingerbankclient.NetworkProfile, error) {
	cacheKey := fmt.Sprintf("fingerbank-api-network-profile-%d", deviceId)
	if o, found := Cache.Get(cacheKey); found {
		fc.logger.Debug("Returning network profile from cache")
		return *(o.(*fingerbankclient.NetworkProfile)), nil
	}

	prof, err := fingerbankclient.GetNetworkProfile(deviceId)
	if err != nil {
		return fingerbankclient.NetworkProfile{}, errors.New(fmt.Sprintf("Cannot communicate with Fingerbank API to fetch network profile of device %d (%s)", deviceId, err))
	}

	Cache.Set(cacheKey, &prof, 24*time.Hour)

	return prof, nil
}

func (fc *FingerbankCollector) DeviceAllChildsIds(deviceId int) ([]int, error) {
	cacheKey := fmt.Sprintf("fingerbank-api-all-childs-ids-%d", deviceId)
	if o, found := Cache.Get(cacheKey); found {
		fc.logger.Debug("Returning child ids from cache")
		return *(o.(*[]int)), nil
	}

	childsIds, err := fingerbankclient.DeviceVirtualChildsIds(deviceId)
	if err != nil {
		return []int{}, errors.New(fmt.Sprintf("Cannot communicate with Fingerbank API to get childs ids of device %d (%s)", deviceId, err))
	}

	Cache.Set(cacheKey, &childsIds, 24*time.Hour)

	return childsIds, nil
}

func (fc *FingerbankCollector) AnalyseEndpoint(ed *EndpointData) ([]string, error) {
	k := ed.mac

	analysisLimitLock.Lock()
	_, found := analysisLimitCache.Get(k)

	if !found {
		analysisLimitCache.Set(k, true, cache.DefaultExpiration)
		analysisLimitLock.Unlock()
		return fc.analyseEndpoint(ed)
	} else {
		analysisLimitLock.Unlock()
		return []string{}, errors.New("Rate limited")
	}
}

func (fc *FingerbankCollector) analyseEndpoint(ed *EndpointData) ([]string, error) {
	if fc.NBAEnabled {
		fc.logger.Info("Analysing endpoint " + ed.mac)
		if res, err := fc.InterrogateAPI(ed.mac, ed); err == nil {
			if res.CloudAPIResult.Device.Id == 0 {
				msg := "Cannot perform analysis of device since we weren't able to profile it correctly"
				fc.logger.Error(msg)
				return []string{}, errors.New(msg)
			}

			prof, err := fc.GetNetworkProfile(res.CloudAPIResult.Device.Id)
			if err != nil {
				fc.logger.Error(fmt.Sprintf("Unable to fetch device profile for device ID %d: %s", res.CloudAPIResult.Device.Id, err))
				return []string{}, err
			}

			events := []string{}
			for _, pol := range fc.nbaPolicies {
				nba := NewNetworkProfileEngine(prof, pol)
				polEvents := nba.AnalyseEndpoint(fc, &res)
				events = append(events, polEvents...)
				if *endpointAnalysisWebhook != "" && len(polEvents) > 0 {
					go fc.performEndpointAnalysisWebhook(*endpointAnalysisWebhook, ed, gin.H{"type": "events", "events": polEvents, "policy": pol.PolicyID})
				}
			}

			return events, nil
		} else {
			fc.logger.Warn(fmt.Sprintf("Error while querying API to analyze hosts of endpoint %s. Error: %s", ed.mac, err))
			return []string{}, err
		}
	} else {
		return []string{}, errors.New("Network Behavior Analysis is not enabled")
	}
}

var webhookHttpClient *http.Client

func init() {
	tr := &http.Transport{
		TLSClientConfig: &tls.Config{InsecureSkipVerify: true},
	}
	webhookHttpClient = &http.Client{Transport: tr}
}

func (fc *FingerbankCollector) performEndpointAnalysisWebhook(url string, ed *EndpointData, data interface{}) {
	id := ed.RLock()

	body, err := json.Marshal(gin.H{"mac": ed.mac, "data": data})
	sharedutils.CheckError(err)

	hmacApiKey := hmac.New(sha256.New, []byte(fc.apiKey))
	hmacApiKey.Write(body)
	bodySum := hex.EncodeToString(hmacApiKey.Sum(nil))

	req, _ := http.NewRequest("POST", url, bytes.NewReader(body))
	req.Header.Set("X-Payload-Signature", bodySum)
	req.Header.Set("Content-Type", "application/json")
	req.SetBasicAuth(*endpointAnalysisWebhookUsername, *endpointAnalysisWebhookPassword)

	ed.RUnlock(id)

	res, err := webhookHttpClient.Do(req)
	if err != nil {
		fc.logger.Error("Error while POSTing host events report to " + url + ". Got error: " + err.Error())
	} else {
		fc.logger.Debug("Sent analysis events report to " + url)
		res.Body.Close()
	}
}

//
// Since we're moving to NBA I commented the code below since it was never actually used
//
//func (fc *FingerbankCollector) AnalyseEndpoint(ed *EndpointData) ([]host_analyser.InvalidHostResult, error) {
//	if res, err := fc.InterrogateAPI(ed.mac, ed); err == nil {
//		return fc.AnalyseQueryResult(res), nil
//	} else {
//		fc.logger.Warn(fmt.Sprintf("Error while querying API to analyze hosts of endpoint %s. Error: %s", ed.mac, err))
//		return []host_analyser.InvalidHostResult{}, err
//	}
//}
//
//func (fc *FingerbankCollector) AnalyseQueryResult(qr QueryResult) []host_analyser.InvalidHostResult {
//	analyser := host_analyser.DefaultHostAnalyser
//	fullHits := []host_analyser.InvalidHostResult{}
//
//	for _, id := range analyser.AffectsDevices() {
//		if is, err := fc.DeviceIsA(qr.CloudAPIResult.Device.Id, id); is {
//			hits := analyser.AnalyseHosts(id, qr.EndpointData.EndpointData.DestinationHosts)
//			fullHits = append(fullHits, hits...)
//			for _, hit := range hits {
//				fc.logger.Warn(fmt.Sprintf("Detected device (ip:%s|mac:%s) triggering event %s on destination host %s. Message: %s", qr.EndpointData.IP, qr.EndpointData.MAC, hit.Type, hit.Host, hit.Message))
//			}
//		} else if err != nil {
//			fc.logger.Warn("Error while determining device type in the analysis of the query result: " + err.Error())
//		}
//	}
//
//	if *webHookHostsEvents != "" {
//		body, err := json.Marshal(fc.HostAnalysisPayload(fullHits))
//		sharedutils.CheckError(err)
//		_, err = http.Post(*webHookHostsEvents, "application/json", bytes.NewReader(body))
//		if err != nil {
//			fc.logger.Error("Error while POSTing host events report to " + *webHookHostsEvents + ". Got error: " + err.Error())
//		} else {
//			fc.logger.Debug("Sent host events report to " + *webHookHostsEvents)
//		}
//	}
//
//	return fullHits
//}
//
//func (fc *FingerbankCollector) HostAnalysisPayload(hits []host_analyser.InvalidHostResult) interface{} {
//	var message string
//	if len(hits) == 0 {
//		message = "No problem was found while analyzing the hosts of this endpoint"
//	} else {
//		message = fmt.Sprintf("Found %d destination hosts problems while analysing the hosts of this endpoint", len(hits))
//	}
//	return map[string]interface{}{"message": message, "results": hits}
//}
