package fccommon

import (
	"fmt"
	"net"
	"sync"
	"time"

	cache "github.com/patrickmn/go-cache"
)

var blacklistingLimitLock = &sync.Mutex{}
var blacklistingLimitCache = cache.New(1*time.Minute, 1*time.Minute)

func (fc *FingerbankCollector) MatchAgainstBlacklist(srcIP, dstIP net.IP, srcPort, dstPort int) {
	found := map[string]bool{}
	for _, ipnet := range fc.BlacklistedIPs {
		if ipnet.Contains(srcIP) {
			if ed, err := fc.FindEndpointDataIP(dstIP.String()); err == nil {
				if fc.AddBlacklistedCommunication(ed, dstIP.String(), dstPort) {
					fc.logger.Info(dstIP.String() + " talking to blacklisted IP " + srcIP.String())
					ed.AddBlacklistedCommunication(srcIP.String(), srcPort)
					found[ed.mac] = true
				}
			}
		}
		if ipnet.Contains(dstIP) {
			if ed, err := fc.FindEndpointDataIP(srcIP.String()); err == nil {
				if fc.AddBlacklistedCommunication(ed, dstIP.String(), dstPort) {
					fc.logger.Info(srcIP.String() + " talking to blacklisted IP " + dstIP.String())
					found[ed.mac] = true
				}
			}
		}
	}

	for _, policy := range fc.nbaPolicies {
		if policy.BlacklistedPorts[uint16(dstPort)] {
			if ed, err := fc.FindEndpointDataIP(srcIP.String()); err == nil {
				if fc.AddBlacklistedCommunication(ed, BlacklistedCommunicationsPortsID, dstPort) {
					fc.logger.Info(fmt.Sprintf("%s talking to blacklisted port %d", srcIP, dstPort))
					found[ed.mac] = true
				}
			}
		}
		if policy.BlacklistedPorts[uint16(srcPort)] {
			if ed, err := fc.FindEndpointDataIP(dstIP.String()); err == nil {
				if fc.AddBlacklistedCommunication(ed, BlacklistedCommunicationsPortsID, srcPort) {
					fc.logger.Info(fmt.Sprintf("%s talking to blacklisted port %d", dstIP, srcPort))
					found[ed.mac] = true
				}
			}
		}
	}

	for mac, _ := range found {
		ed, _ := fc.FindEndpointDataMAC(mac)
		go fc.AnalyseEndpoint(ed)
	}
}

func (fc *FingerbankCollector) AddBlacklistedCommunication(ed *EndpointData, ip string, port int) bool {
	k := fmt.Sprintf("%s-%d", ip, port)

	blacklistingLimitLock.Lock()
	_, found := blacklistingLimitCache.Get(k)

	if !found {
		blacklistingLimitCache.Set(k, true, cache.DefaultExpiration)
		blacklistingLimitLock.Unlock()
		ed.AddBlacklistedCommunication(ip, port)
		return true
	} else {
		blacklistingLimitLock.Unlock()
		return false
	}
}
