package fccommon

import (
	"encoding/json"
	"fmt"
	"io"
	"os"
	"time"

	"github.com/mailru/easyjson/jwriter"
)

func (fc *FingerbankCollector) RestoreIpMaps(filename string, newerThan time.Duration) error {
	file, err := os.Open(filename)
	if err != nil {
		return err
	}
	defer file.Close()

	fileInfo, err := file.Stat()
	if err != nil {
		return err
	}

	if fileInfo.ModTime().Add(newerThan).Before(time.Now()) {
		fc.logger.Info(fmt.Sprintf("IP database %s is older than %s. Will not restore from it", filename, newerThan))
		return nil
	}

	fc.loadIpMaps(file)

	return err
}

func (fc *FingerbankCollector) PersistIpMaps(filename string) error {
	tmpFilename := filename + ".collector-tmp"
	file, err := os.Create(tmpFilename)
	defer file.Close()

	if err != nil {
		return err
	}

	maps := fc.IpMaps()

	enc := json.NewEncoder(file)
	err = enc.Encode(maps)

	if err != nil {
		return err
	}

	err = os.Rename(tmpFilename, filename)

	return err
}

func (fc *FingerbankCollector) PersistEndpointsCache(filename string) error {
	tmpFilename := filename + ".collector-tmp"
	file, err := os.Create(tmpFilename)
	defer file.Close()

	if err != nil {
		return err
	}

	enc := json.NewEncoder(file)
	err = enc.Encode(fc.EndpointsCache)

	if err != nil {
		return err
	}

	err = os.Rename(tmpFilename, filename)

	return err
}

func (fc *FingerbankCollector) RestoreEndpointsCache(filename string) error {
	file, err := os.Open(filename)
	if err != nil {
		return err
	}
	defer file.Close()

	fc.loadEndpointsCache(file)

	return err
}

func (fc *FingerbankCollector) PersistEndpointsData(filename string) error {
	tmpFilename := filename + ".collector-tmp"
	file, err := os.Create(tmpFilename)
	defer file.Close()

	if err != nil {
		return err
	}

	fc.WriteEndpointsData(file)

	err = os.Rename(tmpFilename, filename)

	return err
}

func (fc *FingerbankCollector) WriteEndpointsData(w io.Writer) {
	writer := jwriter.Writer{}

	keys := fc.ListEndpointsIds()

	io.WriteString(w, "{")
	for i, k := range keys {
		lockid := fc.EndpointsDataRLock()
		ed := fc.EndpointsData[k]
		fc.EndpointsDataRUnlock(lockid)

		if ed == nil {
			continue
		}

		lockid2 := ed.RLock()
		io.WriteString(w, `"`+k+`":`)
		ed.MarshalEasyJSON(&writer)
		ed.RUnlock(lockid2)
		writer.Buffer.DumpTo(w)

		if i != len(keys)-1 {
			io.WriteString(w, ",")
		}
	}
	io.WriteString(w, "}")

}

func (fc *FingerbankCollector) RestoreEndpointsData(filename string) error {
	file, err := os.Open(filename)
	if err != nil {
		return err
	}
	defer file.Close()

	fc.loadEndpointsData(file)

	return nil
}
