package fccommon

import (
	"bytes"
	"encoding/hex"
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"net"
	"net/http"
	"strings"
	"time"

	fcreplication "git.inverse.ca/inverse/fingerbank-collector/common/replication"
	"github.com/google/uuid"
	"github.com/inverse-inc/go-utils/sharedutils"
)

type ActionMeta struct {
	ActionID       uint8
	DecodeInStruct interface{}
	Handler        fcreplication.ActionHandler
}

func (fc *FingerbankCollector) ReplicationControllerMeta() []ActionMeta {
	return []ActionMeta{
		ActionMeta{
			ActionID:       fcreplication.REP_ENDPOINT_DATA_UPDATE,
			DecodeInStruct: fcreplication.REPEndpointDataUpdate{},
			Handler:        fc.HandleREPEndpointDataUpdate,
		},
		ActionMeta{
			ActionID:       fcreplication.REP_ENDPOINT_DATA_REPLACE,
			DecodeInStruct: fcreplication.REPEndpointDataReplace{},
			Handler:        fc.HandleREPEndpointDataReplace,
		},
		ActionMeta{
			ActionID:       fcreplication.REP_ENDPOINT_DATA_DELETE,
			DecodeInStruct: fcreplication.REPEndpointDataDelete{},
			Handler:        fc.HandleREPEndpointDataDelete,
		},
		ActionMeta{
			ActionID:       fcreplication.REP_IP_MAC_UPDATE,
			DecodeInStruct: fcreplication.REPIpMacUpdate{},
			Handler:        fc.HandleREPIpMacUpdate,
		},
		ActionMeta{
			ActionID:       fcreplication.REP_ADD_IP_HOST_MAPPING,
			DecodeInStruct: fcreplication.REPAddIpHostMapping{},
			Handler:        fc.HandleREPAddIpHostMapping,
		},
		ActionMeta{
			ActionID:       fcreplication.REP_ADD_DESTINATION_HOST,
			DecodeInStruct: fcreplication.REPAddDestinationHost{},
			Handler:        fc.HandleREPAddDestinationHost,
		},
		ActionMeta{
			ActionID:       fcreplication.REP_RESET_ENDPOINT_CACHE,
			DecodeInStruct: fcreplication.REPResetEndpointCache{},
			Handler:        fc.HandleREPResetEndpointCache,
		},
	}
}

func (fc *FingerbankCollector) BuildReplicationController(repKey []byte, peers []net.IP) *fcreplication.Controller {
	replicationController := fcreplication.NewController(uuid.New(), repKey, peers, *clusterHelloBroadcast)
	replicationController.HelloInterval = *clusterHelloInterval
	for _, meta := range fc.ReplicationControllerMeta() {
		replicationController.ActionMap[meta.ActionID] = meta.Handler
	}
	return replicationController
}

func (fc *FingerbankCollector) setupReplication() {
	repKey, err := hex.DecodeString(fc.apiKey)
	sharedutils.CheckError(err)
	// We ensure that the key is 32 bytes by adding 0s to it
	for i := len(repKey); i < 32; i++ {
		repKey = append(repKey, 0)
	}

	peers := []net.IP{}
	for _, peer := range strings.Split(*contactPeers, ",") {
		peers = append(peers, net.ParseIP(peer))
	}

	fc.ReplicationController = fc.BuildReplicationController(repKey, peers)
	fc.ReplicationController.StartReplication()

	if *odaPeersDiscoveryURL != "" {
		go fc.ReplicationController.DiscoverOdaPeers(*odaPeersDiscoveryURL)
	}

	// Asynchronously we find the oldest peer and sync from it
	go func() {
		initialSyncDone := false
		var lastPeerSyncSuccess time.Time

		for {
			var peerUUID string
			var peer *fcreplication.FingerbankCollectorPeer

			if initialSyncDone {
				if *clusterResyncInterval == 0 {
					fc.logger.Info("Not configured to perform regular resyncs. Disabling them now.")
					return
				}

				time.Sleep(*clusterResyncInterval)

				// If we've detected a cluster change after the last successful sync
				if fc.ReplicationController.GetLastPeerChange().After(lastPeerSyncSuccess) {
					fc.logger.Warn(fmt.Sprintf("Detected a recent peer change (at %s). Will perform a resync with the oldest peer.", fc.ReplicationController.GetLastPeerChange()))
				} else {
					fc.logger.Debug("No recent peer change, won't perform a resync")
					continue
				}

				peerUUID, peer = fc.ReplicationController.FindOldestPeer()

				if peer == nil {
					fc.logger.Debug("Unable to find any peer to do a continious sync from")
				} else if peer.Uptime() <= fc.ReplicationController.Uptime() {
					fc.logger.Debug("The oldest peer that was found has a smaller uptime than this collector. This collector should be used as the source of truth so not doing a sync from anyone.")
					peerUUID = ""
					peer = nil
				}
			} else {
				var peerPtr *fcreplication.FingerbankCollectorPeer
				peerUUID, peerPtr = fc.ReplicationController.FindOldestPeerForInitialSync(*clusterInitialSyncWait)
				if peerPtr != nil {
					// Perform a copy so that if it ever gets expired while we attempt to sync, then we won't be using a nil value
					peerCopy := *peerPtr
					peer = &peerCopy
				}
			}

			if peer == nil {
				fc.logger.Info("Unable to find a peer to sync from")
				continue
			}

			fc.logger.Debug(fmt.Sprintf("Will attempt to sync from %s", peer.IP))

			err := fc.FullSyncFromPeer(peerUUID, peer)

			if err == nil {
				lastPeerSyncSuccess = time.Now()
				fc.logger.Info(fmt.Sprintf("Successfully synced from peer %s at %s", peer.IP, lastPeerSyncSuccess))
				initialSyncDone = true
				continue
			}
		}
	}()
}

func (fc *FingerbankCollector) FullSyncFromPeer(peerUUID string, peer *fcreplication.FingerbankCollectorPeer) error {
	startedAt := time.Now()

	// Try it until it succeeds for 5 seconds, otherwise, return an error and another peer shall be found
	for {
		if time.Since(startedAt) > 5*time.Second {
			msg := fmt.Sprintf("Attempted to sync from %s for too long. Giving up.", peer.IP)
			fc.logger.Error(msg)
			return errors.New(msg)
		}

		// Maps in the block below are not synced when running in ODA mode
		if !fc.ReplicationController.IsInODAMode() {
			if err := fc.syncEndpointsData(peerUUID, peer); err != nil {
				fc.logger.Error(fmt.Sprintf("Error while performing initial sync of the endpoint data: %s", err))
				time.Sleep(5 * time.Second)
				continue
			}

			if err := fc.syncEndpointsCache(peerUUID, peer); err != nil {
				fc.logger.Error(fmt.Sprintf("Error while performing initial sync of the endpoints cache: %s", err))
				time.Sleep(5 * time.Second)
				continue
			}
		}

		// IP2MAC maps are synced whether or not we're running in ODA mode
		if err := fc.syncIpMaps(peerUUID, peer); err != nil {
			fc.logger.Error(fmt.Sprintf("Error while performing initial sync of the IP maps: %s", err))
			time.Sleep(5 * time.Second)
			continue
		}

		return nil
	}
}

func (fc *FingerbankCollector) getSyncData(peer fcreplication.ApiSyncCapablePeer, path string) (*http.Response, error) {
	if !peer.ShouldQuery() {
		msg := fmt.Sprintf("Peer %s is currently marked as dead. Ignoring it", peer)
		fc.logger.Error(msg)
		return nil, errors.New(msg)
	}

	client := peerHttpClient

	defaultProtocol := "http"
	if fc.ApiUsesHTTPS {
		defaultProtocol += "s"
	}

	req, err := http.NewRequest("GET", fmt.Sprintf("%s%s", peer.GetURL(defaultProtocol), path), bytes.NewBuffer([]byte{}))
	sharedutils.CheckError(err)
	req.Header.Set("Authorization", "Token "+fc.apiKey)

	resp, err := client.Do(req)
	if err == nil {
		peer.RecordSuccess()
	} else {
		peer.RecordError()
	}

	return resp, err
}

func (fc *FingerbankCollector) syncIpMaps(peerUUID string, peer *fcreplication.FingerbankCollectorPeer) error {
	resp, err := fc.getSyncData(peer, "/sync/ip_maps")

	if err != nil {
		return err
	}

	if resp.StatusCode != http.StatusOK {
		body, _ := ioutil.ReadAll(resp.Body)
		resp.Body.Close()
		return errors.New(fmt.Sprintf("Error code while trying to perform the initial collector sync for the IP maps. Code is %d. Body is %s", resp.StatusCode, string(body)))
	}

	defer resp.Body.Close()
	return fc.loadIpMaps(resp.Body)
}

func (fc *FingerbankCollector) syncEndpointsCache(peerUUID string, peer *fcreplication.FingerbankCollectorPeer) error {
	resp, err := fc.getSyncData(peer, "/sync/endpoints_cache")

	if err != nil {
		return err
	}

	if resp.StatusCode != http.StatusOK {
		body, _ := ioutil.ReadAll(resp.Body)
		resp.Body.Close()
		return errors.New(fmt.Sprintf("Error code while trying to perform the initial collector sync for the endpoints cache. Code is %d. Body is %s", resp.StatusCode, string(body)))
	}

	defer resp.Body.Close()
	return fc.loadEndpointsCache(resp.Body)
}

func (fc *FingerbankCollector) syncEndpointsData(peerUUID string, peer *fcreplication.FingerbankCollectorPeer) error {
	resp, err := fc.getSyncData(peer, "/sync/endpoint_data")

	if err != nil {
		return err
	}

	if resp.StatusCode != http.StatusOK {
		body, _ := ioutil.ReadAll(resp.Body)
		resp.Body.Close()
		return errors.New(fmt.Sprintf("Error code while trying to perform the initial collector sync for the endpoint data. Code is %d. Body is %s", resp.StatusCode, string(body)))
	}

	defer resp.Body.Close()
	err = fc.loadEndpointsData(resp.Body)

	if err != nil {
		return err
	}

	return nil
}

func (fc *FingerbankCollector) syncIpMacUpdate(ip, mac string) {
	fc.ReplicationController.SendContentToOutboundFriends(
		fcreplication.REP_IP_MAC_UPDATE,
		fcreplication.REPIpMacUpdate{
			Timestamp: time.Now(),
			MAC:       mac,
			IP:        ip,
		},
	)
}

func (fc *FingerbankCollector) syncAddIpHostMapping(endpointId, ip, host string) {
	fc.ReplicationController.SendContentToOutboundFriends(
		fcreplication.REP_ADD_IP_HOST_MAPPING,
		fcreplication.REPAddIpHostMapping{
			Timestamp:  time.Now(),
			EndpointId: endpointId,
			IP:         ip,
			Host:       host,
		},
	)
}

func (fc *FingerbankCollector) syncAddDestinationHost(endpointId, host string, port int, proto string) {
	fc.ReplicationController.SendContentToOutboundFriends(
		fcreplication.REP_ADD_DESTINATION_HOST,
		fcreplication.REPAddDestinationHost{
			Timestamp:  time.Now(),
			EndpointId: endpointId,
			Host:       host,
			Port:       port,
			Proto:      proto,
		},
	)
}

func (fc *FingerbankCollector) syncResetEndpointCache(endpointId string) {
	fc.ReplicationController.SendContentToOutboundFriends(
		fcreplication.REP_RESET_ENDPOINT_CACHE,
		fcreplication.REPResetEndpointCache{
			Timestamp:  time.Now(),
			EndpointId: endpointId,
		},
	)
}

func (fc *FingerbankCollector) syncEndpointDataDelete(ed *EndpointData) {
	// No need to do anything if there is no replication controller
	if fc.ReplicationController == nil {
		return
	}

	fc.ReplicationController.SendContentToOutboundFriends(
		fcreplication.REP_ENDPOINT_DATA_DELETE,
		fcreplication.REPEndpointDataReplace{
			Timestamp:  time.Now(),
			MacAddress: ed.mac,
		},
	)
}

func (fc *FingerbankCollector) EndpointDataODA(ed *EndpointData) *EndpointData {
	if fc.ReplicationController != nil && fc.ReplicationController.IsInODAMode() {
		ed = ed.Copy()
		latestUpdate := ed.LastUpdated
		for _, peer := range fc.ReplicationController.ApiSyncCapablePeers() {
			resp, err := fc.getSyncData(peer, fmt.Sprintf("/sync/endpoint_data/%s/oda_data", ed.mac))
			if err != nil {
				fc.logger.Error(fmt.Sprintf("Unable to communicate with peer %s to obtain ODA data: %s", peer, err))
				continue
			}
			func() {
				defer resp.Body.Close()
				edoda := &EndpointDataODA{EndpointData: ed}
				err := json.NewDecoder(resp.Body).Decode(edoda)
				if err != nil {
					fc.logger.Error(fmt.Sprintf("Unable to JSON decode ODA data from peer %s for MAC %s: %s", peer, ed.mac, err))
				} else {
					fc.logger.Debug(fmt.Sprintf("Aggregated data from %s for endpoint %s", peer, ed.mac))
					if ed.LastUpdated.After(latestUpdate) {
						latestUpdate = ed.LastUpdated
					}
					ed.mergeDestinationHosts(ed.ODAAllHostsCache, edoda.AllHostsCache)
				}
			}()
		}
		ed.LastUpdated = latestUpdate
		return ed
	} else {
		return ed
	}
}
