package fccommon

import (
	"encoding/json"
	"testing"
	"time"

	fcreplication "git.inverse.ca/inverse/fingerbank-collector/common/replication"
	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestCollectorCleanupInactiveEndpointData(t *testing.T) {
	collector := NewFingerbankCollector("")
	collector.FindEndpointDataMAC("001122334455")

	if len(collector.AllEndpointsData()) != 1 {
		t.Error("Failed adding endpoint")
	}

	// Try to cleanup and it shouldn't be removing anything yet
	collector.CleanupInactiveEndpointData(1 * time.Second)

	if len(collector.AllEndpointsData()) != 1 {
		t.Error("Endpoint was removed before it expired")
	}

	// Wait 1 second and we'll expire the node after that second
	time.Sleep(1 * time.Second)

	// Try to cleanup and it shouldn't be removing anything yet
	collector.CleanupInactiveEndpointData(1 * time.Second)

	if len(collector.AllEndpointsData()) != 0 {
		t.Error("Expired endpoint wasn't deleted")
	}

}

func TestCollectorHandleREPEndpointDataUpdate(t *testing.T) {
	collector := NewFingerbankCollector("")
	mac := "001122334455"
	endpoint, _ := collector.FindEndpointDataMAC(mac)

	if endpoint.DhcpFingerprint != "" {
		t.Error("Endpoint doesn't have valid initial value")
	}

	delta := DeltaEndpointData{DhcpFingerprint: "192,168,5,3"}
	deltaJSON, err := json.Marshal(delta)
	sharedutils.CheckError(err)

	request := fcreplication.REPEndpointDataUpdate{Timestamp: time.Now(), DeltaEndpointDataJSON: deltaJSON, MacAddress: mac}

	collector.HandleREPEndpointDataUpdate(request)

	if endpoint.DhcpFingerprint != "192,168,5,3" {
		t.Error("Value isn't correct after updating via HandleREPEndpointDataUpdate")
	}
}

func TestCollectorCheckDhcpVendorChange(t *testing.T) {
	collector := NewFingerbankCollector("")
	ed1, _ := collector.FindEndpointDataMAC("001122334455")
	ed1.DhcpFingerprint = "1,2,3,4,5"
	ed1.DhcpVendor = "test"
	ed1.Hostname = "not-zammits-pc"
	ed1.AddUserAgent("vidange")
	uuidStart := ed1.UUID

	ed1.Update(func(ed *EndpointData) bool {
		ed.DhcpVendor = "test2"
		return true
	})
	collector.CheckDhcpVendorChange(ed1, "test")

	if ed1.DhcpFingerprint != "1,2,3,4,5" {
		t.Error("DHCP information was overwritten after DHCP vendor change although it shouldn't have")
	}

	if ed1.DhcpVendor != "test2" {
		t.Error("Wrong DHCP vendor information")
	}

	if ed1.Hostname != "not-zammits-pc" {
		t.Error("Hostname information destroyed after DHCP vendor change")
	}

	if ed1.UUID != uuidStart {
		t.Error("UUID information destroyed after DHCP vendor change")
	}

	if _, ok := ed1.UserAgents["vidange"]; ok {
		t.Error("User agent was not destroyed after DHCP vendor change")
	}

	ed1.Update(func(ed *EndpointData) bool {
		ed.DhcpFingerprint = "0,0,0,0,0"
		ed.DhcpVendor = "test3"
		return true
	})

	if ed1.DhcpFingerprint != "0,0,0,0,0" {
		t.Error("DHCP information was overwritten after DHCP vendor change although it shouldn't have")
	}

	if ed1.DhcpVendor != "test3" {
		t.Error("Wrong DHCP vendor information")
	}
}

func TestCollectorAddDestinationHost(t *testing.T) {
	collector := NewFingerbankCollector("")
	ed1, _ := collector.FindEndpointDataMAC("001122334455")
	collector.AddDestinationHost(ed1, "otbm.com", 69, "tcp")
	updatedAt := ed1.LastUpdated
	time.Sleep(2 * time.Second)
	collector.AddDestinationHost(ed1, "otbm.com", 69, "tcp")
	if ed1.LastUpdated != updatedAt {
		t.Error("Same dhost combination made twice updated the endpoint data twice")
	}
}
