package fccommon

import (
	"encoding/hex"
	"net"
	"regexp"
)

var mozillaRegexp = regexp.MustCompile(`^Mozilla`)

// Struct that represents the basic info that should be extracted from a packet which will be sent to the handlers
type PacketInfo struct {
	SrcMAC   net.HardwareAddr
	SrcIP    string
	NetSrcIP net.IP
	SrcPort  int
	DstMAC   net.HardwareAddr
	DstIP    string
	NetDstIP net.IP
	DstPort  int
	Proto    string
}

func (pi *PacketInfo) StringSrcMAC() string {
	hexMac := make([]byte, hex.EncodedLen(len(pi.SrcMAC)))
	hex.Encode(hexMac, pi.SrcMAC)
	return string(hexMac)
}

func (pi *PacketInfo) StringDstMAC() string {
	hexMac := make([]byte, hex.EncodedLen(len(pi.DstMAC)))
	hex.Encode(hexMac, pi.DstMAC)
	return string(hexMac)
}

// Interface that represents a collector
type Collector interface {
	UpdateIPMAC(ip string, mac string) error
	FindEndpointDataMAC(mac string) (*EndpointData, error)
	FindEndpointDataIP(ip string) (*EndpointData, error)
	DeleteEndpointData(ed *EndpointData) bool
	IP2MAC(ip string) (string, error)
	MAC2IP(mac string) (string, error)
	AddIpHostMapping(endpointId string, ip string, host string)
	AddDestinationHost(ed *EndpointData, host string, port int, proto string)
	RegisterProtocolHandler(ph ProtocolHandler, bpf string) error
	CheckDhcpVendorChange(ed *EndpointData, dhcpVendorBefore string)
}

// Interface that a protocol handler must implement in order for it to receive collected packets
type ProtocolHandler interface {
	// info is the PacketInfo associated to this connection
	// l4Payload is the bytes of the TCP or UDP packet
	// rawPacket is the bytes of the packet as received on the wire
	Handle(info PacketInfo, l4Payload []byte, rawPacket []byte)
	// Returns the BPF of the packets the protocol handler is interested in
	GetBPF() string
	// Friendly string representation of this protocol handler
	String() string
	// Name that represents this protocol handler
	// Should return a value following the golang package naming standards
	Name() string
	// Method to set the collector of the handler
	SetCollector(Collector)
}

// Basic struct that should be included in all protocol handlers so that they have a collector attribute and other potential attributes this will have
type GenericHandler struct {
	Collector Collector
}

func (gh *GenericHandler) SetCollector(c Collector) {
	gh.Collector = c
}
