package fccommon

import (
	"strconv"
	"testing"
	"time"
)

func TestEndpointDataUpdate(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	now := time.Now()
	// Test update with a different variable name
	ed.Update(func(edl *EndpointData) bool {
		edl.DhcpFingerprint = "vidange"
		return true
	})

	if now.After(ed.GetLastUpdated()) {
		t.Error("Update didn't modify the LastUpdated")
	}

	if ed.DhcpFingerprint != "vidange" {
		t.Error("EndpointData.Update has failed")
	}

	now = time.Now()
	// Test update with the same variable name
	ed.Update(func(ed *EndpointData) bool {
		ed.DhcpFingerprint = "vidange2"
		return false
	})

	if now.Before(ed.GetLastUpdated()) {
		t.Error("Update modified the LastUpdated even though it shouldn't")
	}

	if ed.DhcpFingerprint != "vidange2" {
		t.Error("EndpointData.Update has failed")
	}

	// Test locking
	lockid := ed.lock.Lock()

	c := make(chan int, 1)

	// spawn an update function that will write to c when done
	go func() {
		ed.Update(func(ed *EndpointData) bool {
			ed.DhcpFingerprint = "vidange3"
			c <- 1
			return true
		})
	}()

	// We expect this to timeout, if c gets something in it, then the lock has failed
	select {
	case <-c:
		t.Error("Got something from the update channel that should have failed to lock")
	case <-time.After(1 * time.Second):
	}

	ed.lock.Unlock(lockid)

	// Should now return from the update function
	select {
	case <-c:
	case <-time.After(1 * time.Second):
		t.Error("Didn't get anything from update channel after lock was removed")
	}

}

func TestEndpointDataAddDestinationHost(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)

	// Adding a host that is allowed
	host := "otbm.com"
	port := 0

	if ed.DestinationHosts[host] != nil {
		t.Error("Unseen host counters is not set to nil")
	}

	ed.AddDestinationHost(host, port)

	if ed.DestinationHosts[host][strconv.Itoa(port)] != 1 {
		t.Error("Didn't set the host counter")
	}

	ed.AddDestinationHost(host, port)

	if ed.DestinationHosts[host][strconv.Itoa(port)] != 2 {
		t.Error("Didn't set the host counter")
	}

	// Try a different port
	port = 1234

	ed.AddDestinationHost(host, port)

	if ed.DestinationHosts[host][strconv.Itoa(port)] != 1 {
		t.Error("Didn't set the host counter")
	}

	// Trying to add a host that should be ignored
	host = "zamm.it"

	if ed.DestinationHosts[host] != nil {
		t.Error("Unseen host counters is not set to nil")
	}

	ed.AddDestinationHost(host, port)

	if ed.DestinationHosts[host] != nil {
		t.Error("Set the host counter even though domain should be ignored")
	}

	// Trying with a host that should match a regexp

	host = "centos.mirror.zammitcorp.com"

	if ed.DestinationHosts[host] != nil {
		t.Error("Unseen host counters is not set to nil")
	}

	ed.AddDestinationHost(host, port)

	if ed.DestinationHosts[host][strconv.Itoa(port)] != 1 {
		t.Error("Didn't set the host counter")
	}

	ed.AddDestinationHost(host, port)

	if ed.DestinationHosts[host][strconv.Itoa(port)] != 2 {
		t.Error("Didn't set the host counter")
	}

}

func TestUpdateMostAccurateUserAgent(t *testing.T) {
	ed := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	// When there is no recorded UA, any should do
	ua := "vidange"
	ed.UpdateMostAccurateUserAgent(ua)
	if ed.MostAccurateUserAgent != ua {
		t.Error("User-Agent wasn't updated when it should have been")
	}

	// Updating with another dummy UA shouldn't overwrite this
	ua2 := "vidange2"
	ed.UpdateMostAccurateUserAgent(ua2)
	if ed.MostAccurateUserAgent != ua {
		t.Error("User-Agent was updated when it shouldn't have been")
	}

	// Updating with a Mozilla UA should overwrite it
	ua = "Mozilla vidange"
	ed.UpdateMostAccurateUserAgent(ua)
	if ed.MostAccurateUserAgent != ua {
		t.Error("User-Agent wasn't updated when it should have been")
	}

	// Updating with another Mozilla UA shouldn't overwrite this
	ua2 = "Mozilla vidange2"
	ed.UpdateMostAccurateUserAgent(ua2)
	if ed.MostAccurateUserAgent != ua {
		t.Error("User-Agent was updated when it shouldn't have been")
	}

	// Updating with another dummy UA shouldn't overwrite this
	ua2 = "vidange2"
	ed.UpdateMostAccurateUserAgent(ua2)
	if ed.MostAccurateUserAgent != ua {
		t.Error("User-Agent was updated when it shouldn't have been")
	}
}
