package fccommon

import (
	"testing"
)

func TestUpdateIPMAC(t *testing.T) {
	dc := NewFingerbankCollector("")

	mac := "001122334455"
	ip := "1.2.3.4"

	if _, ok := dc.ip2macMap[ip]; ok {
		t.Error("There is already an entry in the ip2macMap even though we did nothing...")
	}

	if _, ok := dc.mac2ipMap[mac]; ok {
		t.Error("There is already an entry in the mac2ipMap even though we did nothing...")
	}

	dc.UpdateIPMAC(ip, mac)

	if dc.ip2macMap[ip] != mac {
		t.Error("UpdateIPMAC failed to update ip2macMap")
	}

	if dc.mac2ipMap[mac] != ip {
		t.Error("UpdateIPMAC failed to update mac2ipMap")
	}

	// update with an invalid IP
	ip = "invalid"
	mac = "001122334455"
	err := dc.UpdateIPMAC(ip, mac)
	if err == nil {
		t.Error("UpdateIPMAC didn't return an error when provided invalid data")
	}

	if _, ok := dc.ip2macMap[ip]; ok {
		t.Error("Invalid IP was put in the ip2macMap")
	}

	// update with an invalid MAC
	ip = "1.2.3.4"
	mac = "invalid"
	err = dc.UpdateIPMAC(ip, mac)
	if err == nil {
		t.Error("UpdateIPMAC didn't return an error when provided invalid data")
	}

	if _, ok := dc.mac2ipMap[mac]; ok {
		t.Error("Invalid IP was put in the mac2ipMap")
	}

}

func TestUpdateIPMACForIPv6(t *testing.T) {
	dc := NewFingerbankCollector("")

	mac := "001122334455"
	ip := CleanIPv6("3ffe:507:0:1:200:86ff:fe05:80da")

	if _, ok := dc.ipv62macMap[ip]; ok {
		t.Error("There is already an entry in the ipv62macMap even though we did nothing...")
	}

	if _, ok := dc.mac2ipv6Map[mac]; ok {
		t.Error("There is already an entry in the mac2ipv6Map even though we did nothing...")
	}

	dc.UpdateIPMAC(ip, mac)

	if dc.ipv62macMap[ip] != mac {
		t.Error("UpdateIPMAC failed to update ipv62macMap")
	}

	if dc.mac2ipv6Map[mac] != ip {
		t.Error("UpdateIPMAC failed to update mac2ipv6Map")
	}

	// update with an invalid IP
	ip = "invalid"
	mac = "001122334455"
	err := dc.UpdateIPMAC(ip, mac)
	if err == nil {
		t.Error("UpdateIPMAC didn't return an error when provided invalid data")
	}

	if _, ok := dc.ipv62macMap[ip]; ok {
		t.Error("Invalid IP was put in the ipv62macMap")
	}

	// update with an invalid MAC
	ip = "1.2.3.4"
	mac = "invalid"
	err = dc.UpdateIPMAC(ip, mac)
	if err == nil {
		t.Error("UpdateIPMAC didn't return an error when provided invalid data")
	}

	if _, ok := dc.mac2ipv6Map[mac]; ok {
		t.Error("Invalid IP was put in the mac2ipv6Map")
	}

}
func TestFindEndpointDataMAC(t *testing.T) {
	dc := NewFingerbankCollector("")

	ed, _ := dc.FindEndpointDataMAC("001122334455")

	if ed.MostAccurateUserAgent != "" || ed.DhcpFingerprint != "" || ed.DhcpVendor != "" || ed.Dhcp6Fingerprint != "" || ed.Dhcp6Enterprise != "" || ed.OUI != "001122" || len(ed.DestinationHosts) != 0 {
		t.Error("An attribute of a non existing EndpointData was set while it should have been empty")
	}

	ed.MostAccurateUserAgent = "testing"

	ed, _ = dc.FindEndpointDataMAC("001122334455")

	if ed.MostAccurateUserAgent != "testing" {
		t.Error("Couldn't find the edited user agent")
	}

	// Trying to find an invalid MAC will provide an error
	ed, err := dc.FindEndpointDataMAC("invalid")

	if ed != nil || err == nil {
		t.Error("Invalid return values when passing invalid data to FindEndpointDataMAC")
	}
}

func TestFindEndpointDataIP(t *testing.T) {
	dc := NewFingerbankCollector("")

	ed, err := dc.FindEndpointDataIP("1.2.3.4")

	if err == nil {
		t.Error("Didn't get an error when searching for an EndpointData with an IP that has no IP/MAC binding")
	}

	dc.UpdateIPMAC("1.2.3.4", "001122334455")

	ed, err = dc.FindEndpointDataIP("1.2.3.4")

	if err != nil {
		t.Errorf("Got an error while searching for an EndpointData: %s", err)
	}

	if ed.MostAccurateUserAgent != "" || ed.DhcpFingerprint != "" || ed.DhcpVendor != "" || ed.Dhcp6Fingerprint != "" || ed.Dhcp6Enterprise != "" || ed.OUI != "001122" || len(ed.DestinationHosts) != 0 {
		t.Error("An attribute of a non existing EndpointData was set while it should have been empty")
	}

	ed.MostAccurateUserAgent = "testing"

	ed, _ = dc.FindEndpointDataIP("1.2.3.4")

	if ed.MostAccurateUserAgent != "testing" {
		t.Error("Couldn't find the edited user agent")
	}

	// Trying to find an invalid IP will provide an error
	ed, err = dc.FindEndpointDataIP("invalid")

	if ed != nil || err == nil {
		t.Error("Invalid return values when passing invalid data to FindEndpointDataIP")
	}
}

func TestIP2MAC(t *testing.T) {
	dc := NewFingerbankCollector("")

	mac, err := dc.IP2MAC("1.2.3.4")

	if err == nil {
		t.Error("Didn't get an error while searching for an inexisting IP")
	}

	dc.UpdateIPMAC("1.2.3.4", "001122334455")

	mac, err = dc.IP2MAC("1.2.3.4")

	if err != nil {
		t.Errorf("Got an error while searching for a valid IP: %s", err)
	}

	if mac != "001122334455" {
		t.Errorf("Didn't find the right MAC for the IP. Got %s", mac)
	}

	mac, err = dc.IP2MAC("invalid")
	if mac != "" || err == nil {
		t.Error("Got invalid return values while searching IP2MAC with invalid data")
	}
}

func TestMAC2IP(t *testing.T) {
	dc := NewFingerbankCollector("")

	ip, err := dc.MAC2IP("001122334455")

	if err == nil {
		t.Error("Didn't get an error while searching for an inexisting MAC")
	}

	dc.UpdateIPMAC("1.2.3.4", "001122334455")

	ip, err = dc.MAC2IP("001122334455")

	if err != nil {
		t.Errorf("Got an error while searching for a valid MAC: %s", err)
	}

	if ip != "1.2.3.4" {
		t.Errorf("Didn't find the right IP for the MAC. Got %s", ip)
	}

	ip, err = dc.MAC2IP("invalid")
	if ip != "" || err == nil {
		t.Error("Got invalid return values while searching MAC2IP with invalid data")
	}
}
