package fccommon

import (
	"encoding/json"
	"flag"
	"fmt"
	"strconv"
	"strings"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/boot_sequence"
	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
	"git.inverse.ca/inverse/fingerbank-collector/gobpool"
	"git.inverse.ca/inverse/fingerbank-collector/timedlock"

	"github.com/google/uuid"
	"github.com/inverse-inc/go-utils/sharedutils"

	_ "github.com/mailru/easyjson/gen"
)

var GobPool = gobpool.NewGobPool(5, 100)

var maxTrackJA3Domains = flag.Int(`max-track-ja3-domains`, sharedutils.EnvOrDefaultInt("COLLECTOR_MAX_TRACK_JA3_DOMAINS", 10), `The maximum amount of domains to track for the JA3 data (COLLECTOR_MAX_TRACK_JA3_DOMAINS environment variable)`)

var maxTrackDPISignatures = flag.Int(`max-track-dpi-signatures`, sharedutils.EnvOrDefaultInt("COLLECTOR_MAX_TRACK_DPI_SIGNATURES", 100), `The maximum amount of DPI signatures to track per endpoint (COLLECTOR_MAX_TRACK_DPI_SIGNATURES environment variable)`)

var maxInterrogateBootSequence = flag.Int(`max-interrogate-boot-sequence`, sharedutils.EnvOrDefaultInt("COLLECTOR_MAX_INTERROGATE_BOOT_SEQUENCE", 75), `The maximum amount of events to send for a boot sequence while interrogating the API (COLLECTOR_MAX_INTERROGATE_BOOT_SEQUENCE environment variable)`)

const BlacklistedCommunicationsPortsID = "__blacklisted_ports__"

// Special struct for trackable time in the endpoint data (will be considered in the delta)
type TrackableTime struct {
	time.Time
}

// Holds the endpoint data that is collected from the handlers
type EndpointData struct {
	lock                  *timedlock.RWLock
	updateHook            func(edBefore, edAfter *EndpointData, bsEvents []string)
	analysisHook          func(ed *EndpointData)
	runsInLiveReplication func() bool
	CreatedAt             time.Time `json:"created_at"`
	LastUpdated           time.Time `json:"last_updated"`
	UUID                  string
	MostAccurateUserAgent string         `json:"most_accurate_user_agent"`
	UserAgents            map[string]int `json:"user_agents"`
	DhcpFingerprint       string         `json:"dhcp_fingerprint"`
	DhcpVendor            string         `json:"dhcp_vendor"`
	Dhcp6Fingerprint      string         `json:"dhcp6_fingerprint"`
	Dhcp6Enterprise       string         `json:"dhcp6_enterprise"`
	Hostname              string         `json:"hostname"`
	OUI                   string         `json:"oui"`
	// Map of all the seen hosts with the ports it was seen on
	// Also counts the amount of time it was seen on the port
	// {
	//	zammitcorp.com: {
	//   80: 1,
	//   443: 2,
	//  }
	// }
	DestinationHosts map[string]map[string]int `json:"destination_hosts"`
	UPnP             string                    `json:"upnp"`
	// Map UPnP information if it is a service or device, urn and details
	// about this UPnP request
	UPnPURNs          map[string]int `json:"upnp_urn"`
	UPnPServices      map[string]int `json:"upnp_services"`
	UPnPDevices       map[string]int `json:"upnp_devices"`
	UPnPServerStrings map[string]int `json:"upnp_server_strings"`
	UPnPUserAgents    map[string]int `json:"upnp_user_agents"`
	MDNSRequests      map[string]int `json:"mdns_requests"`
	MDNSServices      map[string]int `json:"mdns_services"`
	// TCP signatures
	TCPSynSignatures    map[string]int `json:"tcp_syn_signatures"`
	TCPSynAckSignatures map[string]int `json:"tcp_syn_ack_signatures"`
	//RADIUS
	LLDPCapabilities map[string]bool `json:"lldp_capabilities"`
	CDPCapabilities  map[string]bool `json:"cdp_capabilities"`
	//JA3 fingerprints
	JA3Fingerprints map[string]int `json:"ja3_fingerprints"`

	// Blacklisted IPs communication
	BlacklistedCommunications map[string]map[int]TrackableTime `json:"blacklisted_communications"`

	// HTTP client hints
	ClientHints map[string]map[string]string `json:"client_hints"`

	// JA3 expanded data
	JA3Data map[string]map[string]int `json:"ja3_data"`

	// DPI signatures
	DPISignatures map[string]int `json:"dpi_signatures"`

	// Track action during boot sequence
	BootSequence boot_sequence.BootSequence `json:"boot_sequence"`

	// From and To ports
	FromPorts map[string]int `json:"from_ports"`
	ToPorts   map[string]int `json:"to_ports"`

	// Private, for tracking MAC address directly in the endpoint. Set by the collector while creating the endpoint, should never be touched after that.
	mac string

	// For the all_hosts_cache destination hosts obtained via ODA sync
	ODAAllHostsCache map[string]map[string]int `json:"-"`
}

type EndpointDataODA struct {
	EndpointData  *EndpointData             `json:"endpoint_data"`
	AllHostsCache map[string]map[string]int `json:"all_hosts_cache"`
}

func NewEndpointData(updateHook func(edBefore, edAfter *EndpointData, bsEvents []string), analysisHook func(ed *EndpointData), runsInLiveReplication func() bool) *EndpointData {
	ed := &EndpointData{}

	ed.UUID = uuid.New().String()
	ed.CreatedAt = time.Now()
	ed.LastUpdated = time.Now()

	ed.init(updateHook, analysisHook, runsInLiveReplication)

	return ed
}

func (ed *EndpointData) init(updateHook func(edBefore, edAfter *EndpointData, bsEvents []string), analysisHook func(ed *EndpointData), runsInLiveReplication func() bool) {
	ed.lock = timedlock.NewRWLock()
	ed.lock.Panic = true
	ed.lock.PrintErrors = true

	ed.BootSequence.Init()

	if ed.UserAgents == nil {
		ed.UserAgents = make(map[string]int)
	}
	if ed.DestinationHosts == nil {
		ed.DestinationHosts = make(map[string]map[string]int)
	}
	if ed.UPnPURNs == nil {
		ed.UPnPURNs = make(map[string]int)
	}
	if ed.UPnPServices == nil {
		ed.UPnPServices = make(map[string]int)
	}
	if ed.UPnPDevices == nil {
		ed.UPnPDevices = make(map[string]int)
	}
	if ed.UPnPServerStrings == nil {
		ed.UPnPServerStrings = make(map[string]int)
	}
	if ed.UPnPUserAgents == nil {
		ed.UPnPUserAgents = make(map[string]int)
	}
	if ed.MDNSRequests == nil {
		ed.MDNSRequests = make(map[string]int)
	}
	if ed.MDNSServices == nil {
		ed.MDNSServices = make(map[string]int)
	}
	if ed.TCPSynSignatures == nil {
		ed.TCPSynSignatures = make(map[string]int)
	}
	if ed.TCPSynAckSignatures == nil {
		ed.TCPSynAckSignatures = make(map[string]int)
	}
	if ed.LLDPCapabilities == nil {
		ed.LLDPCapabilities = make(map[string]bool)
	}
	if ed.CDPCapabilities == nil {
		ed.CDPCapabilities = make(map[string]bool)
	}
	if ed.JA3Fingerprints == nil {
		ed.JA3Fingerprints = make(map[string]int)
	}
	if ed.FromPorts == nil {
		ed.FromPorts = make(map[string]int)
	}
	if ed.ToPorts == nil {
		ed.ToPorts = make(map[string]int)
	}
	if ed.BlacklistedCommunications == nil {
		ed.BlacklistedCommunications = make(map[string]map[int]TrackableTime)
	}
	if ed.ClientHints == nil {
		ed.ClientHints = make(map[string]map[string]string)
	}
	if ed.JA3Data == nil {
		ed.JA3Data = make(map[string]map[string]int)
	}
	if ed.DPISignatures == nil {
		ed.DPISignatures = make(map[string]int)
	}
	if ed.ODAAllHostsCache == nil {
		ed.ODAAllHostsCache = make(map[string]map[string]int)
	}

	ed.updateHook = updateHook
	ed.analysisHook = analysisHook
	ed.runsInLiveReplication = runsInLiveReplication
}

func DummyEndpointDataUpdateHook(edBefore, edAfter *EndpointData, bsEvents []string) {

}

func DummyEndpointDataAnalysisHook(ed *EndpointData) {
}

func DummyEndpointDataRunsInLiveReplicationHook() bool {
	return false
}

func (ed *EndpointData) GetMac() string {
	return ed.mac
}

// Get the time at which the endpoint was created
func (ed *EndpointData) GetCreatedAt() time.Time {
	return ed.CreatedAt
}

// Return the duration for which the device was known (now - created date)
func (ed *EndpointData) TimeOnNetwork() time.Duration {
	return time.Since(ed.CreatedAt)
}

// Returns the datetime at which the EndpointData was last modified
func (ed *EndpointData) GetLastUpdated() time.Time {
	return ed.LastUpdated
}

// Acquire a read-lock on the endpoint data
func (ed *EndpointData) RLock() uint64 {
	return ed.lock.RLock()
}

// Unnlock the read-lock on the endpoint data
func (ed *EndpointData) RUnlock(id uint64) {
	ed.lock.RUnlock(id)
}

// Copies the EndpointData by encoding it through gob and decoding it in a new struct
// Takes about a tenth of a millisecond, so its not too expensive but shouldn't be abused
func (ed *EndpointData) Copy() *EndpointData {
	lockid := ed.lock.Lock()
	defer ed.lock.Unlock(lockid)

	id, zegob, err := GobPool.Get()
	sharedutils.CheckError(err)

	err = zegob.Encoder.Encode(ed)
	sharedutils.CheckError(err)

	ned := &EndpointData{}
	err = zegob.Decoder.Decode(ned)
	sharedutils.CheckError(err)

	GobPool.ReleaseItem(id)

	ned.init(ed.updateHook, ed.analysisHook, ed.runsInLiveReplication)
	ned.mac = ed.mac
	ned.BootSequence.ResetLock()

	return ned
}

// Called to perform an update on an EndpointData making sure no-one updates it in the meanwhile
// Helps prevent race conditions when changing data on the struct (especially when appending to DestinationHosts)
// TODO: Add ability to timeout getting the lock and timeout on updating
func (ed *EndpointData) Update(todo func(*EndpointData) bool) {
	cp := ed.Copy()

	lockid := ed.lock.Lock()
	defer ed.lock.Unlock(lockid)

	updated := todo(ed)

	if updated {
		ed.markAsUpdated()
	}
	if updated || ed.BootSequence.IsInBootSeq() {
		bsEvents := ed.postUpdateBootSequence(cp, ed)
		if ed.runsInLiveReplication() {
			ed.updateHook(cp, ed, bsEvents)
		}
	}
}

func (ed *EndpointData) UpdateLight(hasLearnedFunc func(*EndpointData) bool, todo func(*EndpointData)) {
	var hasLearned bool
	func() {
		lockid := ed.lock.Lock()
		defer ed.lock.Unlock(lockid)
		hasLearned = hasLearnedFunc(ed)
	}()

	var cp *EndpointData
	if (hasLearned && ed.runsInLiveReplication()) || ed.BootSequence.IsInBootSeq() {
		cp = ed.Copy()
	}

	lockid := ed.lock.Lock()
	defer ed.lock.Unlock(lockid)

	todo(ed)

	if hasLearned {
		ed.markAsUpdated()
	}
	if cp != nil {
		bsEvents := ed.postUpdateBootSequence(cp, ed)
		if ed.runsInLiveReplication() {
			ed.updateHook(cp, ed, bsEvents)
		}
	}

}

func (ed *EndpointData) postUpdateBootSequence(edBefore, edAfter *EndpointData) []string {
	bsEvents, events := EndpointDataDiffEvents(time.Now(), edBefore, edAfter)
	if edAfter.BootSequence.Add(events...) {
		bsEvents = append(bsEvents, events...)
	}
	return bsEvents
}

func (ed *EndpointData) markAsUpdated() {
	ed.LastUpdated = time.Now()
	ed.analysisHook(ed)
}

func (ed *EndpointData) UpdateFromJSON(jsonBytes []byte) {
	ed.Update(func(ed *EndpointData) bool {
		json.Unmarshal(jsonBytes, ed)
		return true
	})
}

// Add a user agent to the endpoint
func (ed *EndpointData) AddUserAgent(ua string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.UserAgents[ua]
		ed.UserAgents[ua] += 1
		return !exists
	})
	ed.UpdateMostAccurateUserAgent(ua)
}

// Updates the most accurate user agent of the endpoint if needed
func (ed *EndpointData) UpdateMostAccurateUserAgent(userAgent string) {
	ed.Update(func(ed *EndpointData) bool {
		// Anything is better than nothing....
		if ed.MostAccurateUserAgent == "" {
			ed.MostAccurateUserAgent = userAgent
			return true
		}

		// Mozilla user agents have precendence on all other user agents so if the currently recorded one is this, we exit out
		if mozillaRegexp.MatchString(ed.MostAccurateUserAgent) {
			return false
		}

		// If the provided User-Agent is a Mozilla User-Agent, we record it
		if mozillaRegexp.MatchString(userAgent) {
			ed.MostAccurateUserAgent = userAgent
			return true
		}

		return false
	})
}

// Add a destination host to the list of hosts for the EndpointData
// Will ensure it matches a valid host or valid host regexp
// A port can be specified, if no port is associated to the connection (like in DNS), 0 should be provided
func (ed *EndpointData) AddDestinationHost(host string, portInt int) {
	port := fmt.Sprintf("%d", portInt)
	if DefaultHostValidator.Validate(host) {
		ed.UpdateLight(func(ed *EndpointData) bool {
			if ed.DestinationHosts[host] == nil {
				return true
			}
			_, exists := ed.DestinationHosts[host][port]
			return !exists
		}, func(ed *EndpointData) {
			if ed.DestinationHosts[host] == nil {
				ed.DestinationHosts[host] = make(map[string]int)
			}
			ed.DestinationHosts[host][port] += 1
		})
	}
}

func (ed *EndpointData) AddBootSequence(seq ...string) {
	ed.UpdateLight(func(ed *EndpointData) bool { return false }, func(ed *EndpointData) {
		ed.BootSequence.Add(seq...)
	})
}

// Add URN about a UPnP request
func (ed *EndpointData) AddUPnPURN(urn string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.UPnPURNs[urn]
		ed.UPnPURNs[urn] += 1
		return !exists
	})
}

func (ed *EndpointData) AddLLDPCapability(capability string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.LLDPCapabilities[capability]
		ed.LLDPCapabilities[capability] = true
		return !exists
	})
}

func (ed *EndpointData) AddCDPCapability(capability string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.CDPCapabilities[capability]
		ed.CDPCapabilities[capability] = true
		return !exists
	})
}

func (ed *EndpointData) AddJA3Fingerprint(fingerprint string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.JA3Fingerprints[fingerprint]
		ed.JA3Fingerprints[fingerprint] += 1
		return !exists
	})
}

// Add details about a UPnP request type service
func (ed *EndpointData) AddUPnPService(service string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.UPnPServices[service]
		ed.UPnPServices[service] += 1
		return !exists
	})
}

// Add details about a UPnP request type device
func (ed *EndpointData) AddUPnPDevice(device string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.UPnPDevices[device]
		ed.UPnPDevices[device] += 1
		return !exists
	})
}

// Add details about a UPnP User-Agent
func (ed *EndpointData) AddUPnPUserAgent(ua string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.UPnPUserAgents[ua]
		ed.UPnPUserAgents[ua] += 1
		return !exists
	})
}

// Add details about a UPnP Server String
func (ed *EndpointData) AddUPnPServerString(ss string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.UPnPServerStrings[ss]
		ed.UPnPServerStrings[ss] += 1
		return !exists
	})
}

// Add an MDNS service request to the list of MDNSrequests for the EndpointData
func (ed *EndpointData) AddMDNSRequest(service string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.MDNSRequests[service]
		ed.MDNSRequests[service] += 1
		return !exists
	})
}

// Add a list of  MDNS service to the list of MDNServices or the EndpointData
func (ed *EndpointData) AddMDNSServices(services []string) {
	for _, s := range services {
		ed.Update(func(ed *EndpointData) bool {
			_, exists := ed.MDNSServices[s]
			ed.MDNSServices[s] += 1
			return !exists
		})
	}
}

// Add TCP SYN Signature for endpoint
func (ed *EndpointData) AddTCPSynSignature(sig string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.TCPSynSignatures[sig]
		ed.TCPSynSignatures[sig] += 1
		return !exists
	})
}

// Add TCP SYN+ACK Signature for endpoint
func (ed *EndpointData) AddTCPSynAckSignature(sig string) {
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.TCPSynAckSignatures[sig]
		ed.TCPSynAckSignatures[sig] += 1
		return !exists
	})
}

func (ed *EndpointData) purgeMapStringInt(data map[string]int, max int) {
	pl := RankByWordCount(data)
	for i := max; i < len(pl); i++ {
		delete(data, pl[i].Key)
	}
}

func (ed *EndpointData) AddDPISignatures(sigs []string) {
	ed.Update(func(ed *EndpointData) bool {
		changed := false
		for _, sig := range sigs {
			_, exists := ed.DPISignatures[sig]
			if !exists && len(ed.DPISignatures) >= *maxTrackDPISignatures {
				ed.purgeMapStringInt(ed.DPISignatures, *maxTrackDPISignatures-1)
			}

			ed.DPISignatures[sig] += 1
			if !exists {
				changed = true
			}
		}
		return changed
	})
}

func (ed *EndpointData) AddBlacklistedCommunication(dstIP string, port int) {
	ed.Update(func(ed *EndpointData) bool {
		if ed.BlacklistedCommunications[dstIP] == nil {
			ed.BlacklistedCommunications[dstIP] = make(map[int]TrackableTime)
		}
		ed.BlacklistedCommunications[dstIP][port] = TrackableTime{Time: time.Now()}
		return true
	})
}

func (ed *EndpointData) AddJA3SigCouple(domain string, port int, ja3ClientSig string, ja3ServerSig string) {
	ja3Sig := fmt.Sprintf("%s;%s", ja3ClientSig, ja3ServerSig)
	domainPort := fmt.Sprintf("%s:%d", domain, port)
	ed.Update(func(ed *EndpointData) bool {
		_, exists := ed.JA3Data[domainPort][ja3Sig]
		if !exists && len(ed.JA3Data) >= *maxTrackJA3Domains {
			ed.purgeJA3Sigs(ed.JA3Data, *maxTrackJA3Domains-1)
		}

		if ed.JA3Data[domainPort] == nil {
			ed.JA3Data[domainPort] = map[string]int{}
		}

		ed.JA3Data[domainPort][ja3Sig] += 1
		return !exists
	})
}

func (ed *EndpointData) purgeMapIntInt(data map[int]int, max int) {
	pl := RankByWordCountMapIntInt(data)
	for i := max; i < len(pl); i++ {
		delete(data, pl[i].Key)
	}
}

func (ed *EndpointData) purgeJA3Sigs(data map[string]map[string]int, max int) {
	pl := RankByWordCountMapStringMapStringInt(data)
	for i := max; i < len(pl); i++ {
		delete(data, pl[i].Key)
	}
}

func (ed *EndpointData) HasBehavioralData() bool {
	return len(ed.DestinationHosts) > 0 ||
		len(ed.MDNSServices) > 0 ||
		len(ed.UPnPUserAgents) > 0 ||
		len(ed.UPnPServerStrings) > 0 ||
		len(ed.TCPSynSignatures) > 0 ||
		len(ed.TCPSynAckSignatures) > 0
}

func (ed *EndpointData) computeDestinationHosts(fc *FingerbankCollector, dhosts map[string]map[string]int) map[string][]string {
	newDhosts := map[string][]string{}
	for k, v := range dhosts {
		portsCount := 0
		for port, _ := range v {
			//Limit to 10 destination ports
			portsCount += 1
			if portsCount > 10 {
				break
			}
			if _, found := newDhosts[k]; !found {
				newDhosts[k] = []string{port}
			} else {
				newDhosts[k] = append(newDhosts[k], port)
			}
		}
	}
	return newDhosts
}

func (ed *EndpointData) AllHostsCacheMap(fc *FingerbankCollector) map[string]map[string]int {
	merged := map[string]map[string]int{}
	ed.mergeDestinationHosts(merged, ed.ODAAllHostsCache)

	if ahc, found := fc.EndpointsCache.GetEndpointAllHostsCache(ed.mac); found {
		lockid := ahc.RLock()
		defer ahc.RUnlock(lockid)
		ahcMap := ahc.ToMap()
		ed.mergeDestinationHosts(merged, ahcMap)
	} else {
		ed.mergeDestinationHosts(merged, ed.DestinationHosts)
	}
	return merged
}

func (ed *EndpointData) AllDestinationHosts(fc *FingerbankCollector) map[string][]string {
	merged := ed.AllHostsCacheMap(fc)
	return ed.computeDestinationHosts(fc, merged)
}

func (ed *EndpointData) AllDestinationHostsLegacy(fc *FingerbankCollector) map[string][]int {
	dhosts := ed.AllDestinationHosts(fc)
	m2 := map[string][]int{}
	for k, ports := range dhosts {
		m2[k] = []int{}
		for _, protoPortStr := range ports {
			portStrParts := strings.Split(protoPortStr, ":")
			portStr := portStrParts[len(portStrParts)-1]
			portInt, err := strconv.Atoi(portStr)
			sharedutils.CheckError(err)
			m2[k] = append(m2[k], portInt)
		}
	}
	return m2
}

func (ed *EndpointData) RecordPorts(proto string, from, to int) {
	fromStr := fmt.Sprintf("%s:%d", proto, from)
	toStr := fmt.Sprintf("%s:%d", proto, to)
	ed.UpdateLight(func(ed *EndpointData) bool {
		_, existsFrom := ed.FromPorts[fromStr]
		_, existsTo := ed.ToPorts[toStr]
		return !existsFrom || !existsTo
	}, func(ed *EndpointData) {
		if from < *ephemeralPortsStart {
			_, existsFrom := ed.FromPorts[fromStr]
			if !existsFrom && len(ed.FromPorts) >= *maxTrackSrcDstPortsInEndpoint {
				ed.purgeMapStringInt(ed.FromPorts, *maxTrackSrcDstPortsInEndpoint-1)
			}

			ed.FromPorts[fromStr] += 1
		}
		if to < *ephemeralPortsStart {
			_, existsTo := ed.ToPorts[toStr]
			if !existsTo && len(ed.ToPorts) >= *maxTrackSrcDstPortsInEndpoint {
				ed.purgeMapStringInt(ed.ToPorts, *maxTrackSrcDstPortsInEndpoint-1)
			}

			ed.ToPorts[toStr] += 1
		}
	})
}

func (ed *EndpointData) ToInterrogateQuery(fc *FingerbankCollector) fingerbankclient.InterrogateQuery {
	lockid := ed.lock.RLock()
	defer ed.lock.RUnlock(lockid)

	var bestStr PairList

	// The Fingerbank API accepts an array of hosts and not a map like we're collecting here
	var destinationHosts []string
	for host, _ := range ed.DestinationHosts {
		// In case we previously collected hosts that aren't required anymore
		if DefaultHostValidator.Validate(host) {
			destinationHosts = append(destinationHosts, host)
		}
	}

	allDestinationHosts := ed.AllDestinationHostsLegacy(fc)

	bestStr = RankByWordCount(ed.UserAgents)
	var userAgents []string
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		userAgents = append(userAgents, pair.Key)
	}

	bestStr = RankByWordCount(ed.MDNSServices)
	var mdnsServices []string
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		mdnsServices = append(mdnsServices, pair.Key)
	}

	var upnpUserAgents []string
	bestStr = RankByWordCount(ed.UPnPUserAgents)
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		upnpUserAgents = append(upnpUserAgents, pair.Key)
	}

	var upnpServerStrings []string
	bestStr = RankByWordCount(ed.UPnPServerStrings)
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		upnpServerStrings = append(upnpServerStrings, pair.Key)
	}

	var tcpSynSignatures []string
	bestStr = RankByWordCount(ed.TCPSynSignatures)
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		tcpSynSignatures = append(tcpSynSignatures, pair.Key)
	}

	var tcpSynAckSignatures []string
	bestStr = RankByWordCount(ed.TCPSynAckSignatures)
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		tcpSynAckSignatures = append(tcpSynAckSignatures, pair.Key)
	}

	var ja3Fingerprints []string
	bestStr = RankByWordCount(ed.JA3Fingerprints)
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= 5 {
			break
		}
		ja3Fingerprints = append(ja3Fingerprints, pair.Key)
	}

	var ports []string
	bestStr = RankByWordCount(ed.FromPorts)
	for i, pair := range bestStr {
		if i >= *maxTrackSrcDstPortsInEndpoint {
			break
		}
		ports = append(ports, fmt.Sprintf("%s:", pair.Key))
	}
	bestStr = RankByWordCount(ed.ToPorts)
	for i, pair := range bestStr {
		if i >= *maxTrackSrcDstPortsInEndpoint {
			break
		}
		ports = append(ports, fmt.Sprintf(":%s", pair.Key))
	}

	ja3Data := map[string][]string{}
	ed.buildJA3DataSigs(ja3Data)

	var dpiSignatures []string
	bestStr = RankByWordCount(ed.DPISignatures)
	for i, pair := range bestStr {
		// Limit it to 5 items
		if i >= *maxTrackDPISignatures {
			break
		}
		dpiSignatures = append(dpiSignatures, pair.Key)
	}

	return fingerbankclient.InterrogateQuery{
		BootSequence:        ed.BootSequence.GetFirstN(*maxInterrogateBootSequence),
		ClientHints:         ed.ClientHints,
		DestinationHosts:    destinationHosts,
		Destinations:        allDestinationHosts,
		Dhcp6Enterprise:     ed.Dhcp6Enterprise,
		Dhcp6Fingerprint:    ed.Dhcp6Fingerprint,
		DhcpFingerprint:     ed.DhcpFingerprint,
		DhcpVendor:          ed.DhcpVendor,
		DPISignatures:       dpiSignatures,
		Hostname:            ed.Hostname,
		JA3Data:             ja3Data,
		JA3Fingerprints:     ja3Fingerprints,
		MDNSServices:        mdnsServices,
		MacVendor:           ed.OUI,
		Ports:               ports,
		TCPSynAckSignatures: tcpSynAckSignatures,
		TCPSynSignatures:    tcpSynSignatures,
		UPnPServerStrings:   upnpServerStrings,
		UPnPUserAgents:      upnpUserAgents,
		UserAgents:          userAgents,
	}
}

func (ed *EndpointData) buildJA3DataSigs(into map[string][]string) {
	src := ed.JA3Data
	bestDomains := RankByWordCountMapStringMapStringInt(src)
	for i, pair := range bestDomains {
		if i >= *maxTrackJA3Domains {
			break
		}
		domainPort := pair.Key
		sigs := src[pair.Key]
		if into[domainPort] == nil {
			into[domainPort] = []string{}
		}
		bestSigs := RankByWordCount(sigs)
		for i, pair := range bestSigs {
			if i >= 3 {
				break
			}
			if into[domainPort] == nil {
				into[domainPort] = []string{}
			}
			into[domainPort] = append(into[domainPort], pair.Key)
		}
	}
}

func (ed *EndpointData) mergeDestinationHosts(into map[string]map[string]int, toStore map[string]map[string]int) {
	for domainToStore, portsToStore := range toStore {
		if _, found := into[domainToStore]; !found {
			into[domainToStore] = portsToStore
		} else {
			for portToStore, counter := range portsToStore {
				into[domainToStore][portToStore] += counter
			}
		}
	}
}

type ExpandedEndpointData struct {
	EndpointData
	IP               string                    `json:"ip"`
	IPv6             string                    `json:"ipv6"`
	DiscoveredAt     time.Time                 `json:"discovered_at"`
	SecondsOnNetwork float64                   `json:"seconds_on_network"`
	MAC              string                    `json:"mac"`
	IpHostCache      map[string]string         `json:"ip_host_cache"`
	AllHostsCache    map[string]map[string]int `json:"all_hosts_cache"`
}

//easyjson:json
type MapEndpointsData map[string]*EndpointData
