package fccommon

import (
	"reflect"
	"strings"
	"testing"
	"time"

	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestEndpointDataDiffEvents(t *testing.T) {
	{
		ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed1.DhcpFingerprint = "192,168,5,3"
		ed2.DhcpFingerprint = "192,168,5,3"
		ts := time.Now()
		_, events := EndpointDataDiffEvents(ts, ed1, ed2)
		if len(events) != 0 {
			t.Error("Difference between")
		}
	}

	{
		ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2.DhcpFingerprint = "192,168,5,3"
		ts := time.Now()
		_, events := EndpointDataDiffEvents(ts, ed1, ed2)
		if len(events) != 1 {
			t.Error("Difference between")
		}
	}

	{
		ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2.UserAgents = map[string]int{
			"a": 1,
			"b": 2,
		}
		ts := time.Now()
		_, events := EndpointDataDiffEvents(ts, ed1, ed2)
		if len(events) != 2 {
			t.Fatal("Difference between")
		}

		if strings.HasSuffix(events[0], "user_agents:b") {
			if !strings.HasSuffix(events[1], "user_agents:a") {
				t.Fatal("Differences captured")
			}
		} else if strings.HasSuffix(events[0], "user_agents:a") {
			if !strings.HasSuffix(events[1], "user_agents:b") {
				t.Fatal("Differences captured")
			}
		} else {
			t.Fatal("Differences not captured")
		}

	}

	{
		ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed1.UserAgents = map[string]int{
			"a": 1,
			"b": 1,
		}
		ed2.UserAgents = map[string]int{
			"a": 1,
			"b": 2,
		}
		ts := time.Now()
		_, events := EndpointDataDiffEvents(ts, ed1, ed2)
		if len(events) != 1 {
			t.Error("Difference between")
		}

		if !strings.HasSuffix(events[0], "user_agents:b") {
			t.Fatal("Difference of events[0] not captured")
		}
	}

	{
		ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
		ed1.DestinationHosts = map[string]map[string]int{
			"1": map[string]int{
				"80":  1,
				"443": 1,
			},
			"2": map[string]int{
				"80":  1,
				"443": 1,
			},
		}

		ed2.DestinationHosts = map[string]map[string]int{
			"1": map[string]int{
				"80":  1,
				"443": 1,
			},
			"2": map[string]int{
				"80":  1,
				"443": 1,
				"192": 1,
			},
		}

		ts := time.Now()
		_, events := EndpointDataDiffEvents(ts, ed1, ed2)
		if len(events) != 1 {
			t.Error("Difference between")
		}
	}

}

func TestEndpointDataComparatorBuildDeltaEndpoint(t *testing.T) {
	edc := EndpointDataComparator{}

	ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2.DhcpFingerprint = "192,168,5,3"

	changed, delta, err := edc.BuildDeltaEndpoint(ed1, ed2)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("returned no difference found although there is a delta")
	}

	if delta.DhcpFingerprint != "192,168,5,3" {
		t.Error("Delta doesn't contain field that was different")
	}

	ed1 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed1.DhcpFingerprint = "192,168,5,3"
	ed2 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2.DhcpFingerprint = "192,168,5,3"

	changed, delta, err = edc.BuildDeltaEndpoint(ed1, ed2)
	sharedutils.CheckTestError(t, err)

	if changed {
		t.Error("returned a difference found although there isn't a delta")
	}

	if delta.DhcpFingerprint != "" {
		t.Error("Delta value isn't empty for a field that wasn't changed")
	}

	ed1 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2.UserAgents = map[string]int{
		"1": 1,
		"2": 2,
	}

	changed, delta, err = edc.BuildDeltaEndpoint(ed1, ed2)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("returned no difference found although there is a delta")
	}

	if len(delta.UserAgents) != 2 {
		t.Error("Delta value is not correct for a field that changed")
	}

	ed1 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed1.UserAgents = map[string]int{
		"1": 1,
	}
	ed2 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2.UserAgents = map[string]int{
		"1": 1,
		"2": 2,
	}

	changed, delta, err = edc.BuildDeltaEndpoint(ed1, ed2)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("returned no difference found although there is a delta")
	}

	if len(delta.UserAgents) != 1 {
		t.Error("Delta value is not correct for a field that changed")
	}

	ed1 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed1.UserAgents = map[string]int{
		"1": 1,
		"2": 2,
	}
	ed2 = NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2.UserAgents = map[string]int{
		"1": 1,
		"2": 2,
	}

	changed, delta, err = edc.BuildDeltaEndpoint(ed1, ed2)
	sharedutils.CheckTestError(t, err)

	if changed {
		t.Error("returned a difference found although there isn't a delta")
	}

	if len(delta.UserAgents) != 0 {
		t.Error("Delta value is not correct for a field hasn't changed")
	}

}

func TestEndpointDataComparatorDeltaValueField(t *testing.T) {
	edc := EndpointDataComparator{}

	delta := &DeltaEndpointData{}
	rdelta := reflect.ValueOf(delta).Elem()
	f := rdelta.FieldByName("DhcpFingerprint")

	changed, err := edc.deltaValueField(reflect.ValueOf("1"), reflect.ValueOf("2"), f)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("deltaMapField returned no difference found although there is a delta")
	}

	if delta.DhcpFingerprint != "2" {
		t.Error("Changed DhcpFingerprint wasn't set properly in delta")
	}

	edc.deltaValueField(reflect.ValueOf("1"), reflect.ValueOf("1"), f)

	if delta.DhcpFingerprint != "2" {
		t.Error("DhcpFingerprint was changed although the src + dst fields had the same value")
	}
}

func TestEndpointDataComparatorDeltaMapField(t *testing.T) {
	edc := EndpointDataComparator{}

	delta := &DeltaEndpointData{}
	rdelta := reflect.ValueOf(delta).Elem()
	f := rdelta.FieldByName("UserAgents")

	// map[string]int
	changed, err := edc.deltaMapField(
		reflect.ValueOf(map[string]int{
			"1": 1,
			"2": 2,
		}),
		reflect.ValueOf(map[string]int{
			"1": 2,
			"2": 2,
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("deltaMapField returned no difference found although there is a delta")
	}

	if len(delta.UserAgents) != 1 || delta.UserAgents["1"] != 2 {
		t.Error("Delta of endpoint wasn't as expected")
	}

	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]int{
			"1": 1,
			"2": 2,
		}),
		reflect.ValueOf(map[string]int{
			"1": 1,
			"2": 2,
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if changed {
		t.Error("deltaMapField returned a difference found although there isn't a delta")
	}

	if len(delta.UserAgents) != 0 {
		t.Error("Delta of endpoint wasn't as expected")
	}

	// map[string]bool
	f = rdelta.FieldByName("LLDPCapabilities")

	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]bool{
			"1": true,
			"2": true,
		}),
		reflect.ValueOf(map[string]bool{
			"1": false,
			"2": true,
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("deltaMapField returned no difference found although there is a delta")
	}

	if len(delta.LLDPCapabilities) != 1 || delta.LLDPCapabilities["1"] != false {
		t.Error("Delta of endpoint wasn't as expected")
	}

	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]bool{
			"1": false,
			"2": true,
		}),
		reflect.ValueOf(map[string]bool{
			"1": false,
			"2": true,
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if changed {
		t.Error("deltaMapField returned a difference found although there isn't a delta")
	}

	if len(delta.LLDPCapabilities) != 0 {
		t.Error("Delta of endpoint wasn't as expected")
	}

	// map[string]map[int]int
	f = rdelta.FieldByName("DestinationHosts")

	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"80":  1,
				"443": 1,
			},
			"2": map[string]int{
				"80":  1,
				"443": 1,
			},
		}),
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"80":  1,
				"443": 1,
			},
			"2": map[string]int{
				"80":  1,
				"443": 1,
				"192": 1,
			},
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("deltaMapField returned no difference found although there is a delta")
	}

	if len(delta.DestinationHosts) != 1 || delta.DestinationHosts["2"]["192"] != 1 {
		t.Error("Delta of endpoint wasn't as expected")
	}

	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"80":  1,
				"443": 1,
			},
			"2": map[string]int{
				"80":  1,
				"443": 1,
			},
		}),
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"80":  1,
				"443": 1,
			},
			"2": map[string]int{
				"80":  1,
				"443": 1,
			},
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if changed {
		t.Error("deltaMapField returned a difference found although there isn't a delta")
	}

	if len(delta.DestinationHosts) != 0 {
		t.Error("Delta of endpoint wasn't as expected")
	}

	f = rdelta.FieldByName("JA3Data")
	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
			"2": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
		}),
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
			"2": map[string]int{
				"1;1": 1,
				"2;2": 1,
				"3;3": 1,
			},
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if !changed {
		t.Error("deltaMapField returned no difference found although there is a delta")
	}

	if len(delta.JA3Data) != 1 || delta.JA3Data["2"]["3;3"] != 1 {
		t.Error("Delta of endpoint wasn't as expected")
	}

	changed, err = edc.deltaMapField(
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
			"2": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
		}),
		reflect.ValueOf(map[string]map[string]int{
			"1": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
			"2": map[string]int{
				"1;1": 1,
				"2;2": 1,
			},
		}),
		f,
	)
	sharedutils.CheckTestError(t, err)

	if changed {
		t.Error("deltaMapField returned a difference found although there isn't a delta")
	}

	if len(delta.JA3Data) != 0 {
		t.Error("Delta of endpoint wasn't as expected")
	}

}

func TestEndpointDataComparatorCompareString(t *testing.T) {
	edc := EndpointDataComparator{}
	if ok, _ := edc.compareString("1", "2"); ok {
		t.Error("2 different strings are seen as the same")
	}
	if ok, _ := edc.compareString("2", "1"); ok {
		t.Error("2 different strings are seen as the same")
	}
	if ok, _ := edc.compareString("1", "1"); !ok {
		t.Error("2 equal strings are seen as different")
	}
}

func TestEndpointDataComparatorCompareInt(t *testing.T) {
	edc := EndpointDataComparator{}
	if ok, _ := edc.compareInt(1, 2); ok {
		t.Error("2 different ints are seen as the same")
	}
	if ok, _ := edc.compareInt(2, 1); ok {
		t.Error("2 different ints are seen as the same")
	}
	if ok, _ := edc.compareInt(1, 1); !ok {
		t.Error("2 equal ints are seen as different")
	}
}

func TestEndpointDataComparatorCompareBool(t *testing.T) {
	edc := EndpointDataComparator{}
	if ok, _ := edc.compareBool(true, false); ok {
		t.Error("2 different bools are seen as the same")
	}
	if ok, _ := edc.compareBool(false, true); ok {
		t.Error("2 different bools are seen as the same")
	}
	if ok, _ := edc.compareBool(true, true); !ok {
		t.Error("2 equal bools are seen as different")
	}
	if ok, _ := edc.compareBool(false, false); !ok {
		t.Error("2 equal bools are seen as different")
	}
}

func TestEndpointDataComparatorCompareValues(t *testing.T) {
	edc := EndpointDataComparator{}
	var ok bool
	var err error

	ok, err = edc.compareValues(1, 2)
	if ok {
		t.Error("Compare values yielded an invalid result")
	}
	if err != nil {
		t.Error("Comparing values yielded an error", err)
	}

	ok, err = edc.compareValues(1, 1)
	if !ok {
		t.Error("Compare values yielded an invalid result")
	}
	if err != nil {
		t.Error("Comparing values yielded an error", err)
	}

	ok, err = edc.compareValues("1", "2")
	if ok {
		t.Error("Compare values yielded an invalid result")
	}
	if err != nil {
		t.Error("Comparing values yielded an error", err)
	}

	ok, err = edc.compareValues("1", "1")
	if !ok {
		t.Error("Compare values yielded an invalid result")
	}
	if err != nil {
		t.Error("Comparing values yielded an error", err)
	}

	ok, err = edc.compareValues(true, false)
	if ok {
		t.Error("Compare values yielded an invalid result")
	}
	if err != nil {
		t.Error("Comparing values yielded an error", err)
	}

	ok, err = edc.compareValues(true, true)
	if !ok {
		t.Error("Compare values yielded an invalid result")
	}
	if err != nil {
		t.Error("Comparing values yielded an error", err)
	}
}

func TestEndpointDataComparatorCompareMapString(t *testing.T) {
	edc := EndpointDataComparator{}

	var res []string
	var err error

	res, err = edc.compareMapString(
		map[string]interface{}{
			"1": true,
		},
		map[string]interface{}{
			"1": true,
		},
	)

	if len(res) != 0 {
		t.Error("Comparing two equal map[string]interface{} provided differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]interface{} yielded an error", err)
	}

	res, err = edc.compareMapString(
		map[string]interface{}{
			"2": true,
			"1": true,
		},
		map[string]interface{}{
			"1": true,
			"2": true,
		},
	)

	if len(res) != 0 {
		t.Error("Comparing two equal map[string]interface{} provided differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]interface{} yielded an error", err)
	}

	res, err = edc.compareMapString(
		map[string]interface{}{
			"1": true,
			"2": true,
		},
		map[string]interface{}{
			"1": false,
			"2": true,
		},
	)

	if len(res) != 1 {
		t.Error("Comparing two different map[string]interface{} didn't provide differences")
	}

	if res[0] != "1" {
		t.Error("Didn't find the right key in the difference")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]interface{} yielded an error", err)
	}

	res, err = edc.compareMapString(
		map[string]interface{}{
			"1": true,
			"2": false,
		},
		map[string]interface{}{
			"1": false,
			"2": true,
		},
	)

	if len(res) != 2 {
		t.Error("Comparing two different map[string]interface{} didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]interface{} yielded an error", err)
	}

	res, err = edc.compareMapString(
		map[string]interface{}{
			"1": true,
			"2": false,
		},
		map[string]interface{}{
			"1": false,
			"2": true,
			"3": true,
		},
	)

	if len(res) != 3 {
		t.Error("Comparing two different map[string]interface{} didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]interface{} yielded an error", err)
	}

}

func TestEndpointDataComparatorCompareMapStringInt(t *testing.T) {
	edc := EndpointDataComparator{}

	var res []string
	var err error

	res, err = edc.compareMapStringInt(
		map[string]int{
			"1": 1,
			"2": 2,
		},
		map[string]int{
			"1": 2,
			"2": 1,
		},
	)

	if len(res) != 2 {
		t.Error("Comparing two different map[string]int didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]int yielded an error", err)
	}
}
func TestEndpointDataComparatorCompareMapInt(t *testing.T) {
	edc := EndpointDataComparator{}

	var res []int
	var err error

	res, err = edc.compareMapInt(
		map[int]interface{}{
			1: true,
		},
		map[int]interface{}{
			1: true,
		},
	)

	if len(res) != 0 {
		t.Error("Comparing two equal map[int]interface{} provided differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[int]interface{} yielded an error", err)
	}

	res, err = edc.compareMapInt(
		map[int]interface{}{
			2: true,
			1: true,
		},
		map[int]interface{}{
			1: true,
			2: true,
		},
	)

	if len(res) != 0 {
		t.Error("Comparing two equal map[int]interface{} provided differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[int]interface{} yielded an error", err)
	}

	res, err = edc.compareMapInt(
		map[int]interface{}{
			1: true,
			2: true,
		},
		map[int]interface{}{
			1: false,
			2: true,
		},
	)

	if len(res) != 1 {
		t.Error("Comparing two different map[int]interface{} didn't provide differences")
	}

	if res[0] != 1 {
		t.Error("Didn't find the right key in the difference")
	}

	if err != nil {
		t.Error("Comparing two valid map[int]interface{} yielded an error", err)
	}

	res, err = edc.compareMapInt(
		map[int]interface{}{
			1: true,
			2: false,
		},
		map[int]interface{}{
			1: false,
			2: true,
		},
	)

	if len(res) != 2 {
		t.Error("Comparing two different map[int]interface{} didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[int]interface{} yielded an error", err)
	}

	res, err = edc.compareMapInt(
		map[int]interface{}{
			1: true,
			2: true,
		},
		map[int]interface{}{
			1: true,
			2: true,
			3: true,
		},
	)

	if len(res) != 1 {
		t.Error("Comparing two different map[int]interface{} didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[int]interface{} yielded an error", err)
	}

	res, err = edc.compareMapInt(
		map[int]interface{}{
			1: true,
			2: false,
		},
		map[int]interface{}{
			1: false,
			2: true,
			3: true,
		},
	)

	if len(res) != 3 {
		t.Error("Comparing two different map[string]interface{} didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[string]interface{} yielded an error", err)
	}

}

func TestEndpointDataComparatorCompareMapIntInt(t *testing.T) {
	edc := EndpointDataComparator{}

	var res []int
	var err error

	res, err = edc.compareMapIntInt(
		map[int]int{
			1: 1,
			2: 2,
		},
		map[int]int{
			1: 2,
			2: 1,
		},
	)

	if len(res) != 2 {
		t.Error("Comparing two different map[int]int didn't provide differences")
	}

	if err != nil {
		t.Error("Comparing two valid map[int]int yielded an error", err)
	}
}

func BenchmarkBuildDeltaEndpoint(b *testing.B) {
	ed1 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	ed2 := NewEndpointData(DummyEndpointDataUpdateHook, DummyEndpointDataAnalysisHook, DummyEndpointDataRunsInLiveReplicationHook)
	edc := EndpointDataComparator{}
	for i := 0; i < b.N; i++ {
		edc.BuildDeltaEndpoint(ed1, ed2)
	}
}
