package fccommon

import (
	"encoding/json"
	"reflect"
	"sync"

	"git.inverse.ca/inverse/fingerbank-collector/elru"

	"github.com/inverse-inc/go-utils/sharedutils"
)

type EndpointsCache struct {
	lock                   *sync.Mutex
	endpointsIpHostCache   *elru.EncodableLRU
	endpointsAllHostsCache *elru.EncodableLRU
	maxEndpoints           int
	maxIpHostPerEndpoint   int
	maxAllHostsPerEndpoint int
}

func NewEndpointsCache(maxEndpoints int, maxIpHostPerEndpoint int, maxAllHostsPerEndpoint int) *EndpointsCache {
	endpointsIpHostCache, err := elru.NewEncodableLRU(maxEndpoints)
	endpointsAllHostsCache, err := elru.NewEncodableLRU(maxEndpoints)
	sharedutils.CheckError(err)

	ec := &EndpointsCache{
		lock:                   &sync.Mutex{},
		endpointsIpHostCache:   endpointsIpHostCache,
		endpointsAllHostsCache: endpointsAllHostsCache,
		maxEndpoints:           maxEndpoints,
		maxIpHostPerEndpoint:   maxIpHostPerEndpoint,
		maxAllHostsPerEndpoint: maxAllHostsPerEndpoint,
	}
	return ec
}

func (ec *EndpointsCache) MarshalJSON() ([]byte, error) {
	ec.lock.Lock()
	defer ec.lock.Unlock()

	return json.Marshal(map[string]interface{}{
		"endpoints_all_hosts_cache":  ec.endpointsAllHostsCache,
		"endpoints_ip_host_cache":    ec.endpointsIpHostCache,
		"max_endpoints":              ec.maxEndpoints,
		"max_ip_host_per_endpoint":   ec.maxIpHostPerEndpoint,
		"max_all_hosts_per_endpoint": ec.maxAllHostsPerEndpoint,
	})
}

func (ec *EndpointsCache) UnmarshalJSON(data []byte) error {
	ec.lock.Lock()
	defer ec.lock.Unlock()

	m := struct {
		MaxEndpoints           int             `json:"max_endpoints"`
		MaxIpHostPerEndpoint   int             `json:"max_ip_host_per_endpoint"`
		MaxAllHostsPerEndpoint int             `json:"max_all_hosts_per_endpoint"`
		EndpointsIpHostCache   json.RawMessage `json:"endpoints_ip_host_cache"`
		EndpointsAllHostsCache json.RawMessage `json:"endpoints_all_hosts_cache"`
	}{}
	err := json.Unmarshal(data, &m)
	if err != nil {
		return err
	}

	ec.maxEndpoints = m.MaxEndpoints

	ec.maxIpHostPerEndpoint = m.MaxIpHostPerEndpoint
	ec.endpointsIpHostCache = elru.NewEncodableLRUForUnmarshal(reflect.TypeOf(&IpHostCache{}))
	ec.endpointsIpHostCache.UnmarshalJSON(m.EndpointsIpHostCache)

	ec.maxAllHostsPerEndpoint = m.MaxAllHostsPerEndpoint
	ec.endpointsAllHostsCache = elru.NewEncodableLRUForUnmarshal(reflect.TypeOf(&AllHostsCache{}))
	ec.endpointsAllHostsCache.UnmarshalJSON(m.EndpointsAllHostsCache)

	return nil
}

func (ec *EndpointsCache) AddIpHostToEndpoint(endpointId string, ip string, host string) bool {
	if ec.maxIpHostPerEndpoint <= 0 {
		return true
	}

	var ihc *IpHostCache
	ec.lock.Lock()
	defer ec.lock.Unlock()
	if o, found := ec.endpointsIpHostCache.Get(endpointId); !found {
		ihc = NewIpHostCache(ec.maxIpHostPerEndpoint)
		ec.endpointsIpHostCache.Add(endpointId, ihc)
	} else {
		ihc = o.(*IpHostCache)
	}

	return ihc.Add(ip, host)
}

func (ec *EndpointsCache) GetEndpointIpHostCache(endpointId string) (*IpHostCache, bool) {
	ec.lock.Lock()
	defer ec.lock.Unlock()
	if o, found := ec.endpointsIpHostCache.Get(endpointId); found {
		return o.(*IpHostCache), found
	} else {
		return nil, found
	}
}

func (ec *EndpointsCache) AddHostToEndpoint(endpointId string, host string, port string) bool {
	if ec.maxAllHostsPerEndpoint <= 0 {
		return true
	}

	var ahc *AllHostsCache
	ec.lock.Lock()
	defer ec.lock.Unlock()
	if o, found := ec.endpointsAllHostsCache.Get(endpointId); !found {
		ahc = NewAllHostsCache(ec.maxAllHostsPerEndpoint)
		ec.endpointsAllHostsCache.Add(endpointId, ahc)
	} else {
		ahc = o.(*AllHostsCache)
	}

	return ahc.Add(host, port)
}

func (ec *EndpointsCache) GetEndpointAllHostsCache(endpointId string) (*AllHostsCache, bool) {
	ec.lock.Lock()
	defer ec.lock.Unlock()
	if o, found := ec.endpointsAllHostsCache.Get(endpointId); found {
		return o.(*AllHostsCache), found
	} else {
		return nil, found
	}
}

func (ec *EndpointsCache) ResetEndpoint(endpointId string) {
	ec.lock.Lock()
	defer ec.lock.Unlock()

	ec.endpointsAllHostsCache.Remove(endpointId)
	ec.endpointsIpHostCache.Remove(endpointId)
}
