package fccommon

import (
	"encoding/json"
	"fmt"
	"strconv"
	"testing"

	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestEndpointsCacheJSON(t *testing.T) {
	ec := NewEndpointsCache(2, 2, 3)
	endpointId := "endpoint-1"
	values := map[string]string{
		"1.2.3.4": "example.com",
		"2.3.4.5": "zammitcorp.com",
	}

	allHosts := map[string]string{
		"example.com":    "tcp:69",
		"zammitcorp.com": "udp:1234",
		"otbm.com":       "42",
	}

	for k, v := range values {
		ec.AddIpHostToEndpoint(endpointId, k, v)
	}

	for k, v := range allHosts {
		ec.AddHostToEndpoint(endpointId, k, v)
	}

	d, err := json.Marshal(ec)
	sharedutils.CheckError(err)

	ec2 := NewEndpointsCache(10, 10, 10)
	err = json.Unmarshal(d, ec2)

	if ec2.maxEndpoints != 2 {
		t.Error("maxEndpoints wasn't restored properly")
	}

	if ec2.maxIpHostPerEndpoint != 2 {
		t.Error("maxIpHostPerEndpoint wasn't restored properly")
	}

	if ec2.maxAllHostsPerEndpoint != 3 {
		t.Error("maxAllHostsPerEndpoint wasn't restored properly")
	}

	ihc, found := ec2.GetEndpointIpHostCache(endpointId)

	if !found {
		t.Error("Unable to find endpoint in restore EndpointsCache")
	}

	for k, v := range values {
		if foundV, found := ihc.Get(k); !found || v != foundV {
			t.Error("Key", k, "wasn't found or value was unexpected. Expected", v, "Found", foundV)
		}
	}

	ahc, found := ec2.GetEndpointAllHostsCache(endpointId)

	if !found {
		t.Error("Unable to find endpoint in restore EndpointsCache")
	}

	for k, v := range allHosts {
		if foundV, found := ahc.Get(k); !found {
			t.Error("Key", k, "wasn't found")
		} else {
			if foundV[v] != 1 {
				t.Error("Unable to find port value in", k, "expected port", v, "to be present with a count of 1")
			}
		}
	}

}

func TestEndpointsCache(t *testing.T) {
	ec := NewEndpointsCache(2, 5, 5)

	expected := []int{}
	for i := 0; i < 2; i++ {
		expected = append(expected, i)

		stri := strconv.Itoa(i)
		k := "endpoint-" + stri
		ec.AddIpHostToEndpoint(k, "1.2.3."+stri, "example.com")
	}

	if ec.endpointsIpHostCache.Len() != len(expected) {
		t.Error("Resultset doesn't have the expected length")
	}

	for _, i := range expected {
		stri := strconv.Itoa(i)
		k := "endpoint-" + stri
		ihc, found := ec.GetEndpointIpHostCache(k)
		if !found {
			t.Error("Expected to find", k, "but didn't find it")
			continue
		}

		ip := "1.2.3." + stri
		if _, found := ihc.Get(ip); !found {
			t.Error("Expected to find", ip, "in ip host cache of", k, "but didn't find it")
		}
	}

}

func TestEndpointsCacheEmpty(t *testing.T) {
	// The tests below should simply not panic and return properly to succeed

	ec := NewEndpointsCache(10, 10, 0)
	ec.AddHostToEndpoint("1", "otbm.com", "tcp:80")

	ec = NewEndpointsCache(10, 0, 10)
	ec.AddIpHostToEndpoint("1", "1.2.3.4", "otbm.com")

}

func BenchmarkCacheMemUsage1Endpoints1Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		benchmarkCacheMemUsage(b, 1, 1)
	}
}

func BenchmarkCacheMemUsage1Endpoints100Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		benchmarkCacheMemUsage(b, 1, 100)
	}
}

func BenchmarkCacheMemUsage100Endpoints100Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		benchmarkCacheMemUsage(b, 100, 100)
	}
}

func BenchmarkCacheMemUsage500Endpoints50Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		benchmarkCacheMemUsage(b, 500, 50)
	}
}

func BenchmarkCacheMemUsage1000Endpoints100Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		benchmarkCacheMemUsage(b, 1000, 100)
	}
}

func BenchmarkCacheJSONMemUsage100Endpoints100Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		ec := benchmarkCacheMemUsage(b, 100, 100)
		json.Marshal(ec)
	}
}

func BenchmarkCacheJSONMemUsage500Endpoints50Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		ec := benchmarkCacheMemUsage(b, 500, 50)
		json.Marshal(ec)
	}
}

func BenchmarkCacheJSONMemUsage1000Endpoints100Hosts(b *testing.B) {
	for i := 0; i < b.N; i++ {
		ec := benchmarkCacheMemUsage(b, 1000, 100)
		json.Marshal(ec)
	}
}

func benchmarkCacheMemUsage(b *testing.B, endpointsSize, hostSize int) *EndpointsCache {
	ec := NewEndpointsCache(endpointsSize, hostSize, hostSize)

	for i := 0; i < endpointsSize; i++ {
		stri := strconv.Itoa(i)
		endpointId := "00:00:00:00:00:00-" + stri
		for j := 0; j < hostSize; j++ {
			strj := strconv.Itoa(j)
			ip := "1.2.3.4-" + strj
			ec.AddIpHostToEndpoint(endpointId, ip, "fakedomain.com")
			ec.AddHostToEndpoint(endpointId, strj+".fakedomain.com", fmt.Sprintf("tcp:%d", j))
		}
	}
	return ec
}
