package fccommon

import (
	"encoding/json"
	"strconv"
	"testing"

	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestIpHostCache(t *testing.T) {
	ihc := NewIpHostCache(5)

	expected := []string{}
	for i := 0; i < 5; i++ {
		stri := strconv.Itoa(i)
		k := "1.2.3." + stri
		ihc.Add(k, stri+"-example.com")
		expected = append(expected, k)
	}

	ihm := ihc.ToMap()

	if len(ihm) != len(expected) {
		t.Error("Length of ip host map is greater than the max size")
	}

	for _, k := range expected {
		if _, found := ihm[k]; !found {
			t.Error("Unable to find key in ip host map", k)
		}
	}

	// Test starting again from scratch with a range of 10
	// 0-4 should be flushed and 5-9 should be there
	ihc = NewIpHostCache(5)

	expected = []string{}
	for i := 0; i < 10; i++ {
		stri := strconv.Itoa(i)
		k := "1.2.3." + stri
		ihc.Add(k, stri+"-example.com")
		if i >= 5 {
			expected = append(expected, k)
		}
	}

	ihm = ihc.ToMap()

	if len(ihm) != len(expected) {
		t.Error("Length of ip host map is greater than the max size")
	}

	for _, k := range expected {
		if _, found := ihm[k]; !found {
			t.Error("Unable to find key in ip host map", k)
		}
	}

}

func TestIpHostCacheJSON(t *testing.T) {
	ihc := NewIpHostCache(2)
	values := map[string]string{
		"1.2.3.4": "example.com",
		"2.3.4.5": "zammitcorp.com",
	}
	for k, v := range values {
		ihc.Add(k, v)
	}
	d, err := json.Marshal(ihc)
	sharedutils.CheckError(err)

	ihc2 := &IpHostCache{}
	err = json.Unmarshal(d, ihc2)

	if ihc.cache.MaxSize() != ihc2.cache.MaxSize() {
		t.Error("maxSize wasn't restored properly")
	}

	for k, v := range values {
		if foundV, found := ihc2.Get(k); !found || v != foundV {
			t.Error("Key", k, "wasn't found or value was unexpected. Expected", v, "Found", foundV)
		}
	}

	sharedutils.CheckError(err)
}

func TestIPHostCacheMultiAdd(t *testing.T) {
	ihc := NewIpHostCache(2)
	ip := "1.2.3.4"
	domain := "example.com"
	if ihc.Add(ip, domain) {
		t.Error("Adding a non-existing ip/host mapping returned true")
	}
	if !ihc.Add(ip, domain) {
		t.Error("Adding an existing ip/host mapping returned false")
	}
}
