package fccommon

import (
	"fmt"
	"net"
	"regexp"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
)

type NBAPolicy struct {
	PolicyID string
	// Whether or not the policy is enabled
	Status string `ini:"status"`
	//Which devices (and childs) to take into consideration in the policy. Empty means all
	DevicesIncluded []int `ini:"devices_included"`
	//Which devices (and childs) to exclude from the devices_included
	DevicesExcluded []int `ini:"devices_excluded"`
	//Whether or not to check for hosts talking with blacklisted IPs
	WatchBlacklistedIPs string `ini:"watch_blacklisted_ips"`
	//Which IPs (can be CIDR) to ignore in the blacklisted IPs
	WhitelistedIPs []*net.IPNet
	//How many blacklisted hosts can a device talk to before being flagged
	BlacklistedIpHostsThreshold int `ini:"blacklisted_ip_hosts_threshold"`
	///The window to look into for the blacklisted IPs
	BlacklistedIpHostsWindow time.Duration `ini:"blacklisted_ip_hosts_window"`
	//Which attributes to match against the profile data
	WatchedDeviceAttributes map[string]bool
	//Overrides of the hits when an attribute matches or not
	DeviceAttributesDiffThresholdOverrides map[string]int
	//Difference score that is allowed in the attributes
	//Anything lower than this will be considered bad
	DeviceAttributesDiffScore int `ini:"device_attributes_diff_score"`
	//Blacklisted ports
	BlacklistedPorts       map[uint16]bool
	BlacklistedPortsWindow time.Duration `ini:"blacklisted_ports_window"`
}

func (nbap *NBAPolicy) WatchDhcpFingerprint() bool {
	val, ok := nbap.WatchedDeviceAttributes[DHCP_FINGERPRINT]
	return val && ok
}

func (nbap *NBAPolicy) WatchDhcpVendor() bool {
	val, ok := nbap.WatchedDeviceAttributes[DHCP_VENDOR]
	return val && ok
}

func (nbap *NBAPolicy) WatchHostname() bool {
	val, ok := nbap.WatchedDeviceAttributes[HOSTNAME]
	return val && ok
}

func (nbap *NBAPolicy) WatchOUI() bool {
	val, ok := nbap.WatchedDeviceAttributes[OUI]
	return val && ok
}

func (nbap *NBAPolicy) WatchDestinationHosts() bool {
	val, ok := nbap.WatchedDeviceAttributes[DESTINATION_HOSTS]
	return val && ok
}

func (nbap *NBAPolicy) WatchMDNSServices() bool {
	val, ok := nbap.WatchedDeviceAttributes[MDNS_SERVICES]
	return val && ok
}

func (nbap *NBAPolicy) WatchTcpSynSignatures() bool {
	val, ok := nbap.WatchedDeviceAttributes[TCP_SYN_SIGNATURES]
	return val && ok
}

func (nbap *NBAPolicy) WatchTcpSynAckSignatures() bool {
	val, ok := nbap.WatchedDeviceAttributes[TCP_SYN_ACK_SIGNATURES]
	return val && ok
}

func (nbap *NBAPolicy) WatchUpnpServerStrings() bool {
	val, ok := nbap.WatchedDeviceAttributes[UPNP_SERVER_STRINGS]
	return val && ok
}

func (nbap *NBAPolicy) WatchUpnpUserAgents() bool {
	val, ok := nbap.WatchedDeviceAttributes[UPNP_USER_AGENTS]
	return val && ok
}

func (nbap *NBAPolicy) WatchUserAgents() bool {
	val, ok := nbap.WatchedDeviceAttributes[USER_AGENTS]
	return val && ok
}

type NetworkProfileEngine struct {
	Rules  map[string][]*regexp.Regexp
	Policy *NBAPolicy
}

func (npe *NetworkProfileEngine) DiffWithEndpointData(fc *FingerbankCollector, ed *EndpointData) ([]string, []string) {

	matched := []string{}
	unmatched := []string{}
	if npe.Policy.WatchDhcpFingerprint() {
		if len(npe.Rules[DHCP_FINGERPRINT]) != 0 && ed.DhcpFingerprint != "" {
			hasMatched := false
			for _, rule := range npe.Rules[DHCP_FINGERPRINT] {
				if rule.MatchString(ed.DhcpFingerprint) {
					hasMatched = true
				}
			}
			if !hasMatched {
				unmatched = append(unmatched, DHCP_FINGERPRINT)
			} else {
				matched = append(matched, DHCP_FINGERPRINT)
			}
		}
	}

	if npe.Policy.WatchDhcpVendor() {
		if len(npe.Rules[DHCP_VENDOR]) != 0 && ed.DhcpVendor != "" {
			hasMatched := false
			for _, rule := range npe.Rules[DHCP_VENDOR] {
				if rule.MatchString(ed.DhcpVendor) {
					hasMatched = true
				}
			}
			if !hasMatched {
				unmatched = append(unmatched, DHCP_VENDOR)
			} else {
				matched = append(matched, DHCP_VENDOR)
			}
		}
	}

	if npe.Policy.WatchHostname() {
		if len(npe.Rules[HOSTNAME]) != 0 && ed.Hostname != "" {
			hasMatched := false
			for _, rule := range npe.Rules[HOSTNAME] {
				if rule.MatchString(ed.Hostname) {
					hasMatched = true
				}
			}
			if !hasMatched {
				unmatched = append(unmatched, HOSTNAME)
			} else {
				matched = append(matched, HOSTNAME)
			}
		}
	}

	if npe.Policy.WatchOUI() {
		if len(npe.Rules[OUI]) != 0 && ed.OUI != "" {
			hasMatched := false
			for _, rule := range npe.Rules[OUI] {
				if rule.MatchString(ed.OUI) {
					hasMatched = true
				}
			}
			if !hasMatched {
				unmatched = append(unmatched, OUI)
			} else {
				matched = append(matched, OUI)
			}
		}
	}

	if npe.Policy.WatchDestinationHosts() {
		if len(npe.Rules[DESTINATION_HOSTS]) != 0 && len(ed.AllDestinationHosts(fc)) != 0 {
			for v, _ := range ed.AllDestinationHosts(fc) {
				hasMatched := false
				for _, rule := range npe.Rules[DESTINATION_HOSTS] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, DESTINATION_HOSTS)
				} else {
					matched = append(matched, DESTINATION_HOSTS)
				}
			}
		}
	}

	if npe.Policy.WatchMDNSServices() {
		if len(npe.Rules[MDNS_SERVICES]) != 0 && len(ed.MDNSServices) != 0 {
			for v, _ := range ed.MDNSServices {
				hasMatched := false
				for _, rule := range npe.Rules[MDNS_SERVICES] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, MDNS_SERVICES)
				} else {
					matched = append(matched, MDNS_SERVICES)
				}
			}
		}
	}

	if npe.Policy.WatchTcpSynSignatures() {
		if len(npe.Rules[TCP_SYN_SIGNATURES]) != 0 && len(ed.TCPSynSignatures) != 0 {
			for v, _ := range ed.TCPSynSignatures {
				hasMatched := false
				for _, rule := range npe.Rules[TCP_SYN_SIGNATURES] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, TCP_SYN_SIGNATURES)
				} else {
					matched = append(matched, TCP_SYN_SIGNATURES)
				}
			}
		}
	}

	if npe.Policy.WatchTcpSynAckSignatures() {
		if len(npe.Rules[TCP_SYN_ACK_SIGNATURES]) != 0 && len(ed.TCPSynAckSignatures) != 0 {
			for v, _ := range ed.TCPSynAckSignatures {
				hasMatched := false
				for _, rule := range npe.Rules[TCP_SYN_ACK_SIGNATURES] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, TCP_SYN_ACK_SIGNATURES)
				} else {
					matched = append(matched, TCP_SYN_ACK_SIGNATURES)
				}
			}
		}
	}

	if npe.Policy.WatchUpnpServerStrings() {
		if len(npe.Rules[UPNP_SERVER_STRINGS]) != 0 && len(ed.UPnPServerStrings) != 0 {
			for v, _ := range ed.UPnPServerStrings {
				hasMatched := false
				for _, rule := range npe.Rules[UPNP_SERVER_STRINGS] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, UPNP_SERVER_STRINGS)
				} else {
					matched = append(matched, UPNP_SERVER_STRINGS)
				}
			}
		}
	}

	if npe.Policy.WatchUpnpUserAgents() {
		if len(npe.Rules[UPNP_USER_AGENTS]) != 0 && len(ed.UPnPUserAgents) != 0 {
			for v, _ := range ed.UPnPUserAgents {
				hasMatched := false
				for _, rule := range npe.Rules[UPNP_USER_AGENTS] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, UPNP_USER_AGENTS)
				} else {
					matched = append(matched, UPNP_USER_AGENTS)
				}
			}
		}
	}

	if npe.Policy.WatchUserAgents() {
		if len(npe.Rules[USER_AGENTS]) != 0 && len(ed.UserAgents) != 0 {
			for v, _ := range ed.UserAgents {
				hasMatched := false
				for _, rule := range npe.Rules[USER_AGENTS] {
					if rule.MatchString(v) {
						hasMatched = true
					}
				}
				if !hasMatched {
					unmatched = append(unmatched, USER_AGENTS)
				} else {
					matched = append(matched, USER_AGENTS)
				}
			}
		}
	}

	return matched, unmatched
}

func (npe *NetworkProfileEngine) getDiffAttributeHit(attribute string) int {
	if npe.Policy.DeviceAttributesDiffThresholdOverrides != nil {
		if diff, ok := npe.Policy.DeviceAttributesDiffThresholdOverrides[attribute]; ok {
			return diff
		}
	}

	return DIFF_THRESHOLD_HITS[attribute]
}

func (npe *NetworkProfileEngine) DiffScore(fc *FingerbankCollector, ed *EndpointData) int {
	matched, unmatched := npe.DiffWithEndpointData(fc, ed)

	diffScore := 0
	for _, v := range matched {
		diff := npe.getDiffAttributeHit(v)
		diffScore += diff
		fc.logger.Debug(fmt.Sprintf("Positive hit on attribute %s with value %d", v, diff))
	}
	for _, v := range unmatched {
		diff := npe.getDiffAttributeHit(v)
		diffScore -= diff
		fc.logger.Debug(fmt.Sprintf("Negative hit on attribute %s with value %d", v, diff))
	}

	fc.logger.Debug(fmt.Sprintf("Got diffScore %d", diffScore))

	return diffScore
}

func (npe *NetworkProfileEngine) AnalyseEndpoint(fc *FingerbankCollector, res *QueryResult) []string {
	events := []string{}

	if npe.Policy.DevicesExcluded != nil && len(npe.Policy.DevicesExcluded) > 0 {
		for _, id := range npe.Policy.DevicesExcluded {
			if res.CloudAPIResult.Device.Id == id {
				fc.logger.Debug(fmt.Sprintf("Ignoring device %d because its part of the ignored list of policy %s", res.CloudAPIResult.Device.Id, npe.Policy.PolicyID))
				return events
			}
		}
	}

	found := false
	if npe.Policy.DevicesIncluded != nil && len(npe.Policy.DevicesIncluded) > 0 {
		for _, id := range npe.Policy.DevicesIncluded {
			if res.CloudAPIResult.Device.Id == id {
				fc.logger.Debug(fmt.Sprintf("Processing device %d that is part of the included devices of policy %s", res.CloudAPIResult.Device.Id, npe.Policy.PolicyID))
				found = true
			}
		}
	} else {
		fc.logger.Debug(fmt.Sprintf("Processing device %d as policy %s is made for all devices", res.CloudAPIResult.Device.Id, npe.Policy.PolicyID))
		found = true
	}

	if !found {
		fc.logger.Debug(fmt.Sprintf("Ignoring device %d because its not part of the included devices of policy %s", res.CloudAPIResult.Device.Id, npe.Policy.PolicyID))
		return events
	}

	if len(npe.Policy.WatchedDeviceAttributes) != 0 {
		diffScore := npe.DiffScore(fc, &res.EndpointData.EndpointData)

		if diffScore < npe.Policy.DeviceAttributesDiffScore {
			events = append(events, DIFF_SCORE_TOO_LOW_EVENT)
		}
	}

	if npe.Policy.WatchBlacklistedIPs == "enabled" {
		foundBlacklistedIPs := 0
		beginWindow := time.Now().Add(-npe.Policy.BlacklistedIpHostsWindow)
		for ipStr, info := range res.EndpointData.BlacklistedCommunications {
			ip := net.ParseIP(ipStr)
			if ip == nil {
				continue
			}

			ignored := false
			for _, ignoredNet := range npe.Policy.WhitelistedIPs {
				if ignoredNet.Contains(ip) {
					fc.logger.Debug(fmt.Sprintf("Ignoring blacklisted IP %s due to configuration", ip))
					ignored = true
				}
			}
			if !ignored {
				for _, lastSeen := range info {
					if lastSeen.After(beginWindow) {
						fc.logger.Debug(fmt.Sprintf("Found recent blacklisted IP %s", ip))
						foundBlacklistedIPs += 1
						break
					}
				}
			}
		}

		if foundBlacklistedIPs >= npe.Policy.BlacklistedIpHostsThreshold {
			events = append(events, BLACKLISTED_IPS_THRESHOLD_TOO_HIGH)
		}
	}

	if len(npe.Policy.BlacklistedPorts) > 0 {
		beginWindow := time.Now().Add(-npe.Policy.BlacklistedPortsWindow)
		for port, lastSeen := range res.EndpointData.BlacklistedCommunications[BlacklistedCommunicationsPortsID] {
			if lastSeen.After(beginWindow) {
				fc.logger.Debug(fmt.Sprintf("Found recent blacklisted port %d", port))
				events = append(events, BLACKLISTED_PORTS_THRESHOLD_TOO_HIGH)
				break
			}
		}
	}

	return events
}

func (npe *NetworkProfileEngine) regexpForProfilingRule(v string) *regexp.Regexp {
	if string(v[0]) == "/" && string(v[len(v)-1]) == "/" {
		return regexp.MustCompile(string(v[1 : len(v)-1]))
	} else {
		return regexp.MustCompile(`^` + v + `$`)
	}
}

func (npe *NetworkProfileEngine) loadRulesFromProfile(profile fingerbankclient.NetworkProfile) {
	npe.Rules = map[string][]*regexp.Regexp{}

	npe.Rules[DHCP_FINGERPRINT] = []*regexp.Regexp{}
	for _, v := range profile.DhcpFingerprints {
		npe.Rules[DHCP_FINGERPRINT] = append(npe.Rules[DHCP_FINGERPRINT], npe.regexpForProfilingRule(v))
	}

	npe.Rules[DHCP_VENDOR] = []*regexp.Regexp{}
	for _, v := range profile.DhcpVendors {
		npe.Rules[DHCP_VENDOR] = append(npe.Rules[DHCP_VENDOR], npe.regexpForProfilingRule(v))
	}

	npe.Rules[HOSTNAME] = []*regexp.Regexp{}
	for _, v := range profile.Hostnames {
		npe.Rules[HOSTNAME] = append(npe.Rules[HOSTNAME], npe.regexpForProfilingRule(v))
	}

	npe.Rules[OUI] = []*regexp.Regexp{}
	for _, v := range profile.OUIs {
		npe.Rules[OUI] = append(npe.Rules[OUI], npe.regexpForProfilingRule(v))
	}

	npe.Rules[DESTINATION_HOSTS] = []*regexp.Regexp{}
	for _, v := range profile.DestinationHosts {
		npe.Rules[DESTINATION_HOSTS] = append(npe.Rules[DESTINATION_HOSTS], npe.regexpForProfilingRule(v))
	}

	npe.Rules[MDNS_SERVICES] = []*regexp.Regexp{}
	for _, v := range profile.MdnsServices {
		npe.Rules[MDNS_SERVICES] = append(npe.Rules[MDNS_SERVICES], npe.regexpForProfilingRule(v))
	}

	npe.Rules[TCP_SYN_SIGNATURES] = []*regexp.Regexp{}
	for _, v := range profile.TcpSynSignatures {
		npe.Rules[TCP_SYN_SIGNATURES] = append(npe.Rules[TCP_SYN_SIGNATURES], npe.regexpForProfilingRule(v))
	}

	npe.Rules[TCP_SYN_ACK_SIGNATURES] = []*regexp.Regexp{}
	for _, v := range profile.TcpSynAckSignatures {
		npe.Rules[TCP_SYN_ACK_SIGNATURES] = append(npe.Rules[TCP_SYN_ACK_SIGNATURES], npe.regexpForProfilingRule(v))
	}

	npe.Rules[UPNP_SERVER_STRINGS] = []*regexp.Regexp{}
	for _, v := range profile.UpnpServerStrings {
		npe.Rules[UPNP_SERVER_STRINGS] = append(npe.Rules[UPNP_SERVER_STRINGS], npe.regexpForProfilingRule(v))
	}

	npe.Rules[UPNP_USER_AGENTS] = []*regexp.Regexp{}
	for _, v := range profile.UpnpUserAgents {
		npe.Rules[UPNP_USER_AGENTS] = append(npe.Rules[UPNP_USER_AGENTS], npe.regexpForProfilingRule(v))
	}

	npe.Rules[USER_AGENTS] = []*regexp.Regexp{}
	for _, v := range profile.UserAgents {
		npe.Rules[USER_AGENTS] = append(npe.Rules[USER_AGENTS], npe.regexpForProfilingRule(v))
	}

}

func NewNetworkProfileEngine(profile fingerbankclient.NetworkProfile, policy *NBAPolicy) *NetworkProfileEngine {
	npe := &NetworkProfileEngine{Policy: policy}
	npe.loadRulesFromProfile(profile)
	return npe
}
