package fcreplication

import (
	"encoding/json"
	"errors"
	"fmt"
	"testing"
	"time"

	fclog "git.inverse.ca/inverse/fingerbank-collector/log"
	"github.com/inverse-inc/go-utils/sharedutils"
)

func testBatcherRecorder(b *Batcher, todo func(string, []UnencryptedReplicationMessage)) func(string, []byte) {
	return func(id string, data []byte) {
		decoded, err := b.DecodeMessage(id, data)
		sharedutils.CheckError(err)

		todo(id, decoded)
	}
}

func TestBatcherFlushInterval(t *testing.T) {
	b := NewBatcher(fclog.Logger(""))
	b.Start()

	id := "bobby"
	var recorded string
	f := testBatcherRecorder(b, func(id string, decoded []UnencryptedReplicationMessage) {
		for _, data := range decoded {
			recorded += string(data.Content) + ","
		}
	})

	b.f = f

	values := []UnencryptedReplicationMessage{UnencryptedReplicationMessage{Content: json.RawMessage(`"value1"`)}, UnencryptedReplicationMessage{Content: json.RawMessage(`"value2"`)}}
	for _, value := range values {
		jdata, err := json.Marshal(value)
		sharedutils.CheckError(err)
		b.Add(id, jdata)
	}

	time.Sleep(FLUSH_AT_LEAST + 10*time.Millisecond)

	if recorded != `"value1","value2",` {
		t.Error("Unexpected value in batched packet", recorded)
	}
}

func TestBatcherBatchSize(t *testing.T) {
	b := NewBatcher(fclog.Logger(""))

	id := "bobby"
	var recorded string
	f := testBatcherRecorder(b, func(id string, decoded []UnencryptedReplicationMessage) {
		for _, data := range decoded {
			recorded += string(data.Content) + ","
		}
	})

	b.f = f

	b.MaxByteSize = 55

	for i := 0; i < 6; i++ {
		msg := UnencryptedReplicationMessage{Content: json.RawMessage([]byte(fmt.Sprintf(`"0%d"`, i)))}
		jdata, err := json.Marshal(msg)
		sharedutils.CheckError(err)
		b.Add(id, jdata)
	}

	if recorded != `"00","01","02","03","04",` {
		t.Error("Unexpected value in batched packet", recorded)
	}

	b.MaxByteSize = 40

	for i := 0; i < 5; i++ {
		msg := UnencryptedReplicationMessage{Content: json.RawMessage([]byte(fmt.Sprintf(`"0%d"`, i)))}
		jdata, err := json.Marshal(msg)
		sharedutils.CheckError(err)
		b.Add(id, jdata)
	}

	if recorded != `"00","01","02","03","04","05","00","01","02","03",` {
		t.Error("Unexpected value in batched packet", recorded)
	}

	b.flushAllNow()

	if recorded != `"00","01","02","03","04","05","00","01","02","03","04",` {
		t.Error("Unexpected value in batched packet", recorded)
	}

}

func TestBatcherMultiple(t *testing.T) {
	b := NewBatcher(fclog.Logger(""))
	b.Start()

	id1 := "bobby1"
	id2 := "bobby2"
	var recorded1 string
	var recorded2 string
	f := testBatcherRecorder(b, func(id string, decoded []UnencryptedReplicationMessage) {
		for _, data := range decoded {
			if id == id1 {
				recorded1 += string(data.Content) + ","
			} else if id == id2 {
				recorded2 += string(data.Content) + ","
			}
		}
	})

	b.f = f
	countToSend := 5
	b.MaxByteSize = countToSend * 20

	for i := 0; i < countToSend-1; i++ {
		msg := UnencryptedReplicationMessage{Content: json.RawMessage([]byte(fmt.Sprintf(`"%d"`, i)))}
		jdata, err := json.Marshal(msg)
		sharedutils.CheckError(err)
		b.Add(id1, jdata)
	}

	for i := 5; i < 5+countToSend-1; i++ {
		msg := UnencryptedReplicationMessage{Content: json.RawMessage([]byte(fmt.Sprintf(`"%d"`, i)))}
		jdata, err := json.Marshal(msg)
		sharedutils.CheckError(err)
		b.Add(id2, jdata)
	}

	time.Sleep(FLUSH_AT_LEAST + 10*time.Millisecond)

	if recorded1 != `"0","1","2","3",` {
		t.Error("Unexpected value in batched packet", recorded1)
	}

	if recorded2 != `"5","6","7","8",` {
		t.Error("Unexpected value in batched packet", recorded2)
	}

}

func TestBatcherLargeMessages(t *testing.T) {
	b := NewBatcher(fclog.Logger(""))

	id := "bobby"
	var recorded string
	f := testBatcherRecorder(b, func(id string, decoded []UnencryptedReplicationMessage) {
		for _, data := range decoded {
			recorded += string(data.Content)
		}
	})

	b.f = f

	size := MAX_MESSAGE_SIZE - GZIP_HEADER_SIZE - 25
	data := make([]byte, size, size)
	shouldBe := `"`
	data[0] = 0x22
	for i := 1; i < MAX_MESSAGE_SIZE-GZIP_HEADER_SIZE-25-1; i++ {
		data[i] = 0x31
		shouldBe += "1"
	}
	data[len(data)-1] = 0x22
	shouldBe += `"`

	msg := UnencryptedReplicationMessage{Content: data}
	jdata, err := json.Marshal(msg)
	sharedutils.CheckError(err)
	b.Add(id, jdata)
	b.flushAllNow()

	if recorded != shouldBe {
		t.Error("The batcher failed to accept a message that was large")
	}

	recorded = ""
	data = make([]byte, MAX_MESSAGE_SIZE+1, MAX_MESSAGE_SIZE+1)
	for i := 0; i < MAX_MESSAGE_SIZE+1; i++ {
		data[i] = 0x1
	}

	err = b.Add(id, data)
	b.flushAllNow()

	if len(recorded) != 0 {
		t.Error("The batcher accepted a message that is larger than the MAX_MESSAGE_SIZE")
	}

	if !errors.Is(err, ErrorBatcherMessageTooLarge) {
		t.Error("The batcher didn't return an error when the message was too large")
	}

}
