package fcreplication

import (
	"fmt"
	"net"
	"testing"
	"time"

	"github.com/google/uuid"
	"github.com/inverse-inc/go-utils/sharedutils"
)

// Like in our internal example, collector 1 is joining collector 2
// The Hello is handled by collector 2
func TestControllerHandshake(t *testing.T) {
	ctrl1UUID := uuid.New()
	ctrl2UUID := uuid.New()

	fmt.Println("Collector 1 UUID", ctrl1UUID)
	fmt.Println("Collector 2 UUID", ctrl2UUID)

	ctrl1 := NewController(ctrl1UUID, []byte("1234567890123456"), []net.IP{}, true)
	ctrl2 := NewController(ctrl2UUID, []byte("1234567890123456"), []net.IP{}, true)

	helloReplyMsg, err := ctrl2.handlePacketHello(
		net.ParseIP("127.0.0.1"),
		ReplicationMessage{
			Type: REP_HELLO,
		},
		NewREPHello(ctrl1UUID, 1),
	)
	sharedutils.CheckError(err)

	if ctrl2.outboundPeers[ctrl1UUID.String()] == nil {
		t.Error("handlePacketHello failed to created peer that is in the packet")
	}

	twominago := time.Now().Add((-2 * time.Minute))
	if ctrl2.outboundPeers[ctrl1UUID.String()].upSince.Before(twominago) {
		t.Error("Uptime tracker isn't correct after handlePacketHello")
	}

	// The HELLO-REPLY is handled by collector 1
	var helloReply REPHelloReply
	err = ctrl1.decodePacketHelloReply(helloReplyMsg, &helloReply)
	sharedutils.CheckError(err)

	challengeMsg, err := ctrl1.handlePacketHelloReply(
		net.ParseIP("127.0.0.1"),
		helloReplyMsg,
		helloReply,
	)
	sharedutils.CheckError(err)

	if string(ctrl2.outboundPeers[ctrl1UUID.String()].Secret) != string(ctrl1.inboundPeers[ctrl2UUID.String()].Secret) {
		t.Error("The secret isn't the same on both peers after HELLO-REPLY")
	}

	// The CHALLENGE is handled by collector 2
	var challenge REPChallenge
	err = ctrl2.decodePacketChallenge(challengeMsg, &challenge)
	sharedutils.CheckError(err)

	if string(challenge.ChallengeWord()) != string(ctrl1.inboundPeers[ctrl2UUID.String()].challengeWord) {
		t.Error("The challenge word sent to collector 2 isn't the one that was sent by controller 1")
	}

	challengeReplyMsg, err := ctrl2.handlePacketChallenge(
		net.ParseIP("127.0.0.1"),
		challengeMsg,
		challenge,
	)
	sharedutils.CheckError(err)

	if !ctrl2.outboundPeers[ctrl1UUID.String()].friend {
		t.Error("Peer is not a friend after receiving the challenge")
	}

	// The CHALLENGE-REPLY is handled by collector 1
	var challengeReply REPChallenge
	err = ctrl1.decodePacketChallengeReply(challengeReplyMsg, &challengeReply)
	sharedutils.CheckError(err)

	// Make sure all the challenge words match up
	same := sharedutils.AllEquals(
		string(challengeReply.ChallengeWord()),
		string(challenge.ChallengeWord()),
		string(ctrl1.inboundPeers[ctrl2UUID.String()].challengeWord),
		string(ctrl2.outboundPeers[ctrl1UUID.String()].challengeWord),
	)

	if !same {
		t.Error("One of the challenge words didn't match up between the peers")
	}

	err = ctrl1.handlePacketChallengeReply(
		net.ParseIP("127.0.0.1"),
		challengeReplyMsg,
		challengeReply,
	)
	sharedutils.CheckError(err)

	if !ctrl1.inboundPeers[ctrl2UUID.String()].friend {
		t.Error("Peer is not a friend after receiving the challenge-reply")
	}
}

func TestControllerHandshakeFailures(t *testing.T) {
	ctrl := NewController(uuid.New(), []byte("1234567890123456"), []net.IP{}, true)
	localip := net.ParseIP("127.0.0.1")
	// Test a malformed HELLO
	{
		msg := ctrl.NewReplicationMessage(REP_HELLO, []byte{})
		h := REPHello("invalidstuffhere")
		_, err := ctrl.handlePacketHello(localip, msg, h)

		if err == nil {
			t.Error("Invalid HELLO didn't yield an error")
		}
	}

}
