package fcreplication

import (
	"encoding/json"
	"errors"
	"net"
)

type UnencryptedReplicationMessage struct {
	Type    uint8           `json:"type"`
	Content json.RawMessage `json:"content"`
}

type ReplicationMessage struct {
	Type             uint8  `json:"type"`
	CollectorUUID    string `json:"collector_uuid"`
	EncryptedContent []byte `json:"encrypted_content"`
}

func (ctrl *Controller) sendReplicationMessage(peerUUID string, peerIP net.IP, msg ReplicationMessage) error {
	jsonBytes, err := json.Marshal(msg)
	if err != nil {
		return err
	}

	if peer, ok := ctrl.outboundPeers[peerUUID]; ok {
		return peer.SendOutboundMsg(jsonBytes)
	} else {
		return ctrl.sendPacket(peerIP, jsonBytes)
	}
}

func (ctrl *Controller) NewReplicationMessage(msgType uint8, encryptedContent []byte) ReplicationMessage {
	return ReplicationMessage{
		CollectorUUID:    ctrl.UUID.String(),
		Type:             msgType,
		EncryptedContent: encryptedContent,
	}
}

func (ctrl *Controller) decodeReplicationMessage(packet []byte) (ReplicationMessage, error) {
	msg := ReplicationMessage{}
	err := json.Unmarshal(packet, &msg)
	return msg, err
}

// Decode the content of a ReplicationMessage into a pointer's value
func (ctrl *Controller) decodePacketContent(msg ReplicationMessage, into interface{}) error {
	// Decryption of the packet depends on the type of the packet in the handshake
	switch msg.Type {
	case REP_HELLO:
		return ctrl.decodePacketHello(msg, into)
	case REP_HELLO_REPLY:
		return ctrl.decodePacketHelloReply(msg, into)
	case REP_CHALLENGE:
		return ctrl.decodePacketChallenge(msg, into)
	case REP_CHALLENGE_REPLY:
		return ctrl.decodePacketChallengeReply(msg, into)

	// For any other packets, the peer based secret should be used and the packets should be JSON encoded
	default:
		// ACK messages need to find outbound peers instead
		var peer *FingerbankCollectorPeer
		var err error
		if msg.Type == REP_ACK {
			peer, err = ctrl.FindOutboundPeer(msg.CollectorUUID)
		} else {
			peer, err = ctrl.FindInboundPeer(msg.CollectorUUID)
		}
		if err != nil {
			return err
		}

		content, err := ctrl.decryptMessage(peer.Secret, msg.EncryptedContent)
		err = json.Unmarshal(content, into)
		return err
	}

	return errors.New("Couldn't decode packet content, no message type matched")
}

func (ctrl *Controller) decodeUnencryptedPacketContent(msg UnencryptedReplicationMessage, into interface{}) error {
	err := json.Unmarshal(msg.Content, into)
	return err
}
