package fccommon

import (
	"crypto/rand"
	"encoding/hex"
	"errors"
	"fmt"
	"net"
	"regexp"
	"strings"
	"time"
)

const (
	DEFAULT_API_PORT = 8080
)

var validMACRegexp = regexp.MustCompile(`[0-9A-Fa-f]{12}`)

func ValidMAC(mac string) bool {
	mac = CleanMAC(mac)
	return validMACRegexp.Match([]byte(mac))
}

var characterMACExtractionRegexp = regexp.MustCompile(`[0-9A-Fa-f]+`)

func CleanMAC(mac string) string {
	matches := characterMACExtractionRegexp.FindAllString(mac, -1)
	mac = ""
	for _, m := range matches {
		mac += m
	}

	if len(mac) != 12 {
		return ""
	}

	return strings.ToLower(mac)
}

func MacFromBytes(data []byte) (string, error) {
	if len(data) < 6 {
		return "", errors.New("Not enough data to extract mac")
	}
	return hex.EncodeToString(data[:6]), nil
}

var validIPRegexp = regexp.MustCompile(`^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$`)

func ValidIP(ip string) bool {
	return validIPRegexp.Match([]byte(ip))
}

func ValidIPv6(ip string) bool {
	// Shouldn't be an IPv4
	if ValidIP(ip) {
		return false
	}

	netip := net.ParseIP(ip)
	return netip != nil && netip.To16() != nil
}

func CleanIPv6(ip string) string {
	netip := net.ParseIP(ip)
	return netip.To16().String()
}

func RandomMAC() (string, error) {
	buf := make([]byte, 6)
	_, err := rand.Read(buf)
	if err != nil {
		return "", err
	}
	// Set the local bit
	buf[0] |= 2
	return fmt.Sprintf("%02x%02x%02x%02x%02x%02x", buf[0], buf[1], buf[2], buf[3], buf[4], buf[5]), nil
}

// Creates a fake MAC address based on the IP address so that for a specific IP address, the same fake MAC would be generated consistently. This only works with IPv4 addresses
func FakeMAC(ip net.IP) (string, error) {
	localOui := []byte{0x02, 0x00}
	ip4 := ip.To4()

	if ip4 == nil {
		return "", errors.New("IP address is not an IPv4 address")
	}

	buf := append(localOui, ip4...)
	return fmt.Sprintf("%02x%02x%02x%02x%02x%02x", buf[0], buf[1], buf[2], buf[3], buf[4], buf[5]), nil
}

func MustParseDuration(s string) time.Duration {
	duration, err := time.ParseDuration(s)
	if err != nil {
		panic(err)
	}
	return duration
}
