package fccommon

import (
	"net"
	"testing"
)

func TestCleanMAC(t *testing.T) {
	expected := "0011223344aa"

	var mac, result string

	// Test column delimited
	mac = "00:11:22:33:44:aa"
	result = CleanMAC(mac)

	if result != expected {
		t.Errorf("MAC wasn't cleaned properly. Got %s instead of %s", result, expected)
	}

	// Test dash delimited
	mac = "00-11-22-33-44-aa"
	result = CleanMAC(mac)

	if result != expected {
		t.Errorf("MAC wasn't cleaned properly. Got %s instead of %s", result, expected)
	}

	// Test uppercase chars
	mac = "00-11-22-33-44-AA"
	result = CleanMAC(mac)

	if result != expected {
		t.Errorf("MAC wasn't cleaned properly. Got %s instead of %s", result, expected)
	}

	// Test invalid MAC (missing one char)
	mac = "00112233445"
	result = CleanMAC(mac)

	if result != "" {
		t.Errorf("Invalid MAC wasn't returned as empty string")
	}

	// Test invalid MAC (last char is invalid)
	mac = "00:11:22:33:44:5z"
	result = CleanMAC(mac)

	if result != "" {
		t.Errorf("Invalid MAC wasn't returned as empty string")
	}

}

func TestValidMAC(t *testing.T) {
	var mac string

	mac = "001122334455"
	if !ValidMAC(mac) {
		t.Error("Valid MAC was declared invalid")
	}

	mac = "0011223344aa"
	if !ValidMAC(mac) {
		t.Error("Valid MAC was declared invalid")
	}

	mac = "aaaaaaaaaaaa"
	if !ValidMAC(mac) {
		t.Error("Valid MAC was declared invalid")
	}

	mac = "aa-aa-aa-aa-aa-aa"
	if !ValidMAC(mac) {
		t.Error("Valid MAC was declared invalid")
	}

	// Invalid chars (zz)
	mac = "aa-aa-aa-aa-aa-zz"
	if ValidMAC(mac) {
		t.Error("Invalid MAC was declared valid")
	}

	// Missing some chars
	mac = "aa-aa-aa-aa-aa"
	if ValidMAC(mac) {
		t.Error("Invalid MAC was declared valid")
	}

	// Too much chars
	mac = "aa-aa-aa-aa-aa-aa-aa"
	if ValidMAC(mac) {
		t.Error("Invalid MAC was declared valid")
	}
}

func TestValidIP(t *testing.T) {
	var ip string

	ip = "1.2.3.4"
	if !ValidIP(ip) {
		t.Error("Valid IP was declared invalid")
	}

	ip = "12.12.12.12"
	if !ValidIP(ip) {
		t.Error("Valid IP was declared invalid")
	}

	ip = "255.255.255.255"
	if !ValidIP(ip) {
		t.Error("Valid IP was declared invalid")
	}

	// Too many
	ip = "1.2.3.4.5"
	if ValidIP(ip) {
		t.Error("Invalid IP was declared valid")
	}

	// Too few
	ip = "1.2.3"
	if ValidIP(ip) {
		t.Error("Invalid IP was declared valid")
	}

	// Not even close to an IP
	ip = "test.that.valid.ip"
	if ValidIP(ip) {
		t.Error("Invalid IP was declared valid")
	}
}

func TestMacFromBytes(t *testing.T) {
	var mac string
	var err error
	mac, err = MacFromBytes([]byte{0x00, 0x01, 0x02, 0x03, 0x04, 0x05})
	if err != nil {
		t.Error(err)
	}
	if mac != "000102030405" {
		t.Error("Invalid mac parsed")
	}
	mac, err = MacFromBytes([]byte{0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0xff})
	if err != nil {
		t.Error(err)
	}
	if mac != "000102030405" {
		t.Error("Invalid mac parsed")
	}
	mac, err = MacFromBytes([]byte{0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f})
	if err != nil {
		t.Error(err)
	}
	if mac != "0a0b0c0d0e0f" {
		t.Error("Invalid mac parsed")
	}
	mac, err = MacFromBytes([]byte{0x0a, 0x0b, 0x0c, 0x0d, 0x0e})
	if err == nil {
		t.Error("Accepted bad data")
	}
}

func TestValidIPv6(t *testing.T) {
	// An IPv4 is not a valid IPv6
	if ValidIPv6("127.0.0.1") {
		t.Error("Invalid IPv6 was declared valid")
	}

	// Test with a valid IPv6
	if !ValidIPv6("2001:db8::1") {
		t.Error("Valid IPv6 was declared invalid")
	}

	// Test with a valid IPv6
	if !ValidIPv6("00::00") {
		t.Error("Valid IPv6 was declared invalid")
	}
}

func TestCleanIPv6(t *testing.T) {
	// Shortened must stay shortened
	if CleanIPv6("2001:db8::1") != "2001:db8::1" {
		t.Error("Cleanup of IPv6 didn't yield expected result")
	}

	// Expanded must yield shortened
	if CleanIPv6("2001:0db8:85a3:0000:0000:8a2e:0370:7334") != "2001:db8:85a3::8a2e:370:7334" {
		t.Error("Cleanup of IPv6 didn't yield expected result")
	}
}

func TestFakeMAC(t *testing.T) {
	var mac string
	var err error

	mac, err = FakeMAC(net.IPv4(192, 168, 5, 3))

	if mac != "0200c0a80503" {
		t.Error("Generated MAC address is not correct", mac)
	}

	if err != nil {
		t.Error("An error was generated although there shouldn't have been one")
	}

	mac, err = FakeMAC(net.IPv4(0, 0, 0, 0))

	if mac != "020000000000" {
		t.Error("Generated MAC address is not correct", mac)
	}

	if err != nil {
		t.Error("An error was generated although there shouldn't have been one")
	}

	mac, err = FakeMAC(net.ParseIP("2001:db8::68"))

	if mac != "" {
		t.Error("Generated MAC address is not correct", mac)
	}

	if err == nil {
		t.Error("An error wasn't generated although there should have been one")
	}

	mac, err = FakeMAC(nil)

	if mac != "" {
		t.Error("Generated MAC address is not correct", mac)
	}

	if err == nil {
		t.Error("An error wasn't generated although there should have been one")
	}
}
