// +build dpi_qosmos

package dpi

import (
	"errors"
	"fmt"

	"git.inverse.ca/inverse/goqosmos/qosmos"
	"github.com/google/gopacket"
)

const QosmosPrefix = "qos:"

type pktData struct {
	metadata Metadata
	data     []byte
	info     gopacket.CaptureInfo
	linkType int
}

type QosmosEngine struct {
	Engine      *qosmos.Engine
	Bundle      *qosmos.Bundle
	workerChans []chan pktData
	callback    func(metadata Metadata, signatures []string)
}

func CreateQosmosEngine(numWorkers int) (*QosmosEngine, error) {
	e := &QosmosEngine{}
	e.Engine = qosmos.EngineCreate(fmt.Sprintf("nb_workers=%d", numWorkers))
	if !e.Engine.Valid() {
		return nil, errors.New("QosmosEngine: Unable to create DPI engine")
	}
	nbWorkers := e.Engine.ConfigGet("nb_workers")

	e.Bundle = e.Engine.BundleCreate()
	if !e.Bundle.Valid() {
		fmt.Println("Failed to create bundle", qosmos.ErrorGetString(e.Bundle, qosmos.ErrorGet()))
		return nil, errors.New("QosmosEngine: Unable to create bundle: " + qosmos.ErrorGetString(e.Bundle, qosmos.ErrorGet()))
	}

	e.Bundle.Activate()

	err := e.Bundle.SignatureEnableAll()
	if err != nil {
		return nil, errors.New("QosmosEngine: Failed to enable all protocols: " + err.Error())
	}

	for i := int64(0); i < nbWorkers; i++ {
		worker := e.Engine.WorkerCreate()
		if !worker.Valid() {
			return nil, errors.New("QosmosEngine: Unable to create worker")
		}
		workerChan := make(chan pktData)
		e.workerChans = append(e.workerChans, workerChan)
		go e.startWorker(worker, workerChan)
	}

	return e, nil
}

func (e *QosmosEngine) startWorker(worker *qosmos.Worker, workerChan chan pktData) {
	for in := range workerChan {
		ret := worker.PDUSet(in.data, in.info.Length, in.info.Timestamp, in.linkType, qosmos.QMDPI_DIR_DEFAULT, 0)
		if ret < 0 {
			panic("Error in worker.PDUSet")
		}

		ret = qosmos.QMDPI_PROCESS_MORE
		for ret == qosmos.QMDPI_PROCESS_MORE {
			result := qosmos.NewResult()
			ret = worker.Process(nil, result)
			if ret < 0 {
				break
			}
			if result.FlagsShouldProcess() {
				sigs := result.Signatures(worker, e.Bundle)
				for i, sig := range sigs {
					sigs[i] = QosmosPrefix + sig
				}
				if e.callback != nil {
					e.callback(in.metadata, sigs)
				}
			}
		}

		//Allow for necessary flow to expire if necessary
		for i := 0; i < 100; i++ {
			result := qosmos.NewResult()
			ret := worker.FlowExpireNext(in.info.Timestamp, result)
			if int(ret) != 0 {
				break
			}

			if result.FlagsShouldProcess() {
				sigs := result.Signatures(worker, e.Bundle)
				if e.callback != nil {
					e.callback(in.metadata, sigs)
				}
			}
		}
	}
}

func (e *QosmosEngine) SetCallback(callback func(metadata Metadata, signatures []string)) {
	e.callback = callback
}

func (e *QosmosEngine) ProcessPkt(metadata Metadata, gp gopacket.Packet, data []byte, info gopacket.CaptureInfo, linkType int) {
	hashKey := qosmos.PacketHashkeyGet(data, len(data), linkType)
	e.workerChans[hashKey%len(e.workerChans)] <- pktData{data: data, info: info, linkType: linkType, metadata: metadata}
}

func (e *QosmosEngine) Destroy() {

}
