package fingerbankclient

import (
	"fmt"
	"net/http"
	"testing"

	"github.com/inverse-inc/go-utils/sharedutils"
)

func init() {
	FingerbankProtocol = sharedutils.EnvOrDefault("FINGERBANK_API_PROTOCOL", FingerbankProtocol)
	FingerbankHost = sharedutils.EnvOrDefault("FINGERBANK_API_HOST", FingerbankHost)
	FingerbankPort = sharedutils.EnvOrDefaultInt("FINGERBANK_API_PORT", FingerbankPort)
	FingerbankKey = sharedutils.EnvOrDefault("FINGERBANK_API_KEY", FingerbankKey)

}

func TestConcurrency(t *testing.T) {
	concurrency := 10
	max := 100
	concurrencyChan := make(chan int, concurrency)

	if FingerbankKey == "" {
		fmt.Println("Skipping TestInterrogate because there is no API key defined")
		return
	}

	for i := 0; i < max; i++ {
		concurrencyChan <- 1
		go func() {
			req, err := http.NewRequest(
				"OPTIONS",
				fingerbankURI("/api/v2"),
				nil,
			)

			resp, err := doQuery(req, []int{200})

			if err != nil {
				t.Error("Error while testing API concurrency", err)
			}

			resp.Body.Close()

			<-concurrencyChan
		}()
	}
}

func TestInterrogate(t *testing.T) {
	if FingerbankKey == "" {
		fmt.Println("Skipping TestInterrogate because there is no API key defined")
		return
	}

	// Test a fully valid query that will gnerate a match on the API
	res, err := Interrogate(InterrogateQuery{
		UserAgents: []string{"Mozilla/5.0 (Linux; Android 10; SM-G973F Build/QP1A.190711.020; wv) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.152 Mobile Safari/537.36"},
	})

	if err != nil {
		t.Errorf("There was an error testing the Fingerbank API %s", err)
	}

	if res.Device.Id != 63933 {
		t.Errorf("Invalid device ID returned from API (%d)", res.Device.Id)
	}

	if len(res.Device.Parents) != 3 {
		t.Errorf("Invalid parent length returned from API (%d)", len(res.Device.Parents))
	}

	// Test a query with unknown parameters (no match on API)
	res, err = Interrogate(InterrogateQuery{
		UserAgents: []string{"not something valid at all"},
	})

	if err != nil {
		t.Errorf("There was an error testing the Fingerbank API %s", err)
	}

	if res.Device.Id != 0 {
		t.Errorf("Invalid device ID returned from API (%d)", res.Device.Id)
	}

	if len(res.Device.Parents) != 0 {
		t.Errorf("Invalid parent length returned from API (%d)", len(res.Device.Parents))
	}

	// Test a query with an invalid key
	previousKey := FingerbankKey
	FingerbankKey = "invalid"
	res, err = Interrogate(InterrogateQuery{
		UserAgents: []string{"not something valid at all"},
	})

	if err == nil {
		t.Error("Query with invalid key didn't generate an error")
	}
	FingerbankKey = previousKey

}

func TestDestinationHosts(t *testing.T) {
	if FingerbankKey == "" {
		fmt.Println("Skipping TestDestinationHosts because there is no API key defined")
		return
	}

	stringHosts, regexpHosts, err := DestinationHosts()

	if err != nil {
		t.Error("Got error while fetching the destination hosts")
	}

	if len(stringHosts) == 0 {
		t.Error("Destination string hosts are empty. Note that they're fetched from the API so make sure it has hosts in it before troubleshooting this")
	}

	if len(regexpHosts) == 0 {
		t.Error("Destination regexp hosts are empty. Note that they're fetched from the API so make sure it has hosts in it before troubleshooting this")
	}
}
