package gobpool

import (
	"bytes"
	"encoding/gob"
	"errors"
	"sync"

	"github.com/google/uuid"
)

type keyedPool struct {
	lock         *sync.Mutex
	freeKeys     []string
	usedKeys     []string
	items        map[string]interface{}
	initial      int
	max          int
	instantiator func() interface{}

	// Tracks the amount of times the pool was full and a get was performed
	timesFull int
}

// Create a new keyedPool
// Only for internal usage of this package
func newKeyedPool(initial, max int, instantiator func() interface{}) *keyedPool {
	kp := &keyedPool{
		lock:         &sync.Mutex{},
		max:          max,
		initial:      initial,
		instantiator: instantiator,
		items:        make(map[string]interface{}),
	}

	kp.freeKeys = make([]string, initial, initial)
	for i := 0; i < initial; i++ {
		k, _ := kp.newItem()
		kp.freeKeys[i] = k
	}

	return kp
}

// Generate a new key and instantiate a value to go with it, then store it in the items
// Doesn't set it free or used, the consumer should do that
func (kp *keyedPool) newItem() (string, interface{}) {
	k := kp.generateKey()
	v := kp.instantiator()
	kp.items[k] = v

	return k, v
}

// Generate a new key for a new item
func (kp *keyedPool) generateKey() string {
	return uuid.New().String()
}

func (kp *keyedPool) Stats() map[string]interface{} {
	return map[string]interface{}{
		"used_items":   len(kp.usedKeys),
		"free_items":   len(kp.freeKeys),
		"current_size": kp.currentSize(),
		"max_items":    kp.max,
		"times_full":   kp.timesFull,
	}
}

// Get the current size of the pool
func (kp *keyedPool) currentSize() int {
	return len(kp.freeKeys) + len(kp.usedKeys)
}

// Find a free key in the pool and expand the pool if needed
// If the pool is full, returns an error
func (kp *keyedPool) getFreeKey() (string, error) {
	kp.lock.Lock()
	defer kp.lock.Unlock()
	var k string

	if len(kp.freeKeys) > 0 {
		k, kp.freeKeys = kp.freeKeys[0], kp.freeKeys[1:]
	} else if kp.currentSize() < kp.max {
		k, _ = kp.newItem()
	} else {
		kp.timesFull++
		return "", errors.New("Unable to find a free key in the pool")
	}

	kp.usedKeys = append(kp.usedKeys, k)
	return k, nil
}

// Find a free item in the pool and get the associated key as well
// If the pool is full, this returns an error
func (kp *keyedPool) Get() (string, interface{}, error) {
	k, err := kp.getFreeKey()

	kp.lock.Lock()
	defer kp.lock.Unlock()

	if err != nil {
		return "", nil, err
	} else {
		return k, kp.items[k], nil
	}
}

// Release an item back into the free items of the pool
// Returns an error when the key isn't part of the pool
func (kp *keyedPool) ReleaseItem(key string) error {
	kp.lock.Lock()
	defer kp.lock.Unlock()

	keyIndex := -1
	for i, k := range kp.usedKeys {
		if k == key {
			keyIndex = i
			break
		}
	}

	if keyIndex == -1 {
		return errors.New("Couldn't find used item with key " + key)
	} else {
		kp.usedKeys = append(kp.usedKeys[:keyIndex], kp.usedKeys[keyIndex+1:]...)
		kp.freeKeys = append(kp.freeKeys, key)
		return nil
	}
}

type gobPoolObject struct {
	buf     *bytes.Buffer
	Encoder *gob.Encoder
	Decoder *gob.Decoder
}

type GobPool struct {
	keyedPool
}

func NewGobPool(initial, max int) *GobPool {
	gep := &GobPool{}
	gep.keyedPool = *newKeyedPool(initial, max, gep.newGob)
	return gep
}

func (gep *GobPool) newGob() interface{} {
	buf := bytes.NewBuffer([]byte{})
	return &gobPoolObject{
		buf:     buf,
		Encoder: gob.NewEncoder(buf),
		Decoder: gob.NewDecoder(buf),
	}
}

func (gep *GobPool) Get() (string, *gobPoolObject, error) {
	var gpo *gobPoolObject
	id, o, err := gep.keyedPool.Get()
	if o != nil {
		gpo = o.(*gobPoolObject)
	}
	return id, gpo, err
}
