package gobpool

import (
	"testing"

	"github.com/inverse-inc/go-utils/sharedutils"
)

func TestKeyedPool(t *testing.T) {
	{
		// Test a pool that is capped to the initial value
		kp := newKeyedPool(5, 5, func() interface{} { return 1 })

		items := make(map[string]int)

		for i := 0; i < 5; i++ {
			k, v, err := kp.Get()
			sharedutils.CheckError(err)

			if _, found := items[k]; found {
				t.Error("Was given twice the same key while getting items in the pool")
			}

			items[k] = v.(int)
		}

		for _, v := range items {
			if v != 1 {
				t.Error("Value that got out of the pool is wrong")
			}
		}

		// Shouldn't be able to get a free item as its capped
		_, _, err := kp.Get()

		if err == nil {
			t.Error("Didn't get an error from the pool when trying to fetch an item in an exhausted pool")
		}

		// Free up items one by one and try to grab them back
		for k, _ := range items {
			err := kp.ReleaseItem(k)
			sharedutils.CheckTestError(t, err)

			k2, _, err := kp.Get()
			sharedutils.CheckTestError(t, err)

			if k != k2 {
				t.Error("Didn't get the freed up key after releasing it")
			}
		}
	}

	{
		// Test a pool that can expand
		kp := newKeyedPool(5, 10, func() interface{} { return 1 })

		items := make(map[string]int)

		for i := 0; i < 6; i++ {
			k, v, err := kp.Get()
			sharedutils.CheckError(err)

			if _, found := items[k]; found {
				t.Error("Was given twice the same key while getting items in the pool")
			}

			items[k] = v.(int)
		}

		for _, v := range items {
			if v != 1 {
				t.Error("Value that got out of the pool is wrong")
			}
		}

		for i := 0; i < 4; i++ {
			k, v, err := kp.Get()
			sharedutils.CheckError(err)

			if _, found := items[k]; found {
				t.Error("Was given twice the same key while getting items in the pool")
			}

			items[k] = v.(int)
		}

		// release all items in the pool and grab them back
		for k, _ := range items {
			kp.ReleaseItem(k)
		}

		newItems := make(map[string]int)

		for i := 0; i < 10; i++ {
			k, v, err := kp.Get()
			sharedutils.CheckError(err)
			newItems[k] = v.(int)
		}

		if len(newItems) != len(items) {
			t.Error("Refetching all items from the pool that was freed didn't yield the same amount of items")
		}

		for k, v := range items {
			if v2, found := newItems[k]; !found {
				t.Error("A new key has appeared in the pool although it shouldn't have")
			} else if v != v2 {
				t.Error("For the same key, the value in the pool differed")
			}
		}
	}
}
