package fcarphandler

import (
	"encoding/hex"
	"fmt"
	"net"

	"git.inverse.ca/inverse/fingerbank-collector/common"

	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
)

type ARPHandler struct {
	fccommon.GenericHandler
}

const (
	Name       = "arp"
	DefaultBPF = "arp"
)

// Create a new handler
func NewARPHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &ARPHandler{}
	eh.Collector = collector
	return eh
}

// Handle function to process a packet
// info is the PacketInfo associated to this connection
// l4Payload is the bytes of the TCP or UDP packet
// rawPacket is the bytes of the packet as received on the wire
func (h *ARPHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	packet := gopacket.NewPacket(rawPacket, layers.LayerTypeEthernet, gopacket.Default)

	if arpLayer := packet.Layer(layers.LayerTypeARP); arpLayer != nil {
		arp, _ := arpLayer.(*layers.ARP)

		// Source handling
		hexSHA := make([]byte, hex.EncodedLen(len(arp.SourceHwAddress)))
		hex.Encode(hexSHA, arp.SourceHwAddress)
		sourceHwAddress := string(hexSHA)
		sourceProtAddress := net.IPv4(arp.SourceProtAddress[0], arp.SourceProtAddress[1], arp.SourceProtAddress[2], arp.SourceProtAddress[3]).String()
		if sourceHwAddress != "000000000000" && sourceProtAddress != "0.0.0.0" {
			err := h.Collector.UpdateIPMAC(sourceProtAddress, sourceHwAddress)
			if err != nil {
				fmt.Println(err)
				return
			}
		}

		// Destination handling
		hexDHA := make([]byte, hex.EncodedLen(len(arp.DstHwAddress)))
		hex.Encode(hexDHA, arp.DstHwAddress)
		dstHwAddress := string(hexDHA)
		dstProtAddress := net.IPv4(arp.DstProtAddress[0], arp.DstProtAddress[1], arp.DstProtAddress[2], arp.DstProtAddress[3]).String()
		if dstHwAddress != "000000000000" && dstProtAddress != "0.0.0.0" {
			err := h.Collector.UpdateIPMAC(dstProtAddress, dstHwAddress)
			if err != nil {
				fmt.Println(err)
				return
			}
		}
	}

	fmt.Println("ARPHandler has handled the packet")
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *ARPHandler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *ARPHandler) String() string {
	return "ARP Handler"
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h *ARPHandler) Name() string {
	return Name
}
