package fcdhcpv6handler

import (
	"encoding/binary"
	"errors"
	"flag"
	"net"
	"strconv"
	"strings"

	"git.inverse.ca/inverse/fingerbank-collector/common"
)

const (
	Name       = "dhcpv6"
	DefaultBPF = "udp and (port 546 or port 547)"
)

const (
	BOOTREQUEST = 1
	BOOTREPLY   = 2
)

var PROCESS_SOLICIT bool = false

const (
	SOLICIT             = 1
	ADVERTISE           = 2
	REQUEST             = 3
	CONFIRM             = 4
	RENEW               = 5
	REBIND              = 6
	REPLY               = 7
	RELEASE             = 8
	DECLINE             = 9
	RECONFIGURE         = 10
	INFORMATION_REQUEST = 11
	RELAY_FORW          = 12
	RELAY_REPL          = 13
)

// DHCPv6Option rerpresents a DHCP option.
type DHCPv6Option struct {
	Type   uint16
	Length uint16
	Data   []byte
}

type DHCPv6Message struct {
	MessageType   uint8
	TransactionId []byte
	Options       []DHCPv6Option
}

type DHCPv6Relay struct {
	MessageType uint8
	HopCount    uint8
	LinkAddress net.IP
	PeerAddress net.IP
	Options     []DHCPv6Option
}

// Create a new handler
func NewDHCPv6Handler(collector fccommon.Collector) fccommon.ProtocolHandler {
	h := &DHCPv6Handler{fccommon.GenericHandler{Collector: collector}}
	return h
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *DHCPv6Handler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *DHCPv6Handler) String() string {
	return "DHCPv6Handler Handler"
}

// Name that represents this protocol handler
func (h *DHCPv6Handler) Name() string {
	return Name
}

func DHCPv6OptionFromBytes(option *DHCPv6Option, data []byte) error {
	if len(data) < 4 {
		return errors.New("Invalid length for option")
	}
	option.Type = binary.BigEndian.Uint16(data[0:2])
	option.Length = binary.BigEndian.Uint16(data[2:4])
	if len(data) < int(option.Length) {
		return errors.New("Invalid length for option")
	}
	option.Data = data[4 : option.Length+4]
	return nil
}

func DHCPv6MessageFromBytes(dhcp *DHCPv6Message, data []byte) error {
	dhcp.MessageType = data[0]
	length := len(data)
	dhcp.TransactionId = data[1:4]
	start := 4
	for start < length {
		option := DHCPv6Option{}
		err := DHCPv6OptionFromBytes(&option, data[start:])
		if err != nil {
			return err
		}
		dhcp.Options = append(dhcp.Options, option)
		start += 4 + int(option.Length)
	}
	return nil
}

func DHCPv6RelayFromBytes(dhcp *DHCPv6Relay, data []byte) error {
	dhcp.MessageType = data[0]
	dhcp.HopCount = data[1]
	length := len(data)
	dhcp.LinkAddress = net.IP(data[2:18])
	dhcp.PeerAddress = net.IP(data[18:34])
	start := 4
	for start < length {
		option := DHCPv6Option{}
		err := DHCPv6OptionFromBytes(&option, data[start:])
		if err != nil {
			return err
		}
		dhcp.Options = append(dhcp.Options, option)
		start += 4 + int(option.Length)
	}
	return nil
}

type DHCPv6Handler struct {
	fccommon.GenericHandler
}

// The handler for DHCP packets
func (h *DHCPv6Handler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	messageType := l4Payload[0]
	collector := h.Collector
	switch messageType {
	case RELAY_FORW, RELAY_REPL:
		handleRelayMessage(collector, info, l4Payload, rawPacket)
	case SOLICIT:
		if PROCESS_SOLICIT {
			handleMessage(collector, info, l4Payload, rawPacket)
		}
	case REQUEST:
		handleMessage(collector, info, l4Payload, rawPacket)
	case ADVERTISE, CONFIRM, RENEW, REBIND, REPLY, RELEASE, DECLINE, RECONFIGURE, INFORMATION_REQUEST:
	}
}

func handleRelayMessage(collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	var dhcp DHCPv6Relay
	err := DHCPv6RelayFromBytes(&dhcp, l4Payload)
	if err != nil {
		return
	}
	switch dhcp.MessageType {
	case SOLICIT:
		handleRelayForw(&dhcp, collector, info, l4Payload, rawPacket)
	case ADVERTISE:
		handleRelayRepl(&dhcp, collector, info, l4Payload, rawPacket)
	}
}

func handleMessage(collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	var dhcp DHCPv6Message
	err := DHCPv6MessageFromBytes(&dhcp, l4Payload)
	if err != nil {
		return
	}
	switch dhcp.MessageType {
	case SOLICIT:
		handleSolicit(&dhcp, collector, info, l4Payload, rawPacket)
	case ADVERTISE:
		handleAdvertise(&dhcp, collector, info, l4Payload, rawPacket)
	case REQUEST:
		handleRequest(&dhcp, collector, info, l4Payload, rawPacket)
	case CONFIRM:
		handleConfirm(&dhcp, collector, info, l4Payload, rawPacket)
	case RENEW:
		handleRenew(&dhcp, collector, info, l4Payload, rawPacket)
	case REBIND:
		handleRebind(&dhcp, collector, info, l4Payload, rawPacket)
	case REPLY:
		handleReply(&dhcp, collector, info, l4Payload, rawPacket)
	case RELEASE:
		handleRelease(&dhcp, collector, info, l4Payload, rawPacket)
	case DECLINE:
		handleDecline(&dhcp, collector, info, l4Payload, rawPacket)
	case RECONFIGURE:
		handleReconfigure(&dhcp, collector, info, l4Payload, rawPacket)
	case INFORMATION_REQUEST:
		handleInformationRequest(&dhcp, collector, info, l4Payload, rawPacket)
	}
}

type endPointAttributes struct {
	clientId    string
	enterprise  string
	host        string
	fingerprint string
}

func getEndPointAttributes(attrs *endPointAttributes, options []DHCPv6Option) error {
	for _, o := range options {
		switch o.Type {
		case 1:
			duid_type := binary.BigEndian.Uint16(o.Data[0:2])
			offset := 0
			switch duid_type {
			case 1:
				offset = 8
			case 2:
				offset = 6
			case 3:
				offset = 4
			default:
				return errors.New("Invalid DUID type")
			}
			mac, err := fccommon.MacFromBytes(o.Data[offset:])
			if err != nil {
				return err
			}
			attrs.clientId = mac
		case 6:
			var tmp []string
			length := int(o.Length)
			for start := 0; start < length; start += 2 {
				tmp = append(tmp, strconv.FormatUint(uint64(binary.BigEndian.Uint16(o.Data[start:start+2])), 10))
			}
			attrs.fingerprint = strings.Join(tmp, ",")
		case 16:
			attrs.enterprise = strconv.FormatUint(uint64(binary.BigEndian.Uint32(o.Data[0:4])), 10)
		case 39:
			if len(o.Data) > 1 {
				length := int(o.Data[1] + 2)
				attrs.host = string(o.Data[2:length])
			}
		}
	}
	return nil
}

func updateEndpoint(dhcp *DHCPv6Message, collector fccommon.Collector) {

	var attrs endPointAttributes
	err := getEndPointAttributes(&attrs, dhcp.Options)
	if err != nil {
		return
	}
	if attrs.clientId == "" {
		return
	}
	ed, err := collector.FindEndpointDataMAC(attrs.clientId)
	if err != nil {
		return
	}
	ed.Update(func(ed *fccommon.EndpointData) bool {
		changed := false
		if ed.Dhcp6Fingerprint != attrs.fingerprint || ed.Dhcp6Enterprise != attrs.enterprise || ed.Hostname != attrs.host {
			changed = true
		}

		ed.Dhcp6Fingerprint = attrs.fingerprint
		ed.Dhcp6Enterprise = attrs.enterprise
		ed.Hostname = attrs.host
		return changed
	})

}

func handleRequest(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	updateEndpoint(dhcp, collector)
}

func handleReply(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleSolicit(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
	updateEndpoint(dhcp, collector)
}

func handleAdvertise(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleConfirm(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleRenew(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleRebind(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleRelease(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleDecline(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleReconfigure(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleInformationRequest(dhcp *DHCPv6Message, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleRelayForw(dhcp *DHCPv6Relay, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func handleRelayRepl(dhcp *DHCPv6Relay, collector fccommon.Collector, info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {
}

func init() {
	flag.BoolVar(&PROCESS_SOLICIT, "dhcpv6-process-solicit", false, "Process the DHCPv6 solicit packet also")
}
