package fcradiushandler

import (
	"encoding/binary"
	"strconv"
	"strings"
)

type capabilities struct {
	name string
	num  uint16
}

var cdpCapabilitiesMap = []capabilities{
	{"Router", 0x001},
	{"Trans-Bridge", 0x002},
	{"Source-Route-Bridge", 0x004},
	{"Switch", 0x008},
	{"Host", 0x010},
	{"IGMP", 0x020},
	{"Repeater", 0x040},
	{"VoIP-Phone", 0x080},
	{"Remotely-Managed-Device", 0x100},
	{"Supports-STP-Dispute", 0x200},
	{"Two-port Mac Relay", 0x400},
}

var lldpCapabilitiesMap = []capabilities{
	{"other", 0x001},
	{"repeater", 0x002},
	{"bridge", 0x004},
	{"wlanAccessPoint", 0x008},
	{"router", 0x010},
	{"telephone", 0x020},
	{"docsisCableDevice", 0x040},
	{"stationOnly", 0x080},
}

// tlvList contain the global tlv dictionary
type TlvList struct {
	Tlvlist map[string]TlvId
}

type TlvId struct {
	TlvType map[int]TlvType
}

type TlvType struct {
	Desc   string
	Decode tlvDataType
	Extra  tlvExtra
}

var Tlv = TlvList{
	Tlvlist: map[string]TlvId{
		"lldp-tlv": {
			map[int]TlvType{
				0: TlvType{"end-of-lldpdu", tlvZeroSize, extractString},
				1: TlvType{"chassis-id", tlvNstring, extractString},
				2: TlvType{"port-id", tlvBlob, extractString},
				3: TlvType{"time-to-live", tlvRangeByte, extractString},
				4: TlvType{"port-description", tlvNstring, extractString},
				5: TlvType{"system-name", tlvNstring, extractString},
				6: TlvType{"system-description", tlvNstring, extractString},
				7: TlvType{"system-capabilities", tlvShort, extractLLDPCapabilities},
				8: TlvType{"management-address", tlvIpAddr, extractString},
			},
		},
		"cdp-tlv": {
			map[int]TlvType{
				1:  TlvType{"device-name", tlvNstring, extractString},
				2:  TlvType{"address-type", tlvIpAddr, extractString},
				3:  TlvType{"port-id-type", tlvNstring, extractString},
				4:  TlvType{"cdp-capabilities-type", tlvShort, extractCDPCapabilities},
				5:  TlvType{"version-type", tlvNstring, extractString},
				6:  TlvType{"platform-type", tlvNstring, extractString},
				7:  TlvType{"ipprefix-type", tlvIpAddr, extractString},
				8:  TlvType{"protocol-hello-type", tlvNstring, extractString},
				9:  TlvType{"vtp-mgmt-domain-type", tlvNstring, extractString},
				10: TlvType{"native-vlan-type", tlvNstring, extractString},
				11: TlvType{"duplex-type", tlvBool, extractDuplex},
				14: TlvType{"voice-vlan-reply", tlvNstring, extractString},
				15: TlvType{"voice-vlan-query", tlvInt8, extractInt},
				16: TlvType{"power-type", tlvInt8, extractInt},
				17: TlvType{"mtu-type", tlvNstring, extractString},
				18: TlvType{"trust-type", tlvNstring, extractString},
				19: TlvType{"cos-type", tlvNstring, extractString},
				22: TlvType{"mgmt-address-type", tlvIpAddr, extractString},
				23: TlvType{"external_port_id_type", tlvNstring, extractString},
				24: TlvType{"power-request-type", tlvNstring, extractString},
				25: TlvType{"power-available-type", tlvInt8, extractString},
				27: TlvType{"unidirectional-mode-type", tlvNstring, extractString},
				28: TlvType{"secondport-status-type", tlvBool, extractStatusType},
			},
		},
		"http-tlv": {
			map[int]TlvType{
				1: TlvType{"user-agent", tlvNstring, extractString},
			},
		},
		"dhcp-option": {
			map[int]TlvType{
				1:   TlvType{"subnet-mask", tlvIpAddr, extractString},
				2:   TlvType{"time-offset", tlvSTime, extractString},
				3:   TlvType{"routers", tlvIpAddr, extractString},
				4:   TlvType{"time-servers", tlvIpAddr, extractString},
				5:   TlvType{"name-servers", tlvIpAddr, extractString},
				6:   TlvType{"domain-name-servers", tlvIpAddr, extractString},
				7:   TlvType{"log-servers", tlvIpAddr, extractString},
				8:   TlvType{"cookie-servers", tlvIpAddr, extractString},
				9:   TlvType{"lpr-servers", tlvIpAddr, extractString},
				10:  TlvType{"impress-servers", tlvIpAddr, extractString},
				11:  TlvType{"resource-location-servers", tlvIpAddr, extractString},
				12:  TlvType{"host-name", tlvNstring, extractString},
				13:  TlvType{"boot-size", tlvShort, extractString},
				14:  TlvType{"merit-dump", tlvNstring, extractString},
				15:  TlvType{"domain-name", tlvNstring, extractString},
				16:  TlvType{"swap-server", tlvIpAddr, extractString},
				17:  TlvType{"root-path", tlvNstring, extractString},
				18:  TlvType{"extensions-path", tlvNstring, extractString},
				19:  TlvType{"ip-forwarding", tlvBool, extractString},
				20:  TlvType{"non-local-source-routing", tlvBool, extractString},
				21:  TlvType{"policy-filters", tlvIpAddr, extractString},
				22:  TlvType{"max-dgram-reassembly", tlvShort, extractString},
				23:  TlvType{"default-ip-ttl", tlvRangeByte, extractString},
				24:  TlvType{"path-mtu-aging-timeout", tlvSTime, extractString},
				25:  TlvType{"path-mtu-plateau-tables", tlvRangeShort, extractString},
				26:  TlvType{"interface-mtu", tlvRangeShort, extractString},
				27:  TlvType{"all-subnets-local", tlvBool, extractString},
				28:  TlvType{"broadcast-address", tlvIpAddr, extractString},
				29:  TlvType{"perform-mask-discovery", tlvBool, extractString},
				30:  TlvType{"mask-supplier", tlvBool, extractString},
				31:  TlvType{"router-discovery", tlvBool, extractString},
				32:  TlvType{"router-solicitation-address", tlvIpAddr, extractString},
				33:  TlvType{"static-routes", tlvIpAddr, extractString},
				34:  TlvType{"trailer-encapsulation", tlvBool, extractString},
				35:  TlvType{"arp-cache-timeout", tlvSTime, extractString},
				36:  TlvType{"ethernet-encapsulation", tlvBool, extractString},
				37:  TlvType{"default-tcp-ttl", tlvRangeByte, extractString},
				38:  TlvType{"tcp-keepalive-interval", tlvSTime, extractString},
				39:  TlvType{"tcp-keepalive-garbage", tlvBool, extractString},
				40:  TlvType{"nis-domain", tlvNstring, extractString},
				41:  TlvType{"nis-servers", tlvIpAddr, extractString},
				42:  TlvType{"ntp-servers", tlvIpAddr, extractString},
				43:  TlvType{"vendor-encapsulated-options", tlvBlob, extractString},
				44:  TlvType{"netbios-name-servers", tlvIpAddr, extractString},
				45:  TlvType{"netbios-dd-servers", tlvIpAddr, extractString},
				46:  TlvType{"netbios-node-type", tlvRangeByte, extractString},
				47:  TlvType{"netbios-scope", tlvNstring, extractString},
				48:  TlvType{"font-servers", tlvIpAddr, extractString},
				49:  TlvType{"x-display-managers", tlvIpAddr, extractString},
				50:  TlvType{"requested-address", tlvIpAddr, extractString},
				51:  TlvType{"lease-time", tlvSTime, extractString},
				52:  TlvType{"option-overload", tlvOverload, extractString},
				53:  TlvType{"dhcp-message-type", tlvMessage, extractString},
				54:  TlvType{"server-identifier", tlvIpAddr, extractString},
				55:  TlvType{"parameter-request-list", tlvInt8, extractFingerPrint},
				56:  TlvType{"message", tlvNstring, extractString},
				57:  TlvType{"max-message-size", tlvShort, extractString},
				58:  TlvType{"renewal-time", tlvSTime, extractString},
				59:  TlvType{"rebinding-time", tlvSTime, extractString},
				60:  TlvType{"class-identifier", tlvNstring, extractString},
				61:  TlvType{"client-identifier", tlvBlob, extractString},
				62:  TlvType{"netware-ip-domain", tlvNstring, extractString},
				63:  TlvType{"netware-ip-information", tlvBlob, extractString},
				64:  TlvType{"nis-plus-domain", tlvNstring, extractString},
				65:  TlvType{"nis-plus-servers", tlvIpAddr, extractString},
				66:  TlvType{"tftp-server", tlvNstring, extractString},
				67:  TlvType{"boot-file", tlvNstring, extractString},
				68:  TlvType{"mobile-ip-home-agents", tlvIpAddr, extractString},
				69:  TlvType{"smtp-servers", tlvIpAddr, extractString},
				70:  TlvType{"pop3-servers", tlvIpAddr, extractString},
				71:  TlvType{"nntp-servers", tlvIpAddr, extractString},
				72:  TlvType{"www-servers", tlvIpAddr, extractString},
				73:  TlvType{"finger-servers", tlvIpAddr, extractString},
				74:  TlvType{"irc-servers", tlvIpAddr, extractString},
				75:  TlvType{"streettalk-servers", tlvIpAddr, extractString},
				76:  TlvType{"streettalk-directory-assistance-servers", tlvIpAddr, extractString},
				77:  TlvType{"user-class-id", TlvTypeCn, extractString},
				78:  TlvType{"slp-directory-agent", tlvBlob, extractString},
				79:  TlvType{"slp-service-scope", tlvBlob, extractString},
				80:  TlvType{"rapid-commit", tlvZeroSize, extractString},
				81:  TlvType{"client-fqdn", tlvBlob, extractString},
				82:  TlvType{"relay-agent-info", tlvBlob, extractString},
				83:  TlvType{"i-sns", tlvBlob, extractString},
				85:  TlvType{"nds-servers", tlvIpAddr, extractString},
				86:  TlvType{"nds-tree", tlvNstring, extractString},
				87:  TlvType{"nds-context", tlvNstring, extractString},
				88:  TlvType{"bcmcs-servers-d", tlvDnsName, extractString},
				89:  TlvType{"bcmcs-servers-a", tlvIpAddr, extractString},
				90:  TlvType{"authentication", tlvBlob, extractString},
				91:  TlvType{"lq-client-last-transaction-time", tlvSTime, extractString},
				92:  TlvType{"lq-associated-ip", tlvIpAddr, extractString},
				93:  TlvType{"pxe-client-arch", tlvShort, extractString},
				94:  TlvType{"pxe-client-network-id", tlvBlob, extractString},
				95:  TlvType{"ldap-url", tlvNstring, extractString},
				97:  TlvType{"pxe-client-machine-id", tlvBlob, extractString},
				98:  TlvType{"user-auth", tlvNstring, extractString},
				99:  TlvType{"geo-conf-civic", tlvBlob, extractString},
				112: TlvType{"netinfo-parent-server-address", tlvIpAddr, extractString},
				113: TlvType{"netinfo-parent-server-tag", tlvNstring, extractString},
				114: TlvType{"initial-url", tlvNstring, extractString},
				116: TlvType{"auto-configure", tlvRangeByte, extractString},
				117: TlvType{"name-service-search", tlvShort, extractString},
				118: TlvType{"subnet-selection", tlvIpAddr, extractString},
				119: TlvType{"domain-search", tlvBlob, extractString},
				120: TlvType{"sip-servers", tlvBlob, extractString},
				121: TlvType{"classless-static-route", tlvBlob, extractString},
				122: TlvType{"cablelabs-client-configuration", tlvBlob, extractString},
				123: TlvType{"geo-conf", tlvBlob, extractString},
				124: TlvType{"v-i-vendor-class", tlvZeroSize, extractString},
				125: TlvType{"v-i-vendor-opts", tlvZeroSize, extractString},
				128: TlvType{"mcns-security-server", tlvIpAddr, extractString},
				161: TlvType{"cisco-leased-ip", tlvIpAddr, extractString},
				162: TlvType{"cisco-client-requested-host-name", tlvNstring, extractString},
				163: TlvType{"cisco-client-last-transaction-time", tlvZeroSize, extractString},
				185: TlvType{"vpn-id", tlvBlob, extractString},
				220: TlvType{"subnet-alloc", tlvBlob, extractString},
				221: TlvType{"cisco-vpn-id", tlvBlob, extractString},
				251: TlvType{"cisco-auto-configure", tlvRangeByte, extractString},
				255: TlvType{"end", tlvZeroSize, extractString},
			},
		},
	},
}

type tlvDataType interface {
	Value(a []byte) interface{}
	String(a []byte) string
}

type tlvExtra interface {
	extra(Type tlvDataType, a []byte) interface{}
}

var extractString extractStringt

type extractStringt struct{}

func (s extractStringt) extra(Type tlvDataType, a []byte) interface{} {
	return Type.String(a)
}

var extractDuplex extractDuplext

type extractDuplext struct{}

func (s extractDuplext) extra(Type tlvDataType, a []byte) interface{} {
	switch Type.String(a) {
	case "0":
		return "Half-Duplex"
	default:
		return "Full-Duplex"
	}
}

var extractStatusType extractStatusTypet

type extractStatusTypet struct{}

func (s extractStatusTypet) extra(Type tlvDataType, a []byte) interface{} {
	switch Type.Value(a).(bool) {
	case false:
		return "Down"
	default:
		return "Up"
	}
}

var extractFingerPrint extractFingerPrintt

type extractFingerPrintt struct{}

func (s extractFingerPrintt) extra(Type tlvDataType, a []byte) interface{} {
	var tmp []string
	for _, b := range a {
		tmp = append(tmp, strconv.FormatUint(uint64(b), 10))
	}
	fingerprint := strings.Join(tmp, ",")
	return fingerprint
}

var extractInt extractIntt

type extractIntt struct{}

func (s extractIntt) extra(Type tlvDataType, a []byte) interface{} {
	return binary.BigEndian.Uint16(Type.Value(a).([]byte))
}

var extractCDPCapabilities extractCDPCapabilitiest

type extractCDPCapabilitiest struct{}

func (s extractCDPCapabilitiest) extra(Type tlvDataType, a []byte) interface{} {
	Options := make(map[string]bool)
	for attribute := range cdpCapabilitiesMap {
		if (binary.BigEndian.Uint32(a) & uint32(cdpCapabilitiesMap[attribute].num)) != 0 {
			Options[cdpCapabilitiesMap[attribute].name] = true
		}

	}
	return Options
}

var extractLLDPCapabilities extractLLDPCapabilitiest

type extractLLDPCapabilitiest struct{}

func (s extractLLDPCapabilitiest) extra(Type tlvDataType, a []byte) interface{} {
	Options := make(map[string]bool)
	for attribute := range lldpCapabilitiesMap {
		if (binary.BigEndian.Uint32(a) & uint32(lldpCapabilitiesMap[attribute].num)) != 0 {
			Options[lldpCapabilitiesMap[attribute].name] = true
		}

	}
	return Options
}

var tlvRangeByte tlvRangeBytet

type tlvRangeBytet struct{}

func (s tlvRangeBytet) Value(a []byte) interface{} {
	return a
}
func (s tlvRangeBytet) String(a []byte) string {
	return "string"
}

var tlvBlob tlvBlobt

type tlvBlobt struct{}

func (s tlvBlobt) Value(a []byte) interface{} {
	return a
}
func (s tlvBlobt) String(a []byte) string {
	return string(a)
}

var tlvNstring tlvNstringt

type tlvNstringt struct{}

func (s tlvNstringt) Value(a []byte) interface{} {
	return a
}
func (s tlvNstringt) String(a []byte) string {
	return string(a)
}

var tlvShort tlvShortt

type tlvShortt struct{}

func (s tlvShortt) Value(a []byte) interface{} {
	return a
}
func (s tlvShortt) String(a []byte) string {
	return "string"
}

var tlvIpAddr tlvIpAddrt

type tlvIpAddrt struct{}

func (s tlvIpAddrt) Value(a []byte) interface{} {
	return a
}
func (s tlvIpAddrt) String(a []byte) string {
	return "string"
}

var tlvSTime tlvSTimet

type tlvSTimet struct{}

func (s tlvSTimet) Value(a []byte) interface{} {
	return a
}
func (s tlvSTimet) String(a []byte) string {
	return "string"
}

var tlvBool tlvBoolt

type tlvBoolt struct{}

func (s tlvBoolt) Value(a []byte) interface{} {
	switch a[0] {
	case 1:
		return true
	default:
		return false
	}
}
func (s tlvBoolt) String(a []byte) string {
	switch a[0] {
	case 1:
		return "1"
	default:
		return "0"
	}
}

var tlvTime tlvTimet

type tlvTimet struct{}

func (s tlvTimet) Value(a []byte) interface{} {
	return a
}
func (s tlvTimet) String(a []byte) string {
	return "string"
}

var tlvRangeShort tlvRangeShortt

type tlvRangeShortt struct{}

func (s tlvRangeShortt) Value(a []byte) interface{} {
	return a
}
func (s tlvRangeShortt) String(a []byte) string {
	return "string"
}

var tlvOverload tlvOverloadt

type tlvOverloadt struct{}

func (s tlvOverloadt) Value(a []byte) interface{} {
	return a
}
func (s tlvOverloadt) String(a []byte) string {
	return "string"
}

var tlvMessage tlvMessaget

type tlvMessaget struct{}

func (s tlvMessaget) Value(a []byte) interface{} {
	return a
}
func (s tlvMessaget) String(a []byte) string {
	return "string"
}

var tlvInt8 tlvInt8t

type tlvInt8t struct{}

func (s tlvInt8t) Value(a []byte) interface{} {
	return a
}
func (s tlvInt8t) String(a []byte) string {
	return string(binary.BigEndian.Uint16(a))
}

var TlvTypeCn TlvTypeCnt

type TlvTypeCnt struct{}

func (s TlvTypeCnt) Value(a []byte) interface{} {
	return a
}
func (s TlvTypeCnt) String(a []byte) string {
	return "string"
}

var tlvZeroSize tlvZeroSizet

type tlvZeroSizet struct{}

func (s tlvZeroSizet) Value(a []byte) interface{} {
	return a
}
func (s tlvZeroSizet) String(a []byte) string {
	return "string"
}

var tlvDnsName tlvDnsNamet

type tlvDnsNamet struct{}

func (s tlvDnsNamet) Value(a []byte) interface{} {
	return a
}
func (s tlvDnsNamet) String(a []byte) string {
	return "string"
}

var tlvVendorClass tlvVendorClasst

type tlvVendorClasst struct{}

func (s tlvVendorClasst) Value(a []byte) interface{} {
	return a
}
func (s tlvVendorClasst) String(a []byte) string {
	return "string"
}

var tlvVendorOptions tlvVendorOptionst

type tlvVendorOptionst struct{}

func (s tlvVendorOptionst) Value(a []byte) interface{} {
	return a
}
func (s tlvVendorOptionst) String(a []byte) string {
	return "string"
}

var tlvNoLenght tlvNoLenghtt

type tlvNoLenghtt struct{}

func (s tlvNoLenghtt) Value(a []byte) interface{} {
	return a
}
func (s tlvNoLenghtt) String(a []byte) string {
	return "string"
}
