package fctcphandler

import (
	"encoding/binary"
	"errors"
	"fmt"
	"net"
	"time"

	"git.inverse.ca/inverse/fingerbank-collector/common"
	"github.com/google/gopacket"
	"github.com/google/gopacket/layers"
	cache "github.com/patrickmn/go-cache"
)

const (
	DefaultBPF = "tcp[tcpflags] & (tcp-syn|tcp-ack) != 0"
	Name       = "tcp"
	PrettyName = "TCP Handler"
)

const (
	//Config Consts
	/* Maximum number of TCP options we will process (< 256): */
	MAX_TCP_OPT = 24

	//IP common
	/* Protocol versions: */
	IP_VER4 = 0x04
	IP_VER6 = 0x06
	/* IP-level ECN: */
	IP_TOS_CE  = 0x01 /* Congestion encountered          */
	IP_TOS_ECT = 0x02 /* ECN supported                   */
	PROTO_TCP  = 0x06

	//TCP Flags
	TCP_FIN  = 0x01
	TCP_SYN  = 0x02
	TCP_RST  = 0x04
	TCP_PUSH = 0x08
	TCP_ACK  = 0x10
	TCP_URG  = 0x20
	/* ECN stuff: */
	TCP_ECE    = 0x40 /* ECN supported (SYN) or detected */ //Reported through tcp.ECN
	TCP_CWR    = 0x80 /* ECE acknowledgment              */ //Reported through tcp.CWR
	TCP_NS_RES = 0x01 /* ECE notification via TCP        */ //REPORTED through tcp.NS
	/* Notable options: */ //NOTE: All reported through tcp.Options:
	TCPOPT_EOL             = 0 /* End of options (1)              */ //EndList
	TCPOPT_NOP             = 1 /* No-op (1)                       */ //NOP
	TCPOPT_MAXSEG          = 2 /* Maximum segment size (4)        */ //MSS
	TCPOPT_WSCALE          = 3 /* Window scaling (3)              */ //WindowScale
	TCPOPT_SACKOK          = 4 /* Selective ACK permitted (2)     */ //SACKPermitted
	TCPOPT_SACK            = 5 /* Actual selective ACK (10-34)    */ //SACK
	TCPOPT_TSTAMP          = 8 /* Timestamp (10)                  */ //Timestamps

)

var ZERO_IP net.IP = net.IPv4(0, 0, 0, 0)

type TCPHandler struct {
	fccommon.GenericHandler
	seenSigs          *cache.Cache
	seenSigsThreshold int
}

// Create a new handler
func NewTCPHandler(collector fccommon.Collector) fccommon.ProtocolHandler {
	eh := &TCPHandler{}
	eh.Collector = collector

	return eh
}

func (h *TCPHandler) InitRateLimit(threshold int, ignoreDuration time.Duration) {
	h.seenSigs = cache.New(ignoreDuration, 10*time.Minute)
	h.seenSigsThreshold = threshold
}

// Handle function to process a packet
// info is the PacketInfo associated to this connection
// l4Payload is the bytes of the TCP or UDP packet
// rawPacket is the bytes of the packet as received on the wire
func (h *TCPHandler) Handle(info fccommon.PacketInfo, l4Payload []byte, rawPacket []byte) {

	if h.seenSigs != nil {
		if o, found := h.seenSigs.Get(info.SrcIP); found {
			if o.(int) >= h.seenSigsThreshold {
				return
			}
		} else {
			h.seenSigs.SetDefault(info.SrcIP, 0)
		}
	}

	signature := TCPSignature{}
	err := h.ExtractSignature(rawPacket, &signature)
	if err != nil {
		return
	}
	ed, err := h.Collector.FindEndpointDataIP(info.SrcIP)
	if err != nil {
		return
	}
	if signature.TcpType == (TCP_ACK | TCP_SYN) {
		ed.AddTCPSynAckSignature(signature.String())
		if h.seenSigs != nil {
			h.seenSigs.IncrementInt(info.SrcIP, 1)
		}
	} else {
		ed.AddTCPSynSignature(signature.String())
		if h.seenSigs != nil {
			h.seenSigs.IncrementInt(info.SrcIP, 1)
		}
	}

}

func (h *TCPHandler) ExtractSignature(rawPacket []byte, signature *TCPSignature) error {
	var eth layers.Ethernet
	var ip4 layers.IPv4
	var ip6 layers.IPv6
	var tcp layers.TCP
	parser := gopacket.NewDecodingLayerParser(layers.LayerTypeEthernet, &eth, &ip4, &ip6, &tcp)
	decoded := []gopacket.LayerType{}
	err := parser.DecodeLayers(rawPacket, &decoded)
	if err != nil {
		return err
	}
	for _, layerType := range decoded {
		switch layerType {
		case layers.LayerTypeEthernet:
		case layers.LayerTypeIPv4:
			signature.Ver = 4
			signature.Ittl = ip4.TTL
			if (ip4.Flags & layers.IPv4EvilBit) == layers.IPv4EvilBit {
				signature.Quirks |= QuirkNotZeroMustBeZero
			}

			if (ip4.Flags & layers.IPv4DontFragment) == layers.IPv4DontFragment {
				signature.Quirks |= QuirkDontFragment
				if ip4.Id != 0 {
					signature.Quirks |= QuirkNonZeroId
				}
			} else if ip4.Id == 0 {
				signature.Quirks |= QuirkZeroId
			}
		case layers.LayerTypeIPv6:
			signature.Ver = 6
			signature.Ittl = ip6.HopLimit
			if ip6.FlowLabel != 0 {
				signature.Quirks |= QuirkFlow
			}

			if (ip6.TrafficClass & 0x03) != 0 {
				signature.Quirks |= QuirkECN
			}
		case layers.LayerTypeTCP:
			if tcp.SYN {
				signature.TcpType |= TCP_SYN
			}
			if tcp.ACK {
				signature.TcpType |= TCP_ACK
			}
			if signature.TcpType != TCP_SYN && signature.TcpType != TCP_SYN|TCP_ACK {
				return errors.New("Do not process this type of packet")
			}
			signature.WindowSize = int32(tcp.Window)
			if tcp.ECE || tcp.CWR || tcp.NS {
				signature.Quirks |= QuirkECN
			}
			if tcp.Seq == 0 {
				signature.Quirks |= QuirkZeroSeq
			}

			if tcp.ACK {
				if tcp.Ack == 0 {
					signature.Quirks |= QuirkZeroAck
				}
			} else {
				if tcp.Ack != 0 && !tcp.RST {
					signature.Quirks |= QuirkNonZeroAck
				}
			}

			if tcp.URG {
				signature.Quirks |= QuirkUrgent
			} else {
				if tcp.Urgent != 0 {
					signature.Quirks |= QuirkNonZeroUrgent
				}
			}

			if tcp.PSH {
				signature.Quirks |= QuirkPush
			}

			optionsData := tcp.Contents[20:]
			start, end := 0, len(optionsData)
		LOOP_OPTIONS:
			for start < end {
				oType := layers.TCPOptionKind(optionsData[start])

				option := layers.TCPOption{OptionType: oType}
				var length uint8 = 0
				switch oType {
				case layers.TCPOptionKindEndList:
					length = 1
					signature.OptionPad = uint8(end - start)
					for _, tmpByte := range optionsData[start:end] {
						if tmpByte != 0 {
							signature.Quirks |= QuirkOptionNonZeroEol
							break
						}
					}
					start = end
				case layers.TCPOptionKindNop:
					length = 1
				case layers.TCPOptionKindWindowScale:
					if start+3 > end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("layers.TCPOptionKindWindowScale break loop\n")
						break LOOP_OPTIONS
					}
					length = optionsData[start+1]
					if length != 3 {
						signature.Quirks |= QuirkOptionBad
					}
					length = 3
					signature.Scale = int16(optionsData[start+2])
				case layers.TCPOptionKindSACKPermitted:
					if start+2 > end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("layers.TCPOptionKindSACKPermitted break loop\n")
						break LOOP_OPTIONS
					}
					length = optionsData[start+1]
					if length != 2 {
						fmt.Printf("layers.TCPOptionKindSACKPermitted break loop\n")
						signature.Quirks |= QuirkOptionBad
					}
					length = 2
				case layers.TCPOptionKindMSS:
					if start+4 > end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("layers.TCPOptionKindMSS break loop\n")
						break LOOP_OPTIONS
					}
					length = optionsData[start+1]
					if length != 4 {
						signature.Quirks |= QuirkOptionBad
					}
					signature.MSS = int32(binary.BigEndian.Uint16(optionsData[start+2 : start+4]))
				case layers.TCPOptionKindSACK:
					if (start + 1) == end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("Bad length %d\n", oType)
						break LOOP_OPTIONS
					}
					length = optionsData[start+1]
					if length < 10 || length > 34 {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("Bad length %d\n", oType)
						break LOOP_OPTIONS
					}
					if start+int(length) > end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("is too long %d\n, oType", start+int(length))
						break LOOP_OPTIONS
					}

				case layers.TCPOptionKindTimestamps:
					if start+10 > end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("Bad length %d\n", oType)
						break LOOP_OPTIONS
					}
					length = optionsData[start+1]
					if length != 10 {
						signature.Quirks |= QuirkOptionBad
					}
					signature.Timestamp1 = binary.BigEndian.Uint32(optionsData[start+2 : start+6])
					if signature.Timestamp1 == 0 {
						signature.Quirks |= QuirkOptionZeroTS1
					}
					signature.Timestamp2 = binary.BigEndian.Uint32(optionsData[start+6 : start+10])
					if tcp.SYN && signature.Timestamp2 != 0 {
						signature.Quirks |= QuirkOptionNonZeroTS2
					}
				default:
					if start+1 >= end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("Know length %d\n", oType)
						break LOOP_OPTIONS
					}
					length = optionsData[start+1]
					if length < 2 || length > 40 {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("invalid length %d %d\n", oType, length)
						break LOOP_OPTIONS
					}
					if start+int(length) > end {
						signature.Quirks |= QuirkOptionBad
						fmt.Printf("is too long %d\n, oType", start+int(length))
						break LOOP_OPTIONS
					}
				}
				start += int(length)
				option.OptionLength = length
				signature.Options = append(signature.Options, option)
			}

		default:
			return errors.New("Invalid protocol")
		}
	}
	return nil
}

// Returns the BPF of the packets the protocol handler is interested in
func (h *TCPHandler) GetBPF() string {
	return DefaultBPF
}

// Friendly string representation of this protocol handler
func (h *TCPHandler) String() string {
	return PrettyName
}

// Name that represents this protocol handler
// Should return a value following the golang package naming standards
func (h *TCPHandler) Name() string {
	return Name
}
