package fctcphandler

import (
	"encoding/hex"
	"net"
	"testing"

	"git.inverse.ca/inverse/fingerbank-collector/common"
)

type TestData struct {
	info      fccommon.PacketInfo
	l4Payload string
	raw       string
	sig       string
}

var syntests = []TestData{
	TestData{
		info: fccommon.PacketInfo{
			SrcMAC: net.HardwareAddr{0x00, 0x0c, 0x29, 0xa0, 0x33, 0x12},
			SrcIP:  "172.16.101.135",
			DstMAC: net.HardwareAddr{0x00, 0x50, 0x56, 0xed, 0xa1, 0x5f},
			DstIP:  "192.95.20.194",
		},
		l4Payload: "",
		raw:       "005056eda15f000c29a0331208004500003ccf204000400684e2ac106587c05f14c2b6120050a10f426000000000a0027210e6e70000020405b40402080a0c088d400000000001030307",
		sig:       "4:64+0:0:1460:mss*20,7:mss,sok,ts,nop,ws:df,id+:0",
	},
}

var synacktests = []TestData{
	TestData{
		info: fccommon.PacketInfo{
			DstMAC: net.HardwareAddr{0x00, 0x0c, 0x29, 0xa0, 0x33, 0x12},
			DstIP:  "172.16.101.135",
			SrcMAC: net.HardwareAddr{0x00, 0x50, 0x56, 0xed, 0xa1, 0x5f},
			SrcIP:  "192.95.20.194",
		},
		l4Payload: "",
		raw:       "000c29a03312005056eda15f08004500002c009500008006537ec05f14c2ac1065870050b61276018a67a10f42616012faf01c300000020405b40000",
		sig:       "4:128+0:0:1460:mss*44,0:mss::0",
	},
}

func TestExtractSignature(t *testing.T) {
	h := NewTCPHandler(fccommon.NewFingerbankCollector("")).(*TCPHandler)
	for _, test := range append(syntests, synacktests...) {
		raw, err := hex.DecodeString(test.raw)
		if err != nil {
			t.Error(err)
			continue
		}
		sig := TCPSignature{}
		if err = h.ExtractSignature(raw, &sig); err != nil {
			t.Error(err)
		}
		if sig.String() != test.sig {
			t.Errorf("Extracted signature did not match expected (%s) (%s)", sig.String(), test.sig)
		}
	}
}

func TestHandleSyn(t *testing.T) {
	collector := fccommon.NewFingerbankCollector("")
	h := NewTCPHandler(collector).(*TCPHandler)
	for _, test := range syntests {
		l4Payload, err := hex.DecodeString(test.l4Payload)
		if err != nil {
			t.Error(err)
			continue
		}
		raw, err := hex.DecodeString(test.raw)
		if err != nil {
			t.Error(err)
			continue
		}
		err = collector.UpdateIPMAC(test.info.SrcIP, test.info.SrcMAC.String())
		if err != nil {
			t.Error(err)
			continue
		}
		h.Handle(test.info, l4Payload, raw)
		ep, err := collector.FindEndpointDataIP(test.info.SrcIP)
		if err != nil {
			t.Error(err)
			continue
		}
		if _, ok := ep.TCPSynSignatures[test.sig]; !ok {
			t.Error("Did not add the tcp signature")
			continue
		}
	}
}

func TestHandleSynAck(t *testing.T) {
	collector := fccommon.NewFingerbankCollector("")
	h := NewTCPHandler(collector).(*TCPHandler)
	for _, test := range synacktests {
		l4Payload, err := hex.DecodeString(test.l4Payload)
		if err != nil {
			t.Error(err)
			continue
		}
		raw, err := hex.DecodeString(test.raw)
		if err != nil {
			t.Error(err)
			continue
		}
		err = collector.UpdateIPMAC(test.info.SrcIP, test.info.SrcMAC.String())
		if err != nil {
			t.Error(err)
			continue
		}
		h.Handle(test.info, l4Payload, raw)
		ep, err := collector.FindEndpointDataIP(test.info.SrcIP)
		if err != nil {
			t.Error(err)
			continue
		}
		if _, ok := ep.TCPSynAckSignatures[test.sig]; !ok {
			t.Error("Did not add the tcp signature")
			continue
		}
	}
}

var QuirkTests = []struct {
	quirk  QuirkFlags
	squirk string
}{
	{QuirkDontFragment, "df"},
	{QuirkNonZeroId, "id+"},
	{QuirkZeroId, "id-"},
	{QuirkECN, "ecn"},
	{QuirkNotZeroMustBeZero, "0+"},
	{QuirkFlow, "flow"},

	{QuirkZeroSeq, "seq-"},
	{QuirkNonZeroAck, "ack+"},
	{QuirkZeroAck, "ack-"},
	{QuirkNonZeroUrgent, "uptr+"},
	{QuirkUrgent, "urgf+"},
	{QuirkPush, "pushf+"},

	{QuirkOptionZeroTS1, "ts1-"},
	{QuirkOptionNonZeroTS2, "ts2+"},
	{QuirkOptionNonZeroEol, "opt+"},
	{QuirkOptionLargeWindowScale, "exws"},
	{QuirkOptionBad, "bad"},

	{
		QuirkDontFragment | QuirkNonZeroId | QuirkZeroId | QuirkECN | QuirkNotZeroMustBeZero | QuirkFlow |
			QuirkZeroSeq | QuirkNonZeroAck | QuirkZeroAck | QuirkNonZeroUrgent | QuirkUrgent | QuirkPush |
			QuirkOptionZeroTS1 | QuirkOptionNonZeroTS2 | QuirkOptionNonZeroEol | QuirkOptionLargeWindowScale | QuirkOptionBad,
		"df,id+,id-,ecn,0+,flow,seq-,ack+,ack-,uptr+,urgf+,pushf+,ts1-,ts2+,opt+,exws,bad",
	},
}

func TestQuirks(t *testing.T) {
	for i, test := range QuirkTests {
		if test.quirk.String() != test.squirk {
			t.Errorf("%d %s", i, test.squirk)
		}
	}
}
