package iplist

import (
	"fmt"
	"net"
	"regexp"

	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
	fclog "git.inverse.ca/inverse/fingerbank-collector/log"
)

var regexpComment = regexp.MustCompile(`^#`)
var regexpCIDR = regexp.MustCompile(`[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}/[0-9]{1,2}`)

func DownloadBlacklistMap() ([]string, error) {
	rows, err := fingerbankclient.FetchTextFile("/api/v2/download/ip-blacklist")
	if err != nil {
		return []string{}, err
	}

	cidrs := make([]string, 0)
	for _, line := range rows {
		cidrs = append(cidrs, line)
	}

	return cidrs, nil
}

func BlacklistedIPs() ([]*net.IPNet, error) {
	rawMap, err := DownloadBlacklistMap()
	if err != nil {
		return []*net.IPNet{}, nil
	}

	logger := fclog.Logger("")

	excluded := map[string]bool{
		"0.0.0.0/8":      true,
		"10.0.0.0/8":     true,
		"100.64.0.0/10":  true,
		"127.0.0.0/8":    true,
		"169.254.0.0/16": true,
		"172.16.0.0/12":  true,
		"192.0.2.0/24":   true,
		"192.168.0.0/16": true,
		"224.0.0.0/3":    true,
	}

	ipNets := []*net.IPNet{}
	for _, cidr := range rawMap {
		if regexpComment.MatchString(cidr) || cidr == "" {
			continue
		}

		if _, found := excluded[cidr]; found {
			continue
		}

		if !regexpCIDR.MatchString(cidr) {
			cidr += "/24"
		}

		_, ipNet, err := net.ParseCIDR(cidr)
		if err != nil {
			logger.Error(fmt.Sprintf("Ignoring line '%s' in the IP blacklist because its not a valid CIDR", cidr))
			continue
		}
		ipNets = append(ipNets, ipNet)
	}

	return ipNets, nil
}
