package main

import (
	"context"
	"flag"
	"fmt"
	"net/http"
	"os"
	"os/signal"
	"runtime"
	"runtime/debug"
	"runtime/pprof"
	"strings"
	"sync"
	"syscall"
	"time"

	"github.com/inverse-inc/go-utils/sharedutils"

	// Profiling usage
	_ "net/http/pprof"

	"git.inverse.ca/inverse/fingerbank-collector/api"
	fccommon "git.inverse.ca/inverse/fingerbank-collector/common"
	"git.inverse.ca/inverse/fingerbank-collector/fingerbankclient"
	"git.inverse.ca/inverse/fingerbank-collector/iplist"
	"git.inverse.ca/inverse/fingerbank-collector/keyoverride"
	fclog "git.inverse.ca/inverse/fingerbank-collector/log"
	"git.inverse.ca/inverse/fingerbank-collector/packetfence-checks"
)

var collector *fccommon.FingerbankCollector

var fingerbankApiProtocolParam = flag.String(
	"fingerbank-api-protocol",
	sharedutils.EnvOrDefault("FINGERBANK_API_PROTOCOL", fingerbankclient.FingerbankProtocol),
	"The Fingerbank API protocol (http or https) (FINGERBANK_API_PROTOCOL envrionment variable)",
)

var fingerbankApiHostParam = flag.String(
	"fingerbank-api-host",
	sharedutils.EnvOrDefault("FINGERBANK_API_HOST", fingerbankclient.FingerbankHost),
	"The Fingerbank API hostname (FINGERBANK_API_HOST environment variable)",
)

var fingerbankApiPortParam = flag.Int(
	"fingerbank-api-port",
	sharedutils.EnvOrDefaultInt("FINGERBANK_API_PORT", fingerbankclient.FingerbankPort),
	"The Fingerbank API port (FINGERBANK_API_PORT envirionment variable)",
)

var fingerbankApiKeyParam = flag.String(
	"fingerbank-api-key",
	sharedutils.EnvOrDefault("FINGERBANK_API_KEY", ""),
	"The Fingerbank API key to use to communicate with the API (FINGERBANK_API_KEY environment variable)",
)

var profileServer = flag.Bool("profile", sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_PROFILE", "false")), "Enables the profiling server on port 8081")

var cpuProfile = flag.String("cpu-profile", "", "write cpu profile to file")

var profileMutex = flag.Bool("profile-mutex", sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_PROFILE_MUTEX", "false")), "When using the -profile option, this enables full mutex profiling")

var expireEndpoints = flag.Duration(
	"delete-inactive-endpoints",
	fccommon.MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_DELETE_INACTIVE_ENDPOINTS", "0s")),
	"Deletes the endpoints that haven't been seen for more than this duration. By default without this parameter, endpoints will never expire. (COLLECTOR_DELETE_INACTIVE_ENDPOINTS environment variable)",
)

var endpointsCachePath = flag.String("endpoints-cache-path", sharedutils.EnvOrDefault("COLLECTOR_ENDPOINTS_CACHE_PATH", ""), "When set, this will persist the endpoints cache to the specified file. Will also attempt to restore it from this file on startup.")
var persistEndpointsCache bool

var endpointsDataDbPath = flag.String("endpoints-db-path", sharedutils.EnvOrDefault("COLLECTOR_ENDPOINTS_DB_PATH", ""), "When set, this will persist the endpoints data to the specified file. Will also attempt to restore it from this file on startup.")
var persistEndpointsData bool

var ipMapsDbPath = flag.String("ip-maps-db-path", sharedutils.EnvOrDefault("COLLECTOR_IP_MAPS_DB_PATH", ""), "When set, this will persist the IP maps data to the specified file. Will also attempt to restore it from this file on startup.")
var ipMapsOldnessTolerance = flag.Duration(
	"ip-maps-oldness-tolerance",
	fccommon.MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_IP_MAPS_OLDNESS_TOLERANCE", "24h")),
	"Amount of time that is tolerable for the oldness of the IP database during the restoration. Databases older than this duration will not be restored",
)
var persistIpMaps bool

var dbPersistInterval = flag.Duration(
	"db-persistence-interval",
	fccommon.MustParseDuration(sharedutils.EnvOrDefault("COLLECTOR_DB_PERSISTENCE_INTERVAL", "300s")),
	"Interval at which the collector for to persist the data. Used only when endpoints-db-path or ip-maps-db-path is set",
)

var httpCert = flag.String("http-cert", sharedutils.EnvOrDefault("COLLECTOR_HTTP_CERT", ""), "Certificate to use for HTTPS on the collector API. Leaving this parameter empty leaves HTTPS disabled on the API.")

var httpKey = flag.String("http-key", sharedutils.EnvOrDefault("COLLECTOR_HTTP_KEY", ""), "Key to use for HTTPS on the collector API")

var osFreeMemDiff = flag.Int("gc-os-free-after", sharedutils.EnvOrDefaultInt("COLLECTOR_GC_OS_FREE_AFTER", 50), "Amount of MB of unused heap space after which the collector will force the memory to be released to the OS. In high concurrency environments, you will want to set this to a higher number, but this will lead to a larger memory consumption. A value of 0 will disable this feature and the memory will only be released by the Golang GC")

var disableHostValidator = flag.Bool(
	"disable-host-validator",
	sharedutils.IsEnabled(sharedutils.EnvOrDefault("COLLECTOR_DISABLE_HOST_VALIDATOR", "false")),
	"Disables the host validation when tracking the destination hosts. This means absolutely all destination hosts will be collected (versus only the ones recognized by Fingerbank). This should be used carefully since it can lead to large memory usage.",
)

var logger = fclog.Logger("")
var wg *sync.WaitGroup

func setup() {
	wg = &sync.WaitGroup{}

	if *osFreeMemDiff != 0 {
		go func() {
			var memstats runtime.MemStats
			for {
				runtime.ReadMemStats(&memstats)
				if int(memstats.HeapSys-memstats.HeapAlloc) > (*osFreeMemDiff * 1048576) {
					logger.Info("Forcing the release of the memory to the OS.")
					debug.FreeOSMemory()
				}
				time.Sleep(1 * time.Minute)
			}
		}()
	}

	flag.Parse()

	fingerbankclient.FingerbankProtocol = *fingerbankApiProtocolParam
	fingerbankclient.FingerbankHost = *fingerbankApiHostParam
	fingerbankclient.FingerbankPort = *fingerbankApiPortParam

	fingerbankclient.FingerbankKey = *fingerbankApiKeyParam
	// Override the key if necessary
	fingerbankclient.FingerbankKey = keyoverride.Override(fingerbankclient.FingerbankKey)

	if fingerbankclient.FingerbankKey == "" {
		fmt.Println("Missing Fingerbank API key (-fingerbank-api-key flag or FINGERBANK_API_KEY environment variable)")
		os.Exit(1)
	}

	collector = fccommon.NewFingerbankCollector(fingerbankclient.FingerbankKey)

	setupHostValidator()

	setupPersistence()

	setupExitHandler()

	setupEndpointsExpiration()

	setupBlacklistedIPs()
}

func setupBlacklistedIPs() {
	if collector.NBAEnabled {
		go func() {
			for {
				blacklistedIPsNew, err := iplist.BlacklistedIPs()
				if err != nil {
					logger.Error(fmt.Sprintf("ERROR: Wasn't able to fetch the blacklisted IPs map: %s", err))
					time.Sleep(1 * time.Minute)
					continue
				}
				logger.Info("Loading the blacklisted IPs")
				collector.BlacklistedIPs = blacklistedIPsNew
				time.Sleep(1 * time.Hour)
			}
		}()
	}
}

func setupHostValidator() {
	if !*disableHostValidator {
		logger.Info("Running with destination hosts analysis enabled")
		// Setup host validator from the hosts recognized in the API
		go func() {
			for {
				stringHosts, regexpHosts, err := fingerbankclient.DestinationHosts()
				if err != nil {
					logger.Error(fmt.Sprintf("ERROR: Wasn't able to fetch the destination hosts from the Fingerbank API: %s", err))
					time.Sleep(1 * time.Minute)
					continue
				}
				logger.Info("Loading the destination hosts received from the Fingerbank API")

				fccommon.DefaultHostValidator = fccommon.NewHostValidator(stringHosts, regexpHosts)
				time.Sleep(1 * time.Hour)
			}
		}()
	} else {
		logger.Warn("Running with destination host validation disabled.")
	}
}

func setupPersistence() {
	if *endpointsCachePath != "" {
		persistEndpointsCache = true
		err := collector.RestoreEndpointsCache(*endpointsCachePath)
		if err != nil {
			logger.Error(fmt.Sprintf("ERROR: Unable to restore the endpoints from database %s. Error: %s", *endpointsCachePath, err))
		}
	}

	if *endpointsDataDbPath != "" {
		persistEndpointsData = true
		err := collector.RestoreEndpointsData(*endpointsDataDbPath)
		if err != nil {
			logger.Error(fmt.Sprintf("ERROR: Unable to restore the endpoints from database %s. Error: %s", *endpointsDataDbPath, err))
		}
	}

	if *ipMapsDbPath != "" {
		persistIpMaps = true
		err := collector.RestoreIpMaps(*ipMapsDbPath, *ipMapsOldnessTolerance)
		if err != nil {
			logger.Error(fmt.Sprintf("ERROR: Unable to restore the IP maps from database %s. Error: %s", *endpointsDataDbPath, err))
		}
	}

	if !persistIpMaps && !persistEndpointsData && !persistEndpointsCache {
		logger.Info("Not configured to persist any data")
		return
	}

	wg.Add(1)
	go func() {
		for {
			time.Sleep(*dbPersistInterval)
			persist()
		}
	}()
}

func persist() {
	if persistEndpointsCache {
		start := time.Now()
		logger.Debug(fmt.Sprintf("Persisting endpoints cache to %s", *endpointsCachePath))
		err := collector.PersistEndpointsCache(*endpointsCachePath)
		if err != nil {
			logger.Error(fmt.Sprintf("Failed to persist endpoints cache to %s: %s", *endpointsCachePath, err))
		}
		logger.Info(fmt.Sprintf("Persisted endpoints cache in %s", time.Since(start)))
	}

	if persistIpMaps {
		start := time.Now()
		logger.Debug(fmt.Sprintf("Persisting IP maps to %s", *ipMapsDbPath))
		err := collector.PersistIpMaps(*ipMapsDbPath)
		if err != nil {
			logger.Error(fmt.Sprintf("Failed to persist IP maps to %s: %s", *ipMapsDbPath, err))
		}
		logger.Info(fmt.Sprintf("Persisted IP maps in %s", time.Since(start)))
	}

	if persistEndpointsData {
		start := time.Now()
		logger.Debug(fmt.Sprintf("Persisting endpoints to %s", *endpointsDataDbPath))
		err := collector.PersistEndpointsData(*endpointsDataDbPath)
		if err != nil {
			logger.Error(fmt.Sprintf("Failed to persist endpoints to %s: %s", *endpointsDataDbPath, err))
		}
		logger.Info(fmt.Sprintf("Persisted endpoints in %s", time.Since(start)))
	}

}

func setupEndpointsExpiration() {
	if (*expireEndpoints).Nanoseconds() > 0 {
		logger.Info(fmt.Sprintf("Will expire inactive endpoints inactive for more than %s", (*expireEndpoints).String()))
		wg.Add(1)
		go func() {
			for {
				collector.CleanupInactiveEndpointData(*expireEndpoints)
				time.Sleep(1 * time.Minute)
			}
			wg.Done()
		}()
	}

}

func setupExitHandler() {
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)

	wg.Add(1)
	go func() {
		<-sigs
		logger.Info("Exiting...")
		exitCleanup()
		os.Exit(0)
	}()
}

func exitCleanup() {
	netflow5Service.Stop()
	sflowService.Stop()
	persist()
}

func runAPI() {
	// Set the authorization token to the Fingerbank API key
	api.AuthorizationToken = fingerbankclient.FingerbankKey

	port := sharedutils.EnvOrDefaultInt("PORT", fccommon.DEFAULT_API_PORT)

	bindIPs := strings.Split(os.Getenv("BIND_IPS"), ",")

	router := api.Setup(collector)

	for _, bindIP := range bindIPs {
		go func(bindIP string) {
			listenOn := fmt.Sprintf("%s:%d", bindIP, port)
			if *httpCert != "" {
				collector.ApiUsesHTTPS = true
				panic(router.RunTLS(listenOn, *httpCert, *httpKey))
			} else {
				panic(router.Run(listenOn))
			}
		}(bindIP)
	}
}

// Main entry point of the application
func main() {

	setup()

	if *cpuProfile != "" {
		f, err := os.Create(*cpuProfile)
		if err != nil {
			panic(err)
		}
		pprof.StartCPUProfile(f)
		defer pprof.StopCPUProfile()
	}

	go captureAndProcess()

	if *profileServer {
		wg.Add(1)
		if *profileMutex {
			runtime.SetMutexProfileFraction(1)
		}
		go http.ListenAndServe(":8081", http.DefaultServeMux)
	}

	wg.Add(1)
	go netflow5Service.Start()
	go sflowService.Start()
	go runAPI()

	wg.Add(1)
	go func() {
		fails := 0
		for {
			time.Sleep(300 * time.Second)
			if !packetfence.RunsOnPacketFence(context.Background()) {
				fmt.Println("The collector doesn't seem to be running alongside PacketFence.")
				fails++
			} else {
				// Reset the fails count on every success
				fails = 0
			}

			if fails >= 3 {
				fmt.Println("Failed to validate that the collector is running alongside PacketFence. This binary should be used **exclusively** with PacketFence. If you want to use the Fingerbank collector outside of this scope, contact info@inverse.ca")
				os.Exit(42)
			}
		}
	}()

	wg.Wait()
}
